/*
 Implementation of a serial line driver for Allen Bradley PLCs

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include "plclog.h"

#include "vi_string.h"
#define debug 10

#include <stdio.h>
#include "AB_lib.h"
#define ThisModule "Allen_B  : "

#define tmo_normal 331000
#include "allen_b.h"

#define want_write    
/*
    I do not yet know, how Allen Bradley's addressing types and transfer modes
    fit into a closed concept.
    So, to minimize the chaos, we define directly the CMD and FNC and
    other codes here, leaving blank many "tradional" fields of the transfer
    instruction.
*/

class AllenBradleyTI:public transferInstruction {
  public:
    int CMD;			//      Allen B's CMD code
    int FNC;			//      Allen B's FNC code
    int AB_type;		//      Allen B's data type code
    int AB_file;		//      Allen B's file Nr
    int AB_start;		//      Start Address, Allen B's element no.
    int AB_bytes;		//      how many bytes we want. 
    string AllenBradleyCode;
    coord errCell;
    DECLARE_PERSISTENCE(AllenBradleyTI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(AllenBradleyTI, "ALLENB")

void AllenBradleyTI::Read(Engine & e)
{
/*
    To minimize the chaos, we define directly the CMD and FNC codes here
    So direction field is not used.
*/
    string ctc, stc, erc;
    e >> comment >> CMD >> FNC >> AB_type >> AB_file >> AB_start >>
	AB_bytes >> sheet >> stc >> erc >> ctc;
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
    errCell = getPos(erc);
#if debug >3
//    int rfd=0;				// make LOG happy
    LOG2("comment        :%s\n", comment.c_str());
    LOG2("condition      :%s\n", cellN(ctrlCell));
    LOG2("CMD  	      :%d\n", CMD);
    LOG2("FNC  	      :%d\n", FNC);
    LOG2("type           :%x\n", AB_type);
    LOG2("File           :%d\n", AB_file);
    LOG2("first word     :%d\n", AB_start);
    LOG2("length in bytes:%d\n", AB_bytes);
    LOG2("sheet          :%d\n", sheet);
    LOG2("to Cell        :%s\n", cellN(startCell));
    LOG2("Errors to Cell :%s\n", cellN(errCell));
#endif
}

class AllenBradleyDriver:public AllenBradley {
  public:
    AllenBradleyTI *pa;
    AllenBradleyDriver() {};
    virtual void execTI(transferInstruction * p);
    unsigned short master_send(AB_block *, int size);
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    execList();
	    execCount++;
	}
    };
};

void AllenBradleyDriver::intSetup(DriverInfo * i)
{
    DI=i;
    rfd = setPort(i->s, &wfd);
    LOG2("intSetup fd: %d\n", wfd);
    TypeManager::Add("ALLENB", CreateNewAllenBradleyTI);
    PLCdrv::load(i);
    changedMask = registerDriver("AllenBradley", this);
    errorMask = changedMask;
    TNS = 97 * 256;
};

unsigned short AllenBradleyDriver::master_send(AB_block * ab, int size)
{
    ab->loTNS = TNS % 256;
//    ab->hiTNS=TNS/256;
    TNS++;
    return ab_send(ab, size);    
}

void AllenBradleyDriver::execTI(transferInstruction * p)
{
/*
    first we check whether the condition for execution is not zero:
*/	
    bool ok;
    if (numContents(p->sheet, p->ctrlCell, ok) == 0.0) {
#if debug >2
	LOG3("execTI condition %d:%s not true, nothing to do.\n", p->sheet, cellN(p->ctrlCell));
	FLUSH;
#endif	
	return;
    }
/*
    Ok to execute.
*/	    
    coord cell;
    unsigned char b1[1024];
    unsigned char *bs;
    int ii, errorState = 0, res;
    unsigned short rcv_tns, sentTNS;
    pa = (AllenBradleyTI *) p;
#if debug >2
    LOG3("execTI CMD %x FNC %x\n", pa->CMD, pa->FNC);
    FLUSH;
#endif
/*    
 * "protected typed read instruction with three byte address fields"
 * as AB calls it.
 */
    if ((pa->CMD == 0xF) && (pa->FNC == 0xA2)) {	// it's that sort...
#if debug >2
	LOG1("execTI protected typed read with 3 byte address\n");
#endif
	AB_block abb;
	memset(&abb, 0, sizeof(abb));
	abb.CMD = pa->CMD;
	abb.FNC = pa->FNC;
	abb.byteSize = pa->AB_bytes;
	abb.fileNo = pa->AB_file;
	abb.fileType = pa->AB_type;
	abb.eleNo = 0xFF;
	abb.XeleNo = pa->AB_start;
	sentTNS = master_send(&abb, 14);
	res=readAnswer3(b1,&bs,&errorState);
	
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}    
	
	if (errorState)	goto error;
	AB_block *ab = (AB_block *) & bs[2];
	if ((abb.CMD | 0x40) != ab->CMD) {
	    LOG3("*** error Command:%x reply:%x.\n", abb.CMD, ab->CMD);
	    errorState |= 2;
	    goto error;
	}
	rcv_tns = ab->loTNS + 256 * ab->hiTNS;
#if debug>4
	LOG2("execTI() TNS of answer %d.\n", rcv_tns);
#endif
	if (sentTNS != rcv_tns) {
	    LOG3("*** error wrong TNS: mine %d reply:%d.\n", sentTNS, rcv_tns);
	    errorState |= 4;
	    goto error;
	}
//	sendACK();
#if debug>4
	LOG3("execTI() bytes required: %d got:%d.\n", abb.byteSize, res - 12);
#endif
	short *w = (short *) &(ab->FNC);
	float *wf = (float *) &(ab->FNC);
	cell = pa->startCell;
	switch (pa->AB_type) {   // shouldn't this be PA->type ?!
	case 0x89:
	    setShortValues(w,abb.byteSize/2,pa->sheet,cell);
	    break;
	case 0x8a:
	    setFloatValues(wf,abb.byteSize/4,pa->sheet,cell);
	    break;
	default:
	    LOG2("execTI() unknown file type %x\n",  abb.fileType);
	    errorState |= 8;
	}
    }
    
/*    
 * read "diagnostic status"
 * as AB calls it.
 */
    else if ((pa->CMD == 0x6) && (pa->FNC == 0x03)) {	// it's that sort...
#if debug >2
	LOG1("diagnostic status\n");
#endif
	AB_block abb;
	memset(&abb, 0, sizeof(abb));
	abb.CMD = pa->CMD;
	abb.FNC = pa->FNC;
	sentTNS = master_send(&abb, 7);
	res=readAnswer3(b1,&bs,&errorState);
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}    
	if (errorState)	goto error;
	AB_block *ab = (AB_block *) & bs[2];
	if ((abb.CMD | 0x40) != ab->CMD) {
	    LOG3("*** error Command:%x reply:%x.\n", abb.CMD, ab->CMD);
	    errorState |= 2;
	    goto error;
	}
	rcv_tns = ab->loTNS + 256 * ab->hiTNS;
#if debug>4
	LOG2("execTI() TNS of answer %d.\n", rcv_tns);
#endif
	if (sentTNS != rcv_tns) {
	    LOG3("*** error wrong TNS: mine %d reply:%d.\n", sentTNS, rcv_tns);
	    errorState |= 4;
	    goto error;
	}
//	sendACK();
#if debug>4
	LOG3("execTI() bytes required: %d got:%d.\n", abb.byteSize, res - 12);
#endif
	short *w = (short *) &(ab->FNC);
//	float *wf = (float *) &(ab->FNC);
/*
+    So what to do with diagnostic data? If there is an AB char or byte type, it
+    could be used to tell "save it as bytes". For now if you chose integer, the
+    bytes are written as integer to cells.
+    
+*/		
	cell = pa->startCell;
	switch (pa->AB_type) {   
	case 0x89:
	    setShortValues(w,abb.byteSize/2,pa->sheet,cell);
	    break;
	default:
	    LOG2("execTI() unknown file type %x\n", abb.fileType);
	    errorState |= 8;
	}
    }    
/*    
 * that's another read instruction we support:
 * a "unprotected read instruction" for the Micrologix (1000,1200,1500 ?)
 * Attention: For Micrologix 1000, this reads from file N7
 * For Micrologix 1500, this reads from file N9, which you MUST CREATE FIRST.
 */    
    else if ((pa->CMD == 0x1) /*&& (pa->FNC == 0xA2)*/) {	// command has no FNC
#if debug >2
	LOG1("execTI unprotected read from CIF file\n");
#endif
	AB_block2 abb;
	memset(&abb, 0, sizeof(abb));
	abb.CMD = pa->CMD;
	abb.byteSize = pa->AB_bytes;
	abb.loADR = pa->AB_start % 256;
	abb.hiADR = pa->AB_start / 256;
	sentTNS = master_send( (AB_block*) &abb, 9/*14*/); // is this ok??
	
	res=readAnswer3(b1,&bs,&errorState);
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}
	if (errorState)	{
	    errorState=0;
	    LOG1("sending NAK and retry read.\n");
//	    sendNAK();
	    res=readAnswer3(b1,&bs,&errorState);
	}    
	if (errorState)	goto error;
	AB_block2 *ab = (AB_block2 *) & bs[2];
	if ((abb.CMD | 0x40) != ab->CMD) {
	    LOG3("*** error Command:%x reply:%x.\n", abb.CMD, ab->CMD);
	    errorState |= 2;
	    goto error;
	}
	rcv_tns = ab->loTNS + 256 * ab->hiTNS;
#if debug>4
	LOG2("execTI() TNS of answer %d.\n", rcv_tns);
#endif
	if (sentTNS != rcv_tns) {
	    LOG3("*** error wrong TNS: mine %d reply:%d.\n", sentTNS, rcv_tns);
	    errorState |= 4;
	    goto error;
	}
//	sendACK();
#if debug>4
	LOG3("execTI() bytes required: %d got:%d.\n", abb.byteSize, res - 12);
#endif
	short *w = (short *) &(ab->loADR);
	float *wf = (float *) &(ab->loADR);
	cell = pa->startCell;
	switch (pa->AB_type) { // shouldn't this be pa->type ?!
	case 0x89:
	    setShortValues(w,abb.byteSize/2,pa->sheet,cell);
	    break;
	case 0x8a:			// Is this nonsense or can one copy bytes from float
					// to integer Common Interface File?
	    setFloatValues(wf,abb.byteSize/4,pa->sheet,cell);				
	    break;
	default:
	    LOG2("execTI() unknown file type %x\n", pa->AB_type);
	    errorState |= 8;
	}
    }    
#if defined want_write    
    else if ((pa->CMD == 0xF) && (pa->FNC == 0xAA)) {	// it's a write command
#if debug >1
	LOG1("execTI protected typed write with 3 byte address\n");
	FLUSH;
#endif
/*
    preparations	
*/
        cell = pa->startCell;
        int varSize=0;
/*
    Here we do NOT send the whole stuff in one block. Most cells will not change often.
    Hence we test each cell's change flags.
    If the cell has changed, we write only this value to the corresponding PLC variable.
    First determine the byte size per variable depending on AB's type code:
*/	    
        switch (pa->AB_type) {
	    case 0x89: 
	        varSize = 2; // integers are 2 bytes
	    break; 
	    case 0x8a: 
	        varSize = 4; // floats are 4 bytes
	    break; 
	}
#if debug >1
	LOG2("execTI write varSize:%d \n", varSize);
	FLUSH;
#endif	    
/*
    Here we fill the structure AB_block	with the values that do not depend on the value to
    send:
*/	    
	AB_block abb;
	memset(&abb, 0, sizeof(abb));	
	abb.CMD = pa->CMD;			
	abb.FNC = pa->FNC;
	abb.byteSize = varSize;
	abb.fileNo = pa->AB_file;
	abb.fileType = pa->AB_type;
	abb.eleNo = 0xFF;
	abb.XeleNo = pa->AB_start;
	short *w = (short *) &(abb.data);
	float *wf = (float *) &(abb.data);
#if debug >3		    	    
	LOG3("bytes %d varSize %d \n", pa->AB_bytes,varSize);
#endif	    
	for (ii = 0; ii < pa->AB_bytes; ii += varSize) {
	    int flags = getFlags(p->sheet, cell, ok);
#if debug >3		    
	    LOG4("cell %s flags %x mask %x\n", cellN(cell),flags,changedMask);
	    FLUSH;
#endif		    
		
	    if (flags & changedMask) {
	        float d=numContents(pa->sheet,cell,ok);
	        switch (pa->AB_type) {
		    case 0x89: 
		        *w=(int)d;
#if debug >3		    			    
		        LOG1("before master_send.\n");
		        FLUSH;			    
#endif			    
		        sentTNS = master_send(&abb, 14+2); // integers are 2 bytes
//		        res = readAnswer(b1);
			res = readAnswer3(b1,&bs,&errorState);
//		        sendACK();
			break; 
		    case 0x8a: 
		        *wf=d;			
#if debug >3		    			    
		        LOG1("before master_send.\n");
		        FLUSH;
#endif
		        sentTNS = master_send(&abb, 14+4); // floats are 4 bytes    
//			res = readAnswer(b1);    
			res = readAnswer3(b1,&bs,&errorState);
//		        sendACK();
			break; 
		}
#if debug >3		    
		LOG3("clear flags of cell %s mask %x\n", cellN(cell),changedMask);
		FLUSH;
#endif		    		    
		clrPushButton(p->sheet, cell); // change transmitted, now clear the flag
#if debug >3		    		    
		flags = getFlags(p->sheet, cell, ok);
		LOG3("flags of cell %s = %x\n", cellN(cell),flags);
		FLUSH;
#endif		    
	    }
	    cell.row++;    // Next cell
	    abb.XeleNo++;  // Next PLC variable
	}
    }
    else if (pa->CMD == 0x8) {	// it's an unprotected write command
#if debug >1
	LOG1("execTI unprotected write to CIF file\n");
	FLUSH;
#endif
/*
    preparations:
*/
        execError = false;
        cell = pa->startCell;
        int varSize=2;
/*
    Here we do NOT send the whole stuff in one block. Most cells will not change often.
    Hence we test each cell's change flags.
    If the cell has changed, we write only this value to the corresponding PLC variable.
    First determine the byte size per variable depending on AB's type code:
*/	    
        switch (pa->AB_type) {
	    case 0x89: 
	        varSize = 2; // integers are 2 bytes
		break; 
	    case 0x8a: 
	        varSize = 4; // floats are 4 bytes
		break; 
	}
#if debug >1
	LOG2("execTI write varSize:%d \n", varSize);
	FLUSH;
#endif	    
/*
    Here we fill the structure AB_block	with the values that do not depend on the value to
    send:
*/	    
	AB_block2 abb;
	memset(&abb, 0, sizeof(abb));	
	abb.CMD = pa->CMD;		
	abb.byteSize = varSize;
	int PLCadr = pa->AB_start;
	short *w = (short *) &(abb.byteSize); 	// There is no size field here
	float *wf = (float *) &(abb.byteSize);	// There is no size field here
#if debug >3		    	    
	LOG3("bytes %d varSize %d \n",pa->AB_bytes,varSize);
#endif
	for (ii = 0; ii < pa->AB_bytes; ii += varSize) {
#if debug >3		    	    
	    LOG4("bytes %d varSize %d ii %d\n",pa->AB_bytes,varSize,ii);
#endif	    
	    int flags = getFlags(p->sheet, cell, ok);
#if debug >3		    
	    LOG4("testing cell %s flags %x mask %x\n",cellN(cell),flags,changedMask);
	    FLUSH;
#endif		    
		
	    if (flags & changedMask) {
	        abb.loADR = PLCadr % 256;
	        abb.hiADR = PLCadr / 256;
	        float d=numContents(pa->sheet,cell,ok);
#if debug >3		    
	        LOG3("cell %s changed to %g \n",cellN(cell),d);
	        FLUSH;
#endif		    
	        switch (pa->AB_type) {
		    case 0x89: 
		        *w=(int)d;
#if debug >3		    			    
		        LOG1("before master_send.\n");
		        FLUSH;			    
#endif			    
		        sentTNS = master_send( (AB_block*) &abb, 8+2/*14*/); // is this ok??
			res = readAnswer3(b1,&bs,&errorState);
//		        sendACK();
			break; 
		    case 0x8a: 
		        *wf=d;			
#if debug >3		    			    
		        LOG1("before master_send.\n");
		        FLUSH;
#endif
		        sentTNS = master_send( (AB_block*) &abb, 8+4/*14*/); // is this ok??	
			res = readAnswer3(b1,&bs,&errorState);
//			res = readAnswer(b1);
//		        sendACK();
			break; 
		}
#if debug >3		    
		LOG3("clear flags of cell %s mask %x\n",cellN(cell),changedMask);
		FLUSH;
#endif		    		    		    
		clrPushButton(p->sheet, cell); // change transmitted, now clear the flag
#if debug >3		    		    
		flags = getFlags(p->sheet, cell, ok);
		LOG3("flags of cell %s = %x\n",cellN(cell),flags);
		FLUSH;
#endif		    		    
	    }
	    cell.row++;    // Next cell
	    PLCadr+=1;  // Next PLC variable
	}
    }
#endif    
    else {			// implemented commands end here
      error:
	LOG4("*** error %d CMD %d FNC %d !\n", errorState, pa->CMD, pa->FNC);
	setCell(pa->sheet, pa->errCell, errorState, soIoState);
	int res = readChars(b1, tmo_normal);
//	if (res) sendACK();
	usleep(500000);
	res = readChars(b1, tmo_normal);
//	if (res) sendACK();
	usleep(500000);
	res = readChars(b1, tmo_normal);
//	if (res) sendACK();
	usleep(500000);
//	sendACK();
    }				
    setCell(pa->sheet, pa->errCell, errorState, soIoState);
    LOG3("I set Cell %s to %d\n", cellN(pa->errCell),  errorState);
    usleep(21000 * 5);		// try to give time to other threads??
}

#include "thread_makro.cc"
#if debug>0
DTDebug(AllenBradleyDriver)
#else
DT(AllenBradleyDriver)
#endif
/*
    Changes:
    
    02/23/2003  got rid of unused vars
    11/06/2004	keep driverinfo
*/
