/*
 Implementation of a serial line driver for Allen Bradley PLCs

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include "plclog.h"

#include "vi_string.h"

#define debug 4

#include <stdio.h>
#include "AB_lib.h"

#define ThisModule "Allen_BSl: "
#define tmo_normal 999999
//#define sim_broken_transport
#include "allen_b.h"


/*
    The Slaves maps an AB file No and to a sheet and a column, beginning
    at startCell. So lets assume we get a "read 20 bytes as integers from 
    N7:50" command. Then we look for a TI that maps file 7 somewhere. 
    Lets assume we find one that maps file 7 to sheet 1, startCell Q17.
    Now the first value is in Q17 corresponding to N7:0.
    We start reading data at Q67 corresponding to N7:50.
    As 20 bytes are 10 Integers, we take them from Q67 to Q76.
    
    Maybe we could define condition (we wouldn't respond unless it's true)
    and error cell. Error could be inactivity on the connection.
*/

class AB_Slave_TI:public transferInstruction {
  public:
    int AB_file;		//      Allen B's file Nr
    DECLARE_PERSISTENCE(AB_Slave_TI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(AB_Slave_TI, "ALLENB_SL")

void AB_Slave_TI::Read(Engine & e)
{
    string ctc, stc, erc;
    e >> comment >> AB_file >> sheet >> stc;
    startCell = getPos(stc);
#if debug >3
    LOG2("comment        :%s\n", comment.c_str());
    LOG2("File           :%d\n", AB_file);
    LOG2("sheet          :%d\n", sheet);
    LOG2("Map to Column  :%s\n", cellN(startCell));
#endif
}

class AllenBradleySlave:public AllenBradley {
  public:
    AllenBradleySlave() { };
    void execSlave();
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    execSlave();
	    execCount++;
	}
    }
};

void AllenBradleySlave::intSetup(DriverInfo * i)
{
    rfd= setPort(i->s, &wfd);
    LOG2("intSetup write fd: %d\n",wfd);
    TypeManager::Add("ALLENB_SL", CreateNewAB_Slave_TI);
    PLCdrv::load(i);
    changedMask = registerDriver("AllenBradleyS", this);
    errorMask = changedMask;
};

void AllenBradleySlave::execSlave()
{
    LOG1("execSlave() starts.\n");
    FLUSH;
    double rvalue;
    coord cell;
    uc b1[1024];
    uc *bs;
    int i,errorState=0;
    bool ok;
    int res=readAnswer3(b1,&bs,&errorState);
/*    
    if ((b1[0]==DLE)&&(b1[1]==NAK)) {
	 LOG1("got NAK. retransmit.\n");
	 reTransmit();
    }	 
*/    
    if (!errorState){
	AB_block *ab = (AB_block *) & bs[2];
	LOG4("dest: %x source: %x CMD %x\n",ab->dst, ab->src, ab->CMD);
	LOG4("FNC %x FNC %d tns: %d. \n",ab->FNC, ab->FNC, ab->loTNS);       
	if ((ab->CMD == 0xF) && (ab->FNC == 0xa2)) {
//	    sendACK();
	    LOG2(ThisModule"can handle this in principle. Do i have a TFI for file %d?\n",
		   ab->fileNo);
	    AB_Slave_TI *ent=NULL;
	    for (i = 0; i < il->count; i++) {
		ent = (AB_Slave_TI *) (il->items[i]);
#if debug >3
		LOG4("entry: %d %d:%s\n",
		       ent->AB_file, ent->sheet, cellN(ent->startCell)
		    );
#endif
		if ((ent->AB_file == ab->fileNo)) {
		    LOG2(ThisModule"entry no. %d is the right statement\n", i);
		    FLUSH;
		    break;
		}
		ent = NULL;
	    }
//          sx=ent->insertValues();     
	    LOG2("entry: %p is the right statement\n", ent);
	    if (ent) {
		int offset = ab->eleNo;
		if (offset == 0xff) {
		    offset = ab->XeleNo;
		}
		cell = ent->startCell;
		int byteSize = ab->byteSize;
		unsigned char *b = bs;
		short *w = (short *) &(ab->FNC);
		ab->CMD |= 0x40;
		LOG2("entry: %p is the right statement\n",
		       ent);
		switch (ab->fileType) {
		case 0x89:
		    cell.row += offset;
		    LOG2("data begins at %s\n",
			   cellN(cell));

		    for (i = 0; i < byteSize; i += 2) {
			rvalue = numContents(1, cell, ok);
			*w = int (rvalue);
			LOG3("data from %s=%d\n",
			       cellN(cell), *w);
			w++;
			cell.row++;
		    }
#if debug >3
		    res = byteSize + 8;
#include "dumpchars.cc"
#endif
		    ab_send(ab, byteSize + 6);
		    break;

		default:
		    LOG2("cannot handle data type %x!\n", ab->fileType);
//                  sendNAK(); 
		}
	    } else {
		LOG1(ThisModule
		       "No TFI. Don't know where to take data from!\n");
//              sendNAK(); 
	    }
	}
	if (ab->CMD == 0x1) {
//	    sendACK();
	    LOG1(ThisModule"can handle this in principle. Do i have a TFI for CIF?\n");
	    AB_Slave_TI *ent=NULL;
	    for (i = 0; i < il->count; i++) {
		ent = (AB_Slave_TI *) (il->items[i]);
#if debug >3
		LOG4("entry: %d %d:%s\n",
		       ent->AB_file, ent->sheet, cellN(ent->startCell)
		);
#endif
		if ((ent->AB_file == 0)) {
		    LOG2(ThisModule"entry no. %d is the right statement\n", i);
		    FLUSH;
		    break;
		}
		ent = NULL;
	    }
//          sx=ent->insertValues();     
	    LOG2("entry: %p is the right statement\n", ent);
	    if (ent) {
		AB_block2 *ab2= (AB_block2*)ab;
		int offset = ab2->loADR+256*ab2->hiADR;
		cell = ent->startCell;
		int byteSize = ab2->byteSize;
		unsigned char *b = bs;
		short *w = (short *) &(ab->FNC);
		ab->CMD |= 0x40;
		LOG2("entry: %p is the right statement\n",
		       ent);
//		switch (ent->AB_type) {
//		case 0x89:
		    cell.row += offset;
		    LOG2("data begins at %s\n",
			   cellN(cell));

		    for (i = 0; i < byteSize; i += 2) {
			rvalue = numContents(1, cell, ok);
			*w = int (rvalue);
			LOG3("data from %s=%d\n",
			       cellN(cell), *w);
			w++;
			cell.row++;
		    }
#if debug >3
		    res = byteSize + 8;
#include "dumpchars.cc"
#endif
		    ab_send(ab, byteSize + 6);
	    } else {
		LOG1(ThisModule
		       "No TFI. Don't know where to take data from!\n");
//              sendNAK(); 
	    }
	}
	if (ab->CMD == 0x8) {
//	    sendACK();
	    LOG1(ThisModule"can handle this in principle. Do i have a TFI for CIF?\n");
	    AB_Slave_TI *ent=NULL;
	    for (i = 0; i < il->count; i++) {
		ent = (AB_Slave_TI *) (il->items[i]);
#if debug >3
		LOG4("entry: %d %d:%s\n",
		       ent->AB_file, ent->sheet, cellN(ent->startCell)
		);
#endif
		if ((ent->AB_file == 0)) {
		    LOG2(ThisModule"entry no. %d is the right statement\n", i);
		    FLUSH;
		    break;
		}
		ent = NULL;
	    }
//          sx=ent->insertValues();     
	    LOG2("entry: %p is the right statement\n", ent);
	    if (ent) {
/*	    
		AB_block2 *ab2= (AB_block2*)ab;
		int offset = ab2->loADR+256*ab2->hiADR;
*/		
		AB_block2 *ab2= (AB_block2*)(rcvdMessage+2);
		cell = ent->startCell;
		int offset=ab2->loADR + 256* ab2->hiADR;
		LOG3("offset: %d size %d\n",offset,(res-14)/2);
		cell.row+=offset;
		LOG2("first cell: %s\n",cellN(cell));
		short * w=(short*) (&ab2->byteSize);
		setShortValues(w,(res-14)/2,ent->sheet,cell);
		ab->CMD |= 0x40;
#if debug >3
		    res = 6;
		    uc *b=(uc*)ab2;
#include "dumpchars.cc"
#endif
		    ab_send((AB_block*)ab2, res);
		    
	    } else {
		LOG1(ThisModule
		       "No TFI. Don't know where to take data from!\n");
//              sendNAK(); 
	    }
	}
	
    }
}

#include "thread_makro.cc"
#if debug>0
DTDebug(AllenBradleySlave)
#else
DT(AllenBradleySlave)
#endif

