/*
 Implementation of a PLC driver for GE Fanuc SNPX protocol.
 Was originally written for and tested with ALSTOM/Cegelec ALSPA C80-35 PLC. 
 Then I found out, that ALSTOM just sells GE Fanuc 90-30 components under
 this name. Everything is identical up to the part numbers.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include <stdio.h>
#include "plclog.h"

#include "vi_string.h"

#define debug 15

#define ThisModule "Alspadrv: "

class AlspaTI:public transferInstruction {
  public:
    int lastSheet;		// last of multiple sheets when we are 
    // connected to multiple PLCs on RS485.
    coord SNPcell;		// cell we get the SNP id from.
    DECLARE_PERSISTENCE(AlspaTI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(AlspaTI, "ALSPA")

void AlspaTI::Read(Engine & e)
{
    string stc, ctc, nc;
    e >> comment >> sheet >> lastSheet >> startMW >> endMW >> direction >>
	DBNumber >> stc >> ctc >> nc;
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
    SNPcell = getPos(nc);
#if debug >3
    LOG2("comment    :%s\n", comment.c_str());
    LOG2("first sheet:%d\n", sheet);
    LOG2("last sheet :%d\n", lastSheet);
    LOG2("start Adr  :%d\n", startMW);
    LOG2("end   Adr  :%d\n", endMW);
    LOG2("direction  :%d\n", direction);
    LOG2("DBNumber   :%d\n", DBNumber);
    LOG2("to Cell    :%s\n", cellN(startCell));
    LOG2("condition  :%s\n", cellN(ctrlCell));
    LOG2("SNP ID from:%s\n", cellN(SNPcell));
#endif
}

#pragma pack (1)

struct AlspaMessage {
    unsigned char esc;		// allways ESCape 
    unsigned char code;		// allways 0x58 = 'X'
    unsigned char SNP_id[8];	// identify PLC on a RS485 bus
    unsigned char read_write;	// direction
    unsigned char typ;
    word adr;
    word size;
    word w1;
    unsigned char b17;
    unsigned char ddd[8];
};

#pragma pack (1)

struct AlspaAnswer {
    char esc;
    char code;
    char q;
    char a[6];
    char by[1024];
};

class AlspaDriver:public PLCdrv {
  public:
    word DriverNr;
    AlspaTI *pa;
    word DriverState;
    int first, last;
    char aMachine[9];
    AlspaAnswer fa;
    AlspaMessage msg;
    AlspaDriver() {
    };
    virtual void execTI(transferInstruction * p);
    int readChars(int size/*, int *real_size*/);
    bool readCharsLog(int size, char lev1/*, int *real_size*/);
    void sendAlspa(int len);
    bool AlspaOpen();
    virtual void intSetup(DriverInfo * i);
    void execute() {
	int i;
//      PID=getpid();
	while (1) {
	    for (i = first; i <= last; i++) {
		AlspaOpen();
		execList();
	    }
	    execCount++;
	}
    };
    void coreOfsetLoop(int &i, coord & cell);
    void writeByte(word a, char &b, char t);
    void readByte(word a, char *b, char t);
};

void set1stLast(void *ti, void *ad)
{
    if (((AlspaDriver *) ad)->first > ((AlspaTI *) ti)->sheet)
	((AlspaDriver *) ad)->first = ((AlspaTI *) ti)->sheet;
    if (((AlspaDriver *) ad)->last < ((AlspaTI *) ti)->lastSheet)
	((AlspaDriver *) ad)->last = ((AlspaTI *) ti)->lastSheet;
}

void AlspaDriver::intSetup(DriverInfo * i)
{
    aMachine[0] = 0;
    rfd = setPort(i->s, &wfd);
    LOG2("intSetup write fd: %d\n", wfd);
    TypeManager::Add("ALSPA", CreateNewAlspaTI);
    PLCdrv::load(i);
    first = maxSheet;		// no instruction should have a higher number here
    last = 1;			// no instruction should have a lower number here
    changedMask = registerDriver("ALSPA", this);
    errorMask = changedMask;
//
// check the instruction List for the real highest and lowest sheet number:
//    
    il->forEach(set1stLast, this);
    LOG3("first sheet: %d last sheet: %d\n", first, last);
};

void AlspaDriver::coreOfsetLoop(int &i, coord & cell)
{
    LOG1("enter set Loop\n");
    word mVal, j;
    j = 0;
    while (i <= pa->endMW) {
	LOG4("while i %d j:%d  pa->endMW: %d\n", i, j,
	       pa->endMW);
	switch (pa->DBNumber) {
	case 2:
	case 3:
	    mVal = *((word *) & fa.by[j]);
	    break;
	case 1:
	    mVal = fa.by[j];
	    break;
	case 0:
	    mVal = fa.by[j >> 3] & (1 << (j & 7));
	}
	setCell(pa->sheet, cell, mVal, soExternal);
	LOG4("set Cell %d:%s to %d\n", pa->sheet, cellN(cell),
	       mVal);
	cell.row++;
	j++;
	i++;
    }
    LOG1("leave set Loop\n");
}


#define tmo_answ 291
#define tmo_read 291
#define tmo_open 291



int AlspaDriver::readChars(int size/*, int *real_size*/)
{
    fd_set FDS;
    bool charsAvail;
    unsigned char msb, sum, *b;
    int i, res;
//    *real_size = 0;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
    usleep(1000 * tmo_read);
    FD_ZERO(&FDS);
    FD_SET(rfd, &FDS);
    charsAvail = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
    LOG2("readChars %d\n", charsAvail);
    if (!charsAvail) {
	LOG3("%s timeout %d\n", aMachine, charsAvail);
    }
    if (charsAvail) {
	LOG2("readChars reading up to %d chars.\n", size);
	res = read(rfd, &fa, size);
/*	*real_size = res; */
	b = (unsigned char *) &fa;
#include "dumpchars.cc"
	LOG2("readChars %d chars read.\n", res);
	sum = 0;
	for (i = 0; i < res - 1; i++) {
//          LOG2(ThisModule"readChars %d \n",i);
	    sum = sum ^ b[i];
	    msb = sum & 0x80;
	    sum = sum << 1;
	    if (msb != 0)
		sum |= 1;
	}
#if debug > 0
	int chs = b[res - 1];
	LOG3("Checksums %d , %d\n", chs, int (sum));
#endif
	LOG3("%s got %d chars\n", aMachine, res);
	FLUSH;
	return (res);
    } else
	return (0);
}

bool AlspaDriver::readCharsLog(int size, char lev1/*, int *real_size*/)
{
    int rs;
    bool rcl;
#if debug > 0
    LOG1("Readchars_log\n");
    FLUSH;
#endif
    rs = readChars(size/*, real_size*/);
    if (rs != size) {
	lastError.level1 = lev1;
	lastError.level2 = rs;
	lastError.time = time(NULL);
    }
    rcl = (rs != size);
#if debug > 0
    LOG3("readCharsLog: %d:%d\n", rs, size);
    FLUSH;
#endif				/*  */
    return (rcl);
}

void AlspaDriver::sendAlspa(int len)
{
    int i;
    unsigned char msb, sum;
#if debug > 0
    LOG1("sendAlspa enter.\n");
    FLUSH;
#endif				/*  */

    msg.esc = 0x1b;
    msg.code = 0x58;
    for (i = 0; i < 8; i++) {
//        msg.SNP_id[i] = aMachine[i];
	msg.SNP_id[i]=0;
    }
    msg.b17 = 0x17;
    sum = 0;
    char *f = (char *) &msg;
    for (i = 0; i < len; i++) {
	sum = sum ^ ((unsigned char *) f)[i];
	msb = sum & 0x80;
	sum = sum << 1;
	if (msb != 0)
	    sum = sum | 1;
    }
    ((unsigned char *) f)[len] = sum;
    int res = write(wfd, f, len + 1);
    uc* b = (unsigned char *) f;
#include "dumpchars.cc"
    
#if debug > 0
    LOG2("sendAlspa: %d chars send.\n", res);
#endif				/*  */
}

bool AlspaDriver::AlspaOpen()
{
    bool res;
#if debug > 0
    LOG1("AlspaOpen enter.\n");
    FLUSH;
#endif				/*  */
    while (readChars(255) > 0) {
	LOG1("buffer not empty\n");
    }
    memset(&msg, 0, sizeof(msg));
#if debug > 0
    LOG1("AlspaOpen fillChar done.\n");
    FLUSH;
#endif				/*  */
    sendAlspa(23);
#if debug > 0
    LOG1("AlspaOpen sendAlspa done.\n");
    FLUSH;
#endif
    usleep(1000 * tmo_open);
#if debug > 0
    LOG1("AlspaOpen usleep(1000*tmo_open) done.\n");
    FLUSH;
#endif
    res = (readChars(64) == 24);
#if debug > 0
    LOG2("AlspaOpen read done. %d\n", res);
    FLUSH;
#endif				/*  */
    return (res);
}

#define read_from_PLC 0x01
#define write_to_PLC 0x02
#define typ_registers 0x08
#define typ_flags 0x16

void AlspaDriver::writeByte(word a, char &b, char t)
{
    memset(&msg, 0, sizeof(AlspaMessage));
    msg.read_write = write_to_PLC;
    msg.typ = t;
    msg.adr = a;
    msg.size = 1;
    msg.w1 = b;
    sendAlspa(23);
    readCharsLog(15, 5);
}

void AlspaDriver::readByte(word a, char *b, char t)
{
    memset(&msg, 0, sizeof(AlspaMessage));
    msg.read_write = read_from_PLC;
    msg.typ = t;
    msg.adr = a;
    msg.size = 1;
    sendAlspa(23);
    execError = readCharsLog(1 + 15, 2);
    *b = (char) fa.by[0];
#ifdef HAVEOUTWIN
    if Active
	(IoCtrlOut[ENr]) {
	cout << IoCtrlOut[ENr] << "MB: " << a[0] << "," << a[1] << " == "
	    << a[2] << "," << a[3] << endl;
//      fLOG2(ThisModule"cout << IoCtrlOut[ENr] << "MB: " << a[0] << "," << a[1] << " == "
//          << a[2] << "," << a[3] << endl;    
	}
#endif				/*  */
}

void AlspaDriver::execTI(transferInstruction * p)
{
    double rvalue;
    coord cell;
    int flags;
    int i;
    bool ok;
    word workAdr;
    int wordCount;
    word rs;
    char workCopy;		// The copy of a PLC byte, used to write single bits
    LOG1("execTI\n");
    FLUSH;
    execError = false;
    pa = (AlspaTI *) p;
    cell = p->startCell;
    i = p->startMW;
    if (p->direction == 0) {	// its a read instruction
	LOG1("execTI read\n");
	execError = false;
	memset(&msg, 0, sizeof(AlspaMessage));
	msg.read_write = read_from_PLC;
	msg.typ = typ_flags;
	msg.adr = p->startMW - 1;
	wordCount = p->endMW - p->startMW + 1;
	switch (p->DBNumber) {
	case 0:		// bits 
	    msg.adr = ((p->startMW - 1) / 8);
	    wordCount = ((p->endMW - p->startMW) / 8) + 1;
	    break;
/*		
    case 1 is all the standard values	    
		case 1: 		// bytes 
                typ=typ_flags;
                SAdr=StartMW-1;
                ws=EndMW-StartMW+1;
*/ 
	case 2:
	    msg.adr = 2 * msg.adr;
	    wordCount *= 2;
	    break;
	case 3:
	    msg.typ = typ_registers;
	    break;
	default:
	    LOG2("invalid Data Type: %d\n", p->DBNumber);
	}
	msg.size = wordCount;
	LOG2("wordCount: %d\n", wordCount);
	sendAlspa(23);
	if (p->DBNumber == 3) {
	    execError = readCharsLog(2 * wordCount + 15, 1);
	}
	if ((p->DBNumber >= 0) && (p->DBNumber <= 2)) {
	    execError = readCharsLog(wordCount + 15, 4);
	}
#ifdef HAVEOUTWIN
/*
	    if active
	      (IoCtrlOut[ENr])
	      {
		write (IoCtrlOut[ENr] ^, "Q: ", q, " a: " << endl;
			for (ii = 0; ii <= 5; ii++)
			{
			write (IoCtrlOut[ENr] ^, a[ii], "," << endl;
			       cout << IoCtrlOut[ENr] ^);}

*/
#endif

	LOG2("execError after read from PLC: %d\n",
	       execError);
	if (!execError) {
	    coreOfsetLoop(i, cell);
	} else {
#if debug > 0
	    LOG1("*** ExecError !\n");
	    FLUSH;
#endif
	}
    } else {			// read case ends here, write begins
	if (sheets[p->sheet] != NULL) {
	    workAdr = 0xffff;	// shouldn't be a valid address
	    while (i <= p->endMW) {
		flags = getFlags(p->sheet, cell, ok);
		if (ok)
		    if ((flags & changedMask) != 0) {
//			rvalue = sheets[p->sheet]->find(cell)->value;
			rvalue = numContents(p->sheet,cell,ok);
			if (p->DBNumber != 0) {	// Registerworte, Merkerworte, Merkerbytes }
			    memset(&msg, 0, sizeof(AlspaMessage));
			    msg.read_write = write_to_PLC;
			    msg.typ = typ_flags;
			    if (p->DBNumber == 3)
				msg.typ = typ_registers;
			    msg.adr = i - 1;
			    msg.size = 1;
			    if (p->DBNumber == 2) {
				msg.adr = 2 * (i - 1);
				msg.size = 2;
			    }
			    if (rvalue <= 0xffff) {
					// limit to word range
				msg.w1 = int (rvalue);	// use it
				clrPushButton(p->sheet, cell);
			    } else
				msg.w1 = 0xffff;
			    sendAlspa(23);
			    readCharsLog(15, 5);
			} else {	// Merker (single Bits) 
			    rs = i - 1;	// M%001 is at address 0
			    if (workAdr != (rs >> 3)) {	// calculate the byte address
				// this bit flag is in.
				// if it is in a new byte
				if (workAdr != 0xffff) {	// if its not our dummy
				    // address,we have to write
				    // our last workCopy back to
				    // the PLC:
				    writeByte(workAdr, workCopy,
					      typ_flags);
				}
				workAdr = rs >> 3;
					// our new byte address
				    readByte(workAdr, &workCopy,
					     typ_flags);
					// read contents
				    if (execError)
				    return;
					// if read went wrong we
				    // we cannot proceed
			    }
			    
				//
				// from here on, we have a valid workCopy. We copy
				// our value to bit (rs & 7):
				//
				if (rvalue == 0)
				workCopy = workCopy & ~(1 << (rs & 7));
			    
			    else
				workCopy = workCopy | (1 << (rs & 7));
			    clrPushButton(p->sheet, cell);
			    	// maybe, it was from
				// a pushbutton
			}	// bit stuff
		    }		// getFlags successfull
		cell.row++;
		i++;
	    }			// while not endMw
	    // after processing of the whole block, we may be left with
	    // a workCopy of the last byte, if we did write bits. We
	    // then still must write it back to PLC: 
	    if (workAdr != 0xffff) {	//
		writeByte(workAdr, workCopy, typ_flags);
	    }
	}			// sheet present ends
    }				// write ends
    usleep(1000 * 5);		// try to give time to other threads??
}

#include "thread_makro.cc"
#if debug>0
DTDebug(AlspaDriver)
#else
DT(AlspaDriver)
#endif
/*
    Changes:
    
    02/12/2002 	change output statements to printf, this gives 10% less code size.
    02/28/2002 	some changes and corrections to debug output
    05/13/2002 	Object Read & write functions return void now. cout<<... replaced
		by LOG2(..
    06/20/2002 	* got rid of some warnings
		* uses cellN makro in printf
    07/02/2002 	* thread_macro		
    02/23/2003  got rid of unused vars
*/
