/*
 Implementation of a PLC driver for the BECKHOFF BC9000/BK9000 controller. 

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include "lib0.h"
#include <stdio.h>
#include "plclog.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include "nethelp.h"

#include "byteswap.h"


#include "vi_string.h"
#include "plcdrivers/modbus.h"

#define debug 10
#define ThisModule "beckhoff2: "

#define uc unsigned char

class modbusTI:public transferInstruction {
  public:
    coord errCell;
    string format;
    DECLARE_PERSISTENCE(modbusTI)
    void modbusTI::Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(modbusTI, "BECKHOFF")

void modbusTI::Read(Engine & e)
{
    string stc, ctc, erc;
    e >> comment >> sheet >> startMW >> endMW >> DBNumber >> stc >> ctc >> erc >> format;
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
    errCell = getPos(erc);
#if debug >3
    LOG2("comment   :%s\n", comment.c_str());
    LOG2("sheet	:%d\n", sheet);
    LOG2("start Adr :%d\n", startMW);
    LOG2("end   Adr :%d\n", endMW);
    LOG2("DBNumber  :%d\n", DBNumber);
    LOG2("to Cell   :%s\n", cellN(startCell));
    LOG2("Condition :%s\n", cellN(ctrlCell));
    LOG2("Errors to :%s\n", cellN(errCell));
    LOG2("Format         :%s\n", format.c_str());
#endif
}

#pragma pack (1)

class modbusDriver:public PLCdrv {
  public:
    word DriverNr;
    modbusTI *pa;
    word DriverState;
    modbusMessage msg;
    int transaction_id;
    uc ibuf[1500];
    int tmo_read;
    string peer;
#ifdef strangeTests
    coord anaout;
    coord anahold;
    int lastAnaOut;
#endif    
    modbusDriver() {
#ifdef strangeTests    
	anaout=getPos("I1");
	anahold=getPos("M2");
#endif	
    };
    virtual void execTI(transferInstruction * p);
    int readChars(int size);
    void sendmodbus(int len);
    void watchdog_reset();
    int checkAnswer(int i, int count, int func);
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    if (rfd>0) {	// will never be 0
		execList();
		execCount++;
	    } else {
		usleep(500000);	// delay between connect trials
		openSocket();	// try to connect
	    }
	}
    };
    void openSocket();
};

int evalFormat(string f) {
    int j=0,bits=0;
    for (unsigned int i=0;i<f.length();i++) {
	char c=f[i]; //char c=f.at(i);
	switch(c) {
	    case 'X': bits+=1; break;
	    case 'B':
	    case 'C': j+=1; break;
	    case 'I': 
	    case 'W': j+=2; break;
	    case 'F':
	    case 'R': 
	    case 'D': 
	    case 'U': 
	    case 'L': j+=4; break;
	}	    
    }
    j=j+(bits +7)/8;
    return j;	
}



void modbusDriver::openSocket()
{
    sockaddr_in addr;
    socklen_t addrlen;
    peer = noSpace(peer);
    tmo_read = 16291;
#if debug>0
    LOG2("enter OpenSocket for peer (%s)\n", peer.c_str());
    FLUSH;
#endif    
    addr.sin_family = AF_INET;
    addr.sin_port = (((502) & 0xff) << 8) | (((502) & 0xff00) >> 8);
    inet_aton(peer.c_str(), &addr.sin_addr);
    rfd = wfd= socket(AF_INET, SOCK_STREAM, 0);
#if debug>0    
    LOG2("OpenSocket: socket is %d\n", wfd);
#endif
    addrlen = sizeof(addr);
    if (connect(rfd, (sockaddr *) & addr, addrlen)) {
#if debug>0
	LOG2(" Socket error: %s ", strerror(errno));
#endif    	
	close(rfd);
	rfd = wfd= 0;
    } else
#if debug>0    
	LOG2(" Connected to host: %s \n", peer.c_str());
#endif	
    FLUSH;
}

void modbusDriver::intSetup(DriverInfo * i)
{
    TypeManager::Add("BECKHOFF", CreateNewmodbusTI);
    peer=i->s;
    openSocket();
    PLCdrv::load(i);
    changedMask = registerDriver("BECKHOFF", this);
    errorMask = changedMask;
#ifdef strangeTests    
    anaout=getPos("I1");
    anahold=getPos("M2");
#endif    
};

int modbusDriver::readChars(int size)
{
    fd_set FDS;
    bool charsAvail;
    unsigned char *b;
#if debug>1    
    int i;
#endif    
    int res;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = tmo_read;
    usleep(tmo_read);
    FD_ZERO(&FDS);
    FD_SET(rfd, &FDS);
    charsAvail = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
#if debug>1
    LOG2("readChars %d\n", charsAvail);
#endif    
    if (!charsAvail) {
	LOG2(" timeout %d \n", charsAvail);
    }
    if (charsAvail) {
#if debug>1	    
	LOG2("readChars reading up to %d chars.\n", size);
#endif	
	res = read(rfd, &ibuf, size);
	b = (unsigned char *) &ibuf;
#if debug>1	
#include "dumpchars.cc"
	LOG2("readChars %d chars read.\n", res);
	FLUSH;
#endif	
	return (res);
    } else
	return (0);
}

void modbusDriver::sendmodbus(int len) {
    int res;
    res=modbusSend(len, &transaction_id, tmo_read, &msg, wfd);    
    if (res<0) {
	LOG1("sendmodbus: cannot write. Must reconnect\n");
	openSocket();
    }	
}

/*
    This is very special to BECKHOFF couplers:
*/

void modbusDriver::watchdog_reset()
{
    int i;
    LOG1("watchdogReset\n");
    msg.unit_id = 1;
    msg.function = 6;
    msg.ddd[0] = 0x11;
    msg.ddd[1] = 0x21;
    msg.ddd[2] = 0xbe;
    msg.ddd[3] = 0xcf;
    sendmodbus(12);
    i=readChars(261);
    checkAnswer(i, 0, 6);

    msg.unit_id = 1;
    msg.function = 6;
    msg.ddd[0] = 0x11;
    msg.ddd[1] = 0x21;
    msg.ddd[2] = 0xaf;
    msg.ddd[3] = 0xfe;
    sendmodbus(12);
    i=readChars(261);
    checkAnswer(i, 0, 6);
}

int modbusDriver::checkAnswer(int i, int num_regs, int func) {
    int errC=0;
    if (i < 9) {
	if (i == 0) {
	    LOG1("unexpected close of connection at remote end\n");
	    errC|=512;
	} else {
	    LOG2("response was too short - %d chars\n", i);
	    errC|=256;
	}
    } else if (0 != (ibuf[7] & 0x80)) {
	LOG2("MODBUS exception response - type %d\n", ibuf[8]);
	errC|=ibuf[8];
	if (ibuf[8] == 4) {
	    if (num_regs) watchdog_reset();
	}
    } else if (ibuf[7] != func) {
	LOG3("wrong response %d for function %d\n", func, ibuf[7]);
	tmo_read = 10 + (3 * tmo_read) / 2;
	errC|=1024;
	return errC;
    } else if (i != (9 + num_regs)) {
	LOG3("incorrect response size is %d expected%d\n",i, (9 + num_regs));
	errC|=2048;
    } else {
#if debug>1    
	for (i = 0; i < num_regs; i++) {
	    int w = (ibuf[9 + i]);	
	    // + (ibuf[10+i+i]& 0x00FF);
	    LOG3("byte %d = %x\n", i, w);
	} 
#endif	
	if (!tmo_read) tmo_read = 1;
	return errC;
    }
    return errC;
}
extern "C" {

static inline char daveGetByte(uc ** dc){
    char* c= * ((char**) dc);
    (*dc)++;
    return *c;
//    return (*  ((char*) dc)++);
}

static inline unsigned char daveGetChar(uc** dc){
    uc *c= *dc;
    (*dc)++;
    return *c;
//    return (*  ((unsigned char*) dc)++);
}

static inline float daveGetFloat(uc ** dc) {
    union {
	float a;
	int b;
    } f;
    int * i = (int*)(*dc);
    *dc+=sizeof(int);
    f.b=*i;
    f.b=bswap_32(f.b);
//    f.b=(bswap_32(* (   ((int*)( (int*)*dc)++) )));
    return (f.a);
}

static inline int daveGetInteger(uc ** dc){
    int * i = (int*)(*dc);
    *dc+=sizeof(int);
    int j=*i;
    j=bswap_32(j);
    return (j);
//    return (bswap_32(* (((int*)( (int*)dc)++) )));
}    


static inline ui daveGetDWORD(uc ** dc) {
    ui * i = (ui*)(*dc);
    *dc+=sizeof(ui);
    ui j=*i;
    j=bswap_32(j);
    return (j);
//   return (bswap_32(* (   ((int*)( (int*)*dc)++) )));
}

static inline unsigned int daveGetUnsignedInteger(uc ** dc) {
    ui * i = (ui*)(*dc);
    *dc+=sizeof(ui);
    ui j=*i;
    j=bswap_32(j);
    return (j);
//    return (bswap_32(* (   ((unsigned int*)( (unsigned int*)*dc)++) )));
}

static inline unsigned int daveGetWORD(uc ** dc) {
    us * i = (us*)(*dc);
    *dc+=sizeof(us);
    us j=*i;
    j=bswap_16(j);
    return (j);
//    return (bswap_16(* (   ((short*)( (short*)*dc)++) )));
}

static inline int daveGetINT(uc ** dc) {
    short * i = (short*)(*dc);
    *dc+=sizeof(short);
    short j=*i;
    j=bswap_16(j);
    return (j);
//    short s=(bswap_16(* (   ((short*)( (short*)*dc)++) )));
//    return s;
}

static inline float toPLCfloat(float ff) {
    union {
	float a;
	int b;
    } f;
    f.a=ff;
    f.b=(bswap_32(f.b));
    return (f.a);
}
}
void modbusDriver::execTI(transferInstruction * p) {
    double rvalue;
    coord cell;
    int i, ii, count, mask, byte, bc,errorCode;
    bool ok;
    char a[30];
    
    uc *dcc;
    uc**dc;
    int fp;
    int bits;
    double d;
    uc bitContainer;
    
    sprintf(a, "tmo: %d", tmo_read);
    params = a;
#if debug>1    
    LOG1("execTI\n");
    FLUSH;
#endif    
    if (rfd == 0) {
	usleep(20000);
	return;
    }
    execError = false;
    errorCode=0;
    pa = (modbusTI *) p;
    cell = p->startCell;
    ii = p->startMW;
    if (numContents(p->sheet, p->ctrlCell, ok) != 0) {
#if debug>1    
	LOG2("execTI %d\n", p->DBNumber);
#endif	
	execError = false;
	switch (p->DBNumber) {
	case 2:
#if debug>1	
	    LOG1("2: read input status\n");
#endif	    
	    count = p->endMW - p->startMW + 1;
	    memset(&msg, 0, sizeof(modbusMessage));
	    msg.function = p->DBNumber;
	    msg.ddd[0] = p->startMW / 256;
	    msg.ddd[1] = p->startMW % 256;
	    msg.ddd[2] = count / 256;
	    msg.ddd[3] = count % 256;
	    sendmodbus(12);
	    i=readChars(12);
	    errorCode=checkAnswer(i, (count + 7) / 8, p->DBNumber);
	    setCell(p->sheet, pa->errCell, errorCode, soIoState);
	    mask = 1;
	    byte = 0;
	    bc = 0;
	    for (i = 0; i < count; i++) {
		rvalue = (ibuf[9 + byte]) & mask;
		setCell(p->sheet, cell, rvalue, soExternal);
#if debug>0		
		LOG4("set cell %d:%s =%g \n",
		       p->sheet, cellName(cell).c_str(), rvalue
		);
//		LOG3("%x %x\n",ibuf[9 + byte], mask);       
#endif		
		mask <<= 1;
		cell.row++;
		bc++;
		if (bc == 8) {
		    byte++;
		    mask = 1;
		    bc = 0;
		}
	    }
#if debug>0	    
	    LOG2("execError after read from device: %d\n",
		   execError);
#endif		   
	    break;
	    
	case 3:
	case 4:
#if debug>1	
    	    if(p->DBNumber==3) LOG1("3: read holding registers\n");
	    if(p->DBNumber==4)LOG1("4: read input registers\n");
#endif	    
	    count = p->endMW - p->startMW + 1;
	    memset(&msg, 0, sizeof(modbusMessage));
	    msg.function = p->DBNumber;
	    msg.ddd[0] = p->startMW / 256;
	    msg.ddd[1] = p->startMW % 256;
	    msg.ddd[2] = count / 256;
	    msg.ddd[3] = count % 256;
	    sendmodbus(12);
	    i=readChars(1024);
	    if (i==1024)i=readChars(1024);
	    if (i==1024)i=readChars(1024);
	    if (i==1024)i=readChars(1024);
	    if (i==1024)i=readChars(1024);
	    if (i==1024)i=readChars(1024);
	    errorCode=checkAnswer(i, 2*count, p->DBNumber);
	    if(p->DBNumber==3) LOG1("3: read holding registers ");
	    if(p->DBNumber==3)LOG1("4: read input registers ");
	    LOG2(" %d\n",errorCode);
	    setCell(p->sheet, pa->errCell, errorCode, soIoState);
	    ii=9;
	    
	    dcc=ibuf+9;
	    dc=&dcc;
	    fp=0;
	    bits=0;
	    i=0;
	    count = p->endMW - p->startMW + 1;
	    while ((i<2*count)&&(fp<pa->format.length()) /*pa->bytes*/) {
		char fs=pa->format[fp]; //char fs=pa->format.at(fp);
#if debug >15
		LOG3("using format: %c for %d\n",fs,i);
#endif		
		char ic[10];
		sprintf(ic,"%d",i);
		string ss=ic;
		switch(fs) {
		    case 'X': if (bits==0) {
			bitContainer=daveGetByte(dc);
		    }	
		    if((bitContainer & (1 << bits))!=0) {
			d=1.0;
		    } else {
			d=0.0;
		    }
		    ss="DBX"+ss+".";
		    ss=ss+(char)('0'+bits); 
		    bits+=1;
		    if (bits==8) {
			i+=1;
			bits=0;
		    }	
		    break;
		    case 'B': d=daveGetByte(dc);  i+=1; ss="DBB"+ss;  break;
		    case 'C': d=daveGetChar(dc);  i+=1; ss="DBB"+ss;  break;
		    case 'I': d=daveGetINT(dc); i+=2; ss="DBW"+ss;break;
		    case 'W': d=daveGetWORD(dc); i+=2; ss="DBW"+ss;break;
		    
		    case 'F':
		    case 'R': d=daveGetFloat(dc); i+=4; ss="DBD"+ss;break;
		    case 'D': d=daveGetDWORD(dc); i+=4; ss="DBD"+ss;break;
		    case 'U': d=daveGetWORD(dc)+0x10000*daveGetWORD(dc); 
				i+=4; ss="DBD"+ss;break;
		    case 'L': d=daveGetInteger(dc); i+=4; ss="DBD"+ss;break;
		    default:i++;
		}
		if (errorCode!=0) {
		    setState(pa->sheet, cell, errorMask, true);
		} else {
		    setCell(pa->sheet,cell,d,soExternal);
#if debug >5				
		    LOG4("I set cell %s = %g from %s.\n",cellN(cell),d,ss.c_str());
#endif				    
		}    
		setSource(pa->sheet,cell,ss);

		cell.row++;
		fp++;
	    }
//	    break;
//	    }
	    setCell(pa->sheet,pa->errCell,0,soExternal);
//	}
//	 else { 
//	    protOK=false;   
//	    setCell(pa->sheet,pa->errCell,1,soExternal);
//	}    

/*
	    
	    for (i = 0; i < count; i++) {
		rvalue = 256*ibuf[ii]+ibuf[ii+1];
		setCell(p->sheet, cell, rvalue, soExternal);
#ifdef strangeTests		
		if (cell.l==anahold.l) {
		    int rw=(int) rvalue;	
		    if(rw==lastAnaOut) {
			LOG2 ("Output == Holding register %d\n",lastAnaOut);
		    } else {
			LOG3 ("Output %d <> Holding register %d !!!!!!\n",lastAnaOut, rw);
		    }
		}
#endif		
#if debug>0		
		LOG4("set cell %d:%s =%g\n", p->sheet, cellN(cell), rvalue);
#endif
		ii+=2;
		cell.row++;
	    }
*/	    
#if debug>1	    
	    LOG2("execError after read from device: %d\n", execError);
#endif	    
	    break;    
	case 5:		//force single coil
#if debug>1	
	    LOG1("5: force single coil\n");
#endif	    
	    while (ii <= p->endMW) {
		memset(&msg, 0, sizeof(modbusMessage));

		msg.function = p->DBNumber;
		msg.ddd[0] = ii / 256;
		msg.ddd[1] = ii % 256;
		rvalue=numContents(p->sheet, cell, ok);
#if debug>0		
		LOG4("cell %d:%s is %g\n", p->sheet, cellN(cell), rvalue);
#endif		
		if (rvalue != 0)
		    msg.ddd[2] = 255;
		else
		    msg.ddd[2] = 0;
		msg.ddd[3] = 0;
		sendmodbus(12);
		i=readChars(261);
		errorCode=checkAnswer(i, 3, p->DBNumber);
		setCell(p->sheet, pa->errCell, errorCode, soIoState);
		ii++;
		cell.row++;
	    }
	    break;
	case 6:		//preset single register
#if debug>1	
	    LOG1("6: preset single register\n");
#endif	    
	    while (ii <= p->endMW) {
		memset(&msg, 0, sizeof(modbusMessage));
		msg.function = p->DBNumber;
#if debug>1		
		LOG2("register address %d\n",ii);
#endif
		msg.ddd[0] = ii / 256;
		msg.ddd[1] = ii % 256;
		rvalue=numContents(p->sheet, cell, ok);
		int w=(int)rvalue;
#ifdef strangeTests		
		if (cell.l==anaout.l) lastAnaOut=w;
#endif		
#if debug>0		
	        LOG5("cell %d:%s is %g = %d\n", p->sheet, cellN(cell), rvalue, w);
#endif
	        msg.ddd[2] = w/256;
	        msg.ddd[3] = w%256;
		sendmodbus(12);
		i=readChars(261);
		errorCode=checkAnswer(i, 3, p->DBNumber);
		setCell(p->sheet, pa->errCell, errorCode, soIoState);
		ii++;
		cell.row++;
	    }    
	    break;
	case 15:		//write multiple coils
#if debug>1	
	    LOG1("15: write multiple coils\n");
#endif	    
	    while (ii <= p->endMW) {
		memset(&msg, 0, sizeof(modbusMessage));
		msg.function = p->DBNumber;
		msg.ddd[0] = ii / 256;
		msg.ddd[1] = ii % 256;
		rvalue=numContents(p->sheet, cell, ok);
#if debug>0		
		LOG4("cell %d:%s is %g\n", p->sheet, cellN(cell), rvalue);
#endif
		if (rvalue != 0)
		    msg.ddd[2] = 255;
		else
		    msg.ddd[2] = 0;
		msg.ddd[3] = 0;
		sendmodbus(12);
		i=readChars(261);
		errorCode=checkAnswer(i, 3, p->DBNumber);
		setCell(p->sheet, pa->errCell, errorCode, soIoState);
		ii++;
		cell.row++;
	    }
#if debug>1	    
	    LOG2("execError after read from device: %d\n",
		   execError);
#endif		   
	    break;
	}
    } else {			// the condition is not true

    }
//    usleep(1000 * 5);         // try to give time to other threads??
}

#include "thread_makro.cc"
#if debug>0
DTDebug(modbusDriver)
#else
DT(modbusDriver)
#endif
/*
    Changes:
    
    04/30/2002	written
    06/29/2002	got rid of some warnings
    01/17/2003 	implemented function 6, preset single register
    01/28/2003	reconnect automatically
		implemented function 3, read holding registers
		implemented function 4, read input registers
    01/29/2003	connect later if device wasn't available at startup
*/
