/*
 Implementation of a data logger. This one uses a newer file format I
 created for this project.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "lib0.h"
#include "cellList.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#define debug 10
#define ThisModule "log_bin  : "

#include "driverinfo.h"
#include "plclog.h"
#include <math.h>

#include "module.h"

extern "C" void setup(void *);
/*
    This may be redefined when there is a solution where to store old time/value
    for each coord from a range. Note that recordinstruction::read reads endCell
    and discards it then to keep file compatible.
*/


class recordInstruction: public moduleInstruction {
    DECLARE_PERSISTENCE(recordInstruction)
  public:
    double signiTime;		// time difference that will be considered significant
    double signiValue;		// value difference that will be considered significant
    double lastValue;		// last recorded value
    double lastTime;		// time when last value was recorded
    bool firstCall;		// true on start
    recordInstruction();
    virtual void Read(Engine & e);
    virtual bool subscribe(Module * mod);
    virtual void process(ClientInfo * ci, Module * mod, int flags);
};

class aModule:public Module {
  public:
    aModule (DriverInfo* di) {
	LOG1(ThisModule "aModule()\n");
	Module::init(di);
    }
    virtual void addInstructionType(){
	LOG1(ThisModule "add Instruction Type\n");
        TypeManager::Add("RECORD", CreateNewrecordInstruction);
    }	
};

#pragma pack(1)
struct datarecord {
    double time;		// this is seconds since Jan 1st, 1970 with fractional
    // part for microsecond resolution.
    coord c;			// Source coordinate
    int sheet;			// Source data sheet                    
    double v;			// The value
    int state;			// From status flags, marks data that may not be reliable
    int res;			// Reserved, fills it up to 32 byte
};

struct dataheader {
    char nl;			// Length of name (like PASCAL string)
    char Name[255];
    				// Name, taken from source coordinates Label
    double Signi_value;		// Amount of relative change in value, that is considered
    // interesting.
    double Signi_time;		// Amount of change in time, that is considered interesting.
    int reserved[12];		// Reserved, fills it up to 256+64 byte
};


bool recordInstruction::subscribe(Module * mod) {
    if (moduleInstruction::subscribe(mod)) {
	LOG4(ThisModule "I mark: %d:%s my bit mask: %x\n", sheet, cellN(Cell), mod->changedMask);
	string pp = ((aModule*)mod)->path + digits(sheet) + "/" + cellN(Cell) + ".bmd";
	LOG2(ThisModule "I try to open: %s\n", pp.c_str());
	int fd = open(pp.c_str(), O_RDWR | O_CREAT, 0644);
	LOG3(ThisModule "fd: %d error: %s\n", fd, strerror(errno));
	int pos = lseek(fd, 0, SEEK_END);
	LOG2(ThisModule "seek_end returned: %d\n", pos);
	if (pos == 0) {
	    cell *c = sheets[sheet]->find(Cell);
	    if (c != NULL) {
		dataheader h;
		memset(&h, 0, sizeof(h));
		h.nl = c->Label.length();
		strncpy(&h.Name[0], c->Label.c_str(), c->Label.length());
		h.Signi_time = signiTime;
		h.Signi_value = signiValue;
		write(fd, &h, sizeof(h));
	    }	
	}
	close(fd);
	return true;
    } else return false;
}

void recordInstruction::process(ClientInfo * ci, Module * mod, int flags)
{
    bool doRec = true;
	if (firstCall) {
	    firstCall = false;
	} else {
	    double tdiff =
		(time(NULL) - lastTime) / signiTime;
/*
 *	We want to know relative change of value. We refer to old value. If
 *	this is 0.0 we cannot divide and use new value instead. If this is 
 *	0.0 too, we cannot divide and use divisor 1.0.
 */
	    double ref = lastValue;
	    if (ref == 0.0)
		ref = (ci->c->getVal() != 0.0);
	    if (ref == 0.0)
		ref = 1.0;
	    double vdiff =
		(ci->c->getVal() -
		 lastValue) / ref / signiValue;
	    double distance = sqrt(tdiff * tdiff + vdiff * vdiff);
#if debug >4
	    LOG5(ThisModule "old %g %g , new %g %ld\n",
		   lastValue,
		   lastTime, ci->c->getVal(), time(NULL)
		);
	    LOG4(ThisModule
		   "norm'ed tdiff %g norm'ed vdiff %g dist. %g\n", tdiff,
		   vdiff, distance);
	    FLUSH;
#endif
	    doRec = (distance >= 1.0);	// significant change in data/time
	}
    if (doRec) {
	datarecord r;
	struct timeval tv;
	memset(&r, 0, sizeof(r));
	r.c = ci->c->k;
	r.v = ci->c->getVal();
	r.sheet = ci->sheet;
	r.state = ci->c->getState();

	gettimeofday(&tv, NULL);
	r.time = tv.tv_sec + 1e-6 * tv.tv_usec;
	string pp = mod->path + digits(ci->sheet) + "/" + cellName(ci->c->k) + ".bmd";
	LOG2(ThisModule "I try to open: %s\n", pp.c_str());
	int fd = open(pp.c_str(), O_WRONLY | O_APPEND);
	write(fd, &r, sizeof(r));
	lastTime = time(NULL);
	lastValue = ci->c->getVal();
#if debug >3
	LOG3(ThisModule "fd: %d %s\n", fd, strerror(errno));
#endif
	close(fd);
	FLUSH;
    }
}

void setup(void *arg)
{
    DriverInfo *di = (DriverInfo *) arg;
    LOG3(ThisModule "enter setup. list: %s path: %s\n", di->s1.c_str(), di->s.c_str());
    Module *drv = new aModule(di);
    if (!drv)
	LOG1(ThisModule "no instance generated!\n");
    FLUSH;
}

recordInstruction::recordInstruction()
{
};

IMPLEMENT_PERSISTENCE(recordInstruction, "RECORD")

void recordInstruction::Read(Engine & e)
{
    string stc, etc;
    e >> comment >> sheet >> stc >> etc >> signiTime >> signiValue;
    Cell = getPos(stc);
    lastValue = 0;
    lastTime = 0;
    firstCall = true;
}

/*
    Changes:
    
    06/21/2002  created this file.
    04/14/2003 	Moved common module code to module.cc
    05/20/2003 	Bug fixes, removed members variables which are already defined in module.h
*/
