/*
 Implementation of a data logger that stores data into a MySQL database.
 
 Host,user,password and database are common to all instructions and must be 
 specified in a comma separated list in visual.cfg.
 
 Every instruction has an SQL statement. (Usually INSERT)
 This version can also deal with select statements. If the trigger cell changes value,
 the SELECT is executed and the values from the database are stored into the corresponding
 cells.
 The table name is specified within the SQL statement.
 Variables are specified like with printf().  A query can have as many variables as you like,
 but the output from sprintf() must fit into queryStringBufferSize (2000).
 
 Example '"INSERT INTO oven values(0,NOW(),%0.3f,%0.3f)",T1,T2'
 The actual values from cells T1 and T2 are inserted, giving the actual query:
 "INSERT INTO oven values(0,NOW(),123.456,124.567)"
 The single quotes enclose the string recognized by the dri-list loader.
 The double quotes enclose the SQL part with format specifiers.
 Currently, all values from cells are retrieved as numbers of type double,
 so only format specifiers that can take a doublew are appropriate.

 This is from a real life application:
 <RECORD> 'save input data:' '"insert into machine4 values(0,0,now(),%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.1f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f,%0.8f)",E1,N10,N11,K12,N1,N2,N3,N4,K13,N26,N27,M22,M23,M24,M25,E7,N28,N29,N30,N31,N32,N33' 1 0 0 'E3' 'H2' 'H2' <RECORD>
 <RECORD> 'delete older data:' '"delete from machine4 where unix_timestamp(zeit) < (unix_timestamp(now())-600) and (id % 100)<>0 and keep=0 and comerrors=0",XA55' 1 1 0 'E4' 'H2' 'H2' <RECORD>
 <RECORD> 'save last data from delete:' '"update machine4 set keep=1",XA55' 1 1 0 'E6' 'H2' 'H2' <RECORD>
 The first line stores data every time E3 changes, which is whenever one of the other variables
 change value. It just stores all data relevant to know the state of that machine. This is done
 about 5 times per second. The first zero goes to an id field with auto_increment attribute.
 The second row is called every minute and deletes 99 out of 100 rows for data older than 10 
 Minutes. It thus leaves representative samples of the machine's state for about every 20
 seconds.
 The third row is invoked, when the machine stops. It marks all data present with the "keep"
 flag, preventing it from being deleted by the second statement. This results in leaving
 high resolution data for the last ten minutes before a breakdown.
 
 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"

#include "vi_string.h"

#include "driverinfo.h"
#include "config.h"
extern "C" {
#include <mysql/mysql.h>
}

#include "plclog.h"
#include "module.h"

#define debug 5
#define  ThisModule "logmysql3:"

extern "C" void setup(void *);

#define stINSERT 0
#define stSELECT 1

#define queryStringBufferSize 2000

class recordInstruction: public moduleInstruction {
    DECLARE_PERSISTENCE(recordInstruction)
  public:
    string SQL;			// an SQL statement to use
    int type;			// statement type, either write (INSERT, UPDATE) or read (SELECT)
    int maxRows;		// puts a limit on rows that will be copied from a SELECT result 
				// to s sheet
    int varCount;
    coord resultCell;		// this cell takes up the database operaton result
    coord validRowsCell;	// this cell gets the number of valid rows
    coord * Variables;		// memory space for coordinates		
    double * actValues;		// memory space for actual values
    recordInstruction(){};
    void prepareVars();		// counts variables, reserves memory and translates coordinates
				// to their binary representations
    string insertValues();	// gets actual values from the spreadsheet and fills them into the
				// SQL string
    virtual void Read(Engine & e);
    virtual void process(ClientInfo * ci, Module * mod, int flags);
};
    
class aModule:public Module {
  public:
    string host;
    string user;
    string passwd;
    string database;
    virtual void init(DriverInfo * di);
    virtual ~ aModule() {};				//pro forma destrutor
    MYSQL connection;
    MYSQL_RES *result;
    aModule (DriverInfo* di) {
	LOG1("aModule(DriverInfo)\n");
	init(di);
    }
};

void aModule::init(DriverInfo * di)
{
    string s = di->s;
    int p = s.Find(",");
    host = noSpace(s.substr(0,p));
    LOG2("host: %s\n", host.c_str());
    s = s.after(p);
    p = s.Find(",");
    user = noSpace(s.substr(0,p));
    LOG2("user: %s\n", user.c_str());
    s = s.after(p);
    p = s.Find(",");
    passwd = noSpace(s.substr(0,p));
    LOG2("password: %s\n", passwd.c_str());
    database = noSpace(s.after(p));
    LOG2("database: %s\n", database.c_str());
    Module::init(di);
    
    if (!mysql_init(&connection)) {
	LOG1("Error initializing client. mysql_init() failed.\n");
    } else {
	LOG1("Client initialized.\n");
    }
    FLUSH;

    if (!mysql_real_connect(&connection, host.c_str(),	// host
			    user.c_str(),	// user
			    passwd.c_str(),	// password
			    database.c_str(),	//"testdata", // database
			    0,	// port
			    NULL,	// unix socket
			    0	// client flags         
	)) {
	LOG2("Connection error.%s\n",
	       mysql_error(&connection));
    } else {
	LOG1("Connection established.\n");
    }
}

string recordInstruction::insertValues()
{
    bool ok;
    char buf[queryStringBufferSize];
//    LOG2("insertValues: %s\n",ds->SQLformat->c_str());
    for (int i = 0; i < varCount; i++) {
	actValues[i] = numContents(sheet, Variables[i], ok);
    }
//    LOG4("Variables: %12.2f %12.2f %12.2f \n",parms[0],parms[1],parms[2]);
    vsnprintf(buf, queryStringBufferSize, SQL.c_str(), (char*)actValues);
//    result=SQL;    
    return string(buf);
}

void recordInstruction::process(ClientInfo * ci, Module * mod, int flags)
{
    aModule * m=(aModule *)mod;
    string sx;
    sx = insertValues();
    if (type == 0) {	// INSERT type
#if debug >3
	LOG3("I will do: %s. qlen:%d\n", sx.c_str(), sx.length());
	LOG2("The connection address is:%p\n",&(m->connection));
#endif
	int qres = mysql_real_query(&(m->connection), sx.c_str(), sx.length());
//	if (qres)
	    LOG3("Query returned %d %s\n", qres,
		   mysql_error(&(m->connection)));
#if debug >3
	LOG2("Query result %d\n", qres);
#endif
	bool ok;
	LOG2("VRC %s\n", cellN(validRowsCell));
	if (numContents(sheet, validRowsCell, ok)>0) {
	    LOG2("VRC %s\n", cellN(validRowsCell));
	    setCell(sheet, Cell, 0, 0);
	}    
    } else if (type == 1) {	// SELECT one row type
	LOG3("I will do %s. qlen:%d\n", sx.c_str(), sx.length());
	FLUSH;
	int qres = mysql_real_query(&(m->connection), sx.c_str(), sx.length());
	if (qres) {
	    LOG3("Query returned %d %s\n", qres, mysql_error(&(m->connection)));
	    FLUSH;
	} else {
	    m->result = mysql_use_result(&(m->connection));
	    if (m->result) {
		int i, num_fields = mysql_num_fields(m->result), rowCount = 0;
		while (MYSQL_ROW resRow = mysql_fetch_row(m->result)) {
		    LOG2(
			   "fetchrow returned a row of %d fields.\n",
			   num_fields);
		    FLUSH;   
		    int useFields = num_fields;
//		    if (useFields > varCount) useFields = varCount;
		    LOG3("I use %d fields of %d\n", useFields, num_fields);
		    if (rowCount < maxRows) {
			for (i = 0; i < useFields; i++) {	// begin with 0
			    coord ci = resultCell;
			    ci.col += i;
			    ci.row += rowCount;
//			    coord ci = vars[i];
			    string vs = resRow[i];
			    setCell(sheet, ci, vs, soExternal);
			    LOG4("I set %d%s to %s\n",
				   sheet, cellName(ci).c_str(),
				   vs.c_str()
				);
			}
		    } else
			LOG3("wasting excess row: %d limit %d\n", rowCount, maxRows);
		    FLUSH;
		    rowCount++;
		}
		LOG2("row count: %d\n", rowCount);
		mysql_free_result(m->result);
		setCell(sheet, validRowsCell, rowCount, soExternal);
//!		setSource(sheet, validRowsCell, "valid rows in query");
		
	    } else
		LOG1("Query returned empty result.\n");
	}
    }
    FLUSH;
}

void setup(void *arg)
{
    DriverInfo *di = (DriverInfo *) arg;
    LOG3("enter setup. list: %s path: %s\n",
	   di->s1.c_str(), di->s.c_str());
    Module *drv = new aModule(di);
    if (!drv)
	LOG1("no instance generated!\n");
    FLUSH;
}

void recordInstruction::prepareVars()
{
    string varList,varList2;
    string var = "";
    const char * p1;
    int i;
    int fbeg = SQL.Find('"');
    SQL = SQL.after(fbeg);
    int fend = SQL.Find('"');
    varList = SQL.from(fend + 2);
    SQL = SQL.before(fend);
    LOG2("prepareVars format : %s\n", SQL.c_str());
    LOG2("prepareVars varList: %s\n", varList.c_str());
    p1=varList.c_str();
    varCount=1;
    while (*p1) {
	if (*p1==',') varCount++;
	p1++;
    }
    LOG2("prepareVars varCount: %d\n", varCount);
    Variables=(coord*)malloc(varCount*sizeof(coord));
    actValues=(double *)malloc(varCount*sizeof(double));
//    varList=varList2;
    int vc=0;
    i = varList.Find(',');
    while ((i > -1) && (!varList.empty())) {
	var = varList.substr(0,i);
	varList = varList.after(i);
	i = varList.Find(',');
#if debug > 5	
	LOG2("prepareVars Variable: %s\n", var.c_str());
	LOG2("prepareVars varList: %s\n", varList.c_str());
#endif
	Variables[vc] = getPos(var);
	vc++;
    }
    LOG2("prepareVars last Variable: %s\n", varList.c_str());
    Variables[vc] = getPos(varList);
}

IMPLEMENT_PERSISTENCE(recordInstruction, "RECORD")

void recordInstruction::Read(Engine & e)
{
    string stc, ctc, vtc;
    e >> comment >> SQL >> sheet >> type >> maxRows >> ctc >> stc >> vtc;
    Cell = getPos(ctc);			// execution trigger as in all modules
    resultCell = getPos(stc);		// stores query result code
    validRowsCell = getPos(vtc);	// stores query result code
    LOG3("type: %d SQL %s\n", type, SQL.c_str());
    prepareVars();
}
/*
    Changes:
    
    04/14/2003 	Moved common module code to module.cc
    05/20/2003 	Bug fixes, removed members variables which are already defined in module.h
    09/26/2003 	Overcome limitation to 20 variables per query. Next limit is size of query
		string buffer (2000);
    11/10/2003 	Newer gcc (3.?.?) wants a char * as last parameter in vsnprintf().
    02/28/2004 	In INSERT queries, set the cell that takes up result row count in other queries
		to 0.
    11/08/2004 	removed changedMask. This is already defined in class module.
*/
