/*
 Implementation of a data logger that stores data into a text file.
 This one is more flexible as log_text. You can specify a filename
 per instruction. Filenames can contain variables from the spreadsheet.
 This allows to log data to different files, e.g. dependent of a production
 lot number.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"

#include "vi_string.h"

#include "driverinfo.h"
#include "plclog.h"

#include "module.h"

#define queryStringBufferSize 2000

extern "C" void setup(void *);

#define debug 14
#define ThisModule "log_text2: "

#define stINSERT 0
#define stSELECT 1

class recordInstruction:public moduleInstruction {
    DECLARE_PERSISTENCE(recordInstruction)
  public:
    string formatString[2];	// a format string to use. index 0 is for the
				// text, index 1 for the filename.
    int varCount[2];
    coord * Variables[2];		// memory space for coordinates		
    double * actValues[2];		// memory space for actual values
    recordInstruction();
    void prepareVars(int n);
    string insertValues(int n);
    virtual void process(ClientInfo * ci, Module * mod, int flags);
    virtual void Read(Engine & e);
};

class aModule:public Module {
  public:
    aModule (DriverInfo* di) {
	LOG1(ThisModule "aModule()\n");
	Module::init(di);
        FLUSH;
    }
    
    virtual void addInstructionType(){
	LOG1(ThisModule "add Instruction Type\n");
        TypeManager::Add("RECORD", CreateNewrecordInstruction);
    }	
};

string recordInstruction::insertValues(int n)
{
    bool ok;
    char buf[queryStringBufferSize];
    for (int i = 0; i < varCount[n]; i++) {
	actValues[n][i] = numContents(sheet, Variables[n][i], ok);
    }
    vsnprintf(buf, queryStringBufferSize, formatString[n].c_str(), (char*)(actValues[n]));
    return string(buf);
}

void recordInstruction::process(ClientInfo * ci, Module * mod, int flags)
//void Module::callBack(ClientInfo * ci)
{
    int fd, res;
    string sx;
#if debug >3
    LOG4(ThisModule "%d:%s changed to: %g\n",
	   ci->sheet, cellName(ci->c->k).c_str(), ci->c->getVal());
    FLUSH;
#endif
    sx = insertValues(0);
    sx += "\n";
#if debug >3
    LOG2(ThisModule "I will do %s\n", sx.c_str());
#endif
    string fileName=mod->path+insertValues(1);
	
    fd = open(fileName.c_str(), O_APPEND | O_RDWR, 0644);
    if (fd<0)
	fd = open(fileName.c_str(), O_CREAT | O_RDWR,0644);
    LOG2(ThisModule "fd is: %d\n", fd);
    res = write(fd, sx.c_str(), sx.length());
    LOG3(ThisModule "write returned: %d %s\n", res, strerror(errno));
    close(fd);
    
}

void setup(void *arg)
{
    DriverInfo *di = (DriverInfo *) arg;
    LOG3(ThisModule "enter setup. list: %s path: %s\n",
	   di->s1.c_str(), di->s.c_str());
    Module *drv = new aModule(di);
    if (!drv)
	LOG1(ThisModule "no instance generated!\n");
    FLUSH;
}

void recordInstruction::prepareVars(int n)
{
    LOG3(ThisModule "prepareVars(%d) %s\n", n, formatString[n].c_str());
    string varList;
    string var = "";
    const char * p1;
    int i;
    int fbeg = formatString[n].Find('"');
    formatString[n] = formatString[n].after(fbeg);
    ui fend = formatString[n].Find('"');
    LOG2(ThisModule "prepareVars fend: %d\n", fend);
    FLUSH;    
    if(formatString[n].length() > (fend + 2)) {
	varList = formatString[n].from(fend + 2);
    } else {
	varCount[n]=0;
	Variables[n]=NULL;
	actValues[n]=NULL;
	LOG1(ThisModule "prepareVars: no values\n");
	FLUSH;    
	formatString[n] = formatString[n].before(fend);
	return;
    }	
    formatString[n] = formatString[n].before(fend);
    LOG2(ThisModule "prepareVars format : %s\n", formatString[n].c_str());
    LOG2(ThisModule "prepareVars varList: %s\n", varList.c_str());
    FLUSH;	
    p1=varList.c_str();
    varCount[n]=1;
    while (*p1) {
	if (*p1==',') varCount[n]++;
	p1++;
    }
    LOG2(ThisModule "prepareVars varCount: %d\n", varCount[n]);
    Variables[n]=(coord*)malloc(varCount[n]*sizeof(coord));
    actValues[n]=(double *)malloc(varCount[n]*sizeof(double));
//    varList=varList2;
    int vc=0;
    i = varList.Find(',');
    while ((i > -1) && (!varList.empty())) {
	var = varList.substr(0,i);
	varList = varList.after(i);
	i = varList.Find(',');
#if debug > 5	
	LOG2(ThisModule "prepareVars Variable: %s\n", var.c_str());
	LOG2(ThisModule "prepareVars varList: %s\n", varList.c_str());
#endif
	Variables[n][vc] = getPos(var);
	vc++;
    }
    LOG2(ThisModule "prepareVars last Variable: %s\n", varList.c_str());
    Variables[n][vc] = getPos(varList);
}


recordInstruction::recordInstruction()
{
};

IMPLEMENT_PERSISTENCE(recordInstruction, "RECORD")

void recordInstruction::Read(Engine & e)
{
    string stc,fnc;
    e >> comment >>  sheet >> stc >> formatString[1] >> formatString[0];
    Cell = getPos(stc);
    LOG3(ThisModule "file name: %s formatString %s\n", 
	formatString[1].c_str(), formatString[0].c_str());
    prepareVars(0);
    prepareVars(1);
}

/*
    Changes:
    02/13/2004	Adaptions to new current version: use vi_string.h, use .c_str() instead of 
		.chars(), remove modeRead, use .Find() instead of .index(), 
		use getFlags() and setFlags()
    02/14/2004	Filename can now include variable values the way the text can.
		Fixed bug in prepareVars when no variable fields given. 
		This module can now exactly bevave like log_text, only
		the filename is given in the instruction line and not in
		configuraton.
*/
