/*
 Implementation of a data logger that stores data into a text file.
 This one is more flexible as log_text. You can specify a filename
 per instruction. Filenames can contain variables from the spreadsheet.
 This allows to log data to different files, e.g. dependent of a production
 lot number.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"

#include "vi_string.h"

#include "driverinfo.h"
#include "plclog.h"

#include "module.h"

#define queryStringBufferSize 2000

extern "C" void setup(void *);

#define debug 14
#define ThisModule "log_text3: "

//#define stINSERT 0
//#define stSELECT 1

#define ctimeBufferLength 30

class aModule;

class recordInstruction:public moduleInstruction {
    DECLARE_PERSISTENCE(recordInstruction)
  public:
    string formatString[2];	// a format string to use. index 0 is for the
				// text, index 1 for the filename.
    int varCount[2];
    coord * Variables[2];	// memory space for coordinates		
    double * actValues[2];	// memory space for actual values
    char * modes[2];		// memory space for modes (numeric,text)
    int ctimeBuffersNeeded;	// ctime uses it's own static (global) buffer.
				// This gives unexpected results in a statement
				// like: 
				// printf("%s %s",ctime(&timea),ctime(&timeb));
				// ctime_r can use our buffer, but thus we
				// need an unlimited number of such buffers
				// to support an unlimited number of timestamps
				// in a one log string.
    recordInstruction() {
	ctimeBuffersNeeded=0;
    }	
    void prepareVars(int n);
    string insertValues(int n, aModule * mod);
    virtual void process(ClientInfo * ci, Module * mod, int flags);
    virtual void Read(Engine & e);
};

void setMaxCtimeBuffers(void *ri, void *mod);

class aModule:public Module {
  public:
    int maxCtimeBuffersNeeded;	// maximum from all instructions
    char * ctimeBuffers; 
    aModule (DriverInfo* di) {
	LOG1(ThisModule "aModule()\n");
	Module::init(di);
	maxCtimeBuffersNeeded=0;
        FLUSH;
    }
    
    virtual void addInstructionType(){
	LOG1(ThisModule "add Instruction Type\n");
        TypeManager::Add("RECORD", CreateNewrecordInstruction);
    }	
    
    virtual void setupAfterListLoad(){
	il->forEach(setMaxCtimeBuffers, this);
	if (maxCtimeBuffersNeeded>0)
	    ctimeBuffers=(char *)malloc(ctimeBufferLength * maxCtimeBuffersNeeded);
#if debug >3
	LOG2(ThisModule "Max buffers for ctime: %d\n", maxCtimeBuffersNeeded);
#endif
    };	
    
};

void setMaxCtimeBuffers(void *ri, void *mod)
{
    if (((recordInstruction *) ri)->ctimeBuffersNeeded >
	    ((aModule *) mod)->maxCtimeBuffersNeeded) {
	((aModule *) mod)->maxCtimeBuffersNeeded=
	    ((recordInstruction *) ri)->ctimeBuffersNeeded;
    }
}

string recordInstruction::insertValues(int n, aModule * mod) {
    bool ok;
    char buf[queryStringBufferSize];
    const char * buf2;
    time_t ct;
    char * ctimeChars=mod->ctimeBuffers;
    uc * values =(uc*) &(actValues[n][0]);
    for (int i = 0; i < varCount[n]; i++) {
	coord c=Variables[n][i];
	LOG3(ThisModule "trying to get value %c of %s\n", modes[n][i], cellN(c));
	switch(modes[n][i]) {
	    case 't':
		    *((double*) values) = getTimestamp(sheet, Variables[n][i]);
//		    ((double*) values)++;
		    values+=sizeof(double);
		    break;
	    case 'n':
		    *((double*) values) = numContents(sheet, Variables[n][i], ok);
//		    ((double*) values)++;
		    values+=sizeof(double);
		    break;
	    case 'c':
		    ct =(time_t)(getTimestamp(sheet, Variables[n][i]));
		    ctime_r(&ct, ctimeChars);
		    *((const char**) values) = ctimeChars;
		    ctimeChars[strlen(ctimeChars)-2]=0;	//remove trailing newline
		    ctimeChars+=ctimeBufferLength;
//		    ((char**) values)++;
		    values+=sizeof(char*);
		    break;	    
	    case 'l':
		    buf2 =symbolName(sheet, Variables[n][i], 0).c_str();
		    *((const char**) values) = buf2;
//		    ((char**) values)++;
		    values+=sizeof(char*);		    
		    break;
	    case 's':
		    buf2 =stringContents(sheet, Variables[n][i], 0,0).c_str();
		    *((const char**) values) = buf2;
//		    ((char**) values)++;
		    values+=sizeof(char*);
		    break;
	    case 'p':
		    *((double*) values) = getPreVal(sheet, Variables[n][i]);
//		    ((double*) values)++;
		    values+=sizeof(double);		    
		    break;
	    case 'u':
		    *((double*) values) = getState(sheet, Variables[n][i],ok);
//		    ((double*) values)++;
		    values+=sizeof(double);
		    break;
	    case 'v':
		    *((double*) values) = getFlags(sheet, Variables[n][i], ok);
//		    ((double*) values)++;
		    values+=sizeof(double);
		    break;	    
	    case 'w':
		    *((double*) values) = getError(sheet, Variables[n][i]);
//		    ((double*) values)++;
		    values+=sizeof(double);
		    break;	    
	    default:	    
		    actValues[n][i] = 0;
	}
	
    }
    vsnprintf(buf, queryStringBufferSize, formatString[n].c_str(), (char*)(actValues[n]));
    return string(buf);
}

void recordInstruction::process(ClientInfo * ci, Module * mod, int flags)
{
    int fd, res;
    string sx;
#if debug >3
    LOG4(ThisModule "%d:%s changed to: %g\n",
	   ci->sheet, cellName(ci->c->k).c_str(), ci->c->getVal());
    FLUSH;
#endif
    sx = insertValues(0, (aModule *) mod);
    sx += "\n";
#if debug >3
    LOG2(ThisModule "I will do %s\n", sx.c_str());
#endif
    string fileName=mod->path+insertValues(1, (aModule *) mod);
	
    fd = open(fileName.c_str(), O_APPEND | O_RDWR, 0644);
    if (fd<0)
	fd = open(fileName.c_str(), O_CREAT | O_RDWR,0644);
    LOG2(ThisModule "fd is: %d\n", fd);
    res = write(fd, sx.c_str(), sx.length());
    LOG3(ThisModule "write returned: %d %s\n", res, strerror(errno));
    close(fd);
    
}

void setup(void *arg)
{
    DriverInfo *di = (DriverInfo *) arg;
    LOG3(ThisModule "enter setup. list: %s path: %s\n",
	   di->s1.c_str(), di->s.c_str());
    Module *drv = new aModule(di);
    if (!drv)
	LOG1(ThisModule "no instance generated!\n");
    FLUSH;
}

void recordInstruction::prepareVars(int n)
{
    LOG3(ThisModule "prepareVars(%d) %s\n", n, formatString[n].c_str());
    string varList;
    string var = "";
    const char * p1;
    int i,i2;
    int fbeg = formatString[n].Find('"');
    formatString[n] = formatString[n].after(fbeg);
    ui fend = formatString[n].Find('"');
    LOG2(ThisModule "prepareVars fend: %d\n", fend);
    FLUSH;    
    if(formatString[n].length() > (fend + 2)) {
	varList = formatString[n].from(fend + 2);
    } else {
	varCount[n]=0;
	Variables[n]=NULL;
	actValues[n]=NULL;
	LOG1(ThisModule "prepareVars: no values\n");
	FLUSH;    
	formatString[n] = formatString[n].before(fend);
	return;
    }	
    formatString[n] = formatString[n].before(fend);
    LOG2(ThisModule "prepareVars format : %s\n", formatString[n].c_str());
    LOG2(ThisModule "prepareVars varList: %s\n", varList.c_str());
    FLUSH;	
    p1=varList.c_str();
    varCount[n]=1;
    while (*p1) {
	if (*p1==',') varCount[n]++;
	p1++;
    }
    LOG2(ThisModule "prepareVars varCount: %d\n", varCount[n]);
    Variables[n]=(coord*)malloc(varCount[n]*sizeof(coord));
    actValues[n]=(double *)malloc(varCount[n]*sizeof(double));
    modes[n]=(char *)malloc(varCount[n]*sizeof(char));
//    varList=varList2;
    int vc=0;
    i = varList.Find(',');
    while ((i > -1) && (!varList.empty())) {
	var = varList.substr(0,i);
	varList = varList.after(i);
	i = varList.Find(',');
#if debug > 5	
	LOG2(ThisModule "prepareVars Variable: %s\n", var.c_str());
	LOG2(ThisModule "prepareVars varList: %s\n", varList.c_str());
#endif
	modes[n][vc] = 'n';
	i2 = var.Find('.');
	if (i2>0) {
	    string suffix=var.after(i2);
	    var=var.before(i2);
	    LOG2(ThisModule "prepareVars got suffix: %s\n", suffix.c_str());
	    if (suffix=="time") modes[n][vc] = 't';
	    else if (suffix=="prev") modes[n][vc] = 'p';
	    else if (suffix=="ctime") {
		modes[n][vc] = 'c';
		ctimeBuffersNeeded++;
	    }	
	    else if (suffix=="label") modes[n][vc] = 'l';
	    else if (suffix=="string") modes[n][vc] = 's';
	    else if (suffix=="state") modes[n][vc] = 'u';
	    else if (suffix=="flags") modes[n][vc] = 'v';
	    else if (suffix=="error") modes[n][vc] = 'w';
	}
	Variables[n][vc] = getPos(var);
	vc++;
    }
    modes[n][vc] = 'n';
    var=varList;
    LOG2(ThisModule "prepareVars last Variable: %s\n", var.c_str());
    i2 = var.Find('.');
    if (i2>0) {
        string suffix=var.after(i2);
        var=var.before(i2);
        LOG2(ThisModule "prepareVars got suffix: %s\n", suffix.c_str());
	if (suffix=="time") modes[n][vc] = 't';
	else if (suffix=="prev") modes[n][vc] = 'p';
	else if (suffix=="ctime") {
		modes[n][vc] = 'c';
		ctimeBuffersNeeded++;
	    }
	else if (suffix=="label") modes[n][vc] = 'l';
	else if (suffix=="string") modes[n][vc] = 's';
	else if (suffix=="state") modes[n][vc] = 'u';
	else if (suffix=="flags") modes[n][vc] = 'v';
	else if (suffix=="error") modes[n][vc] = 'w';
    }
    Variables[n][vc] = getPos(var);
}

IMPLEMENT_PERSISTENCE(recordInstruction, "RECORD")

void recordInstruction::Read(Engine & e)
{
    string stc,fnc;
    int ctbn;
    e >> comment >>  sheet >> stc >> formatString[1] >> formatString[0];
    Cell = getPos(stc);
    LOG3(ThisModule "file name: %s formatString %s\n", 
	formatString[1].c_str(), formatString[0].c_str());
    prepareVars(0);
    ctbn=ctimeBuffersNeeded;
    prepareVars(1);
    if (ctbn>ctimeBuffersNeeded) {
	ctimeBuffersNeeded=ctbn;
    }	
}

/*
    Changes:
    02/13/2004	Adaptions to new current version: use vi_string.h, use .c_str() instead of 
		.chars(), remove modeRead, use .Find() instead of .index(), 
		use getFlags() and setFlags()
    02/14/2004	Filename can now include variable values the way the text can.
		Fixed bug in prepareVars when no variable fields given. 
		This module can now exactly bevave like log_text, only
		the filename is given in the instruction line and not in
		configuraton.
    02/14/2004# Can now retrieve other data than values from cells.
		A1.time will retrieve a timestamp from this cell as a number.
		(The cell must have one or result is 0.0). Use %f format.
		A1.ctime will retrieve a timestamp as formatted by ctime(3).
		Use %s format.
		A1.label will retrieve the label from a cell.
		Use %s format.
		A1.string will retrieve the string contents from a cell, e.g. 
		the formula. Use %s format.
		A1.prev will retrieve the previous value from this cell as 
		a number. Use %f format.
		A1.state will retrieve the state flags from this cell as 
		a number. Use %f format.
		A1.flags will retrieve the "changed" flags from this cell as 
		a number. Use %f format.
		A1.error will retrieve the error code from this cell as 
		a number. Use %f format.
*/
