/* Code by Paolo Baroni

 Implementation of a module to handle redundancy, interacting with another
 (and more in the future??) instance of VISUAL.

 Part of VISUAL, a human machine interface and data acquisition program

*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fstream.h>
#include "mytypes.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include "nethelp.h"
#include "driverinfo.h"
#ifndef DONT_USE_GETHOSTBYNAME
#include <netdb.h>		// for gethostbyname
#endif

#include "vi_string.h"
#include "plclog.h"
#include <math.h>
#include "mod_rdn.h"

#define debug 5
#define coordtime 400000        //first guess values :)
#define WaitRetry 500000        //coordtime is used as delay when switching status, to avoid possible conflicts.
#define Interval 100000         //Interval is the "intended" interval between Master's "still alive" signals (not sent
#define WaitMaster 1500000      //when values updates are frequent enough). WaitMaster is the time to declare dead the
                                //Master, when no signal is received from the spare. Right now is longer than WaitRetry,
				//so if the connection fails for some reason the Master has time for a few retries.

#define ThisModule "Redundancy  : "
#define soRDNmodule 12          // Used to exclude the CurrentMaster flagcell from callbacks. Another way is to simply
                                // check if a callback points to that cell, and in case stop the operation.
				// The new source type might anyway may be useful for future expansions or needs

extern "C" void * setup(void *);
// I need to declare mst here, since it cannot be passed to DriverThread (since that interprets its argument as
// DriverInfo), and so it cannot be passed to parseLine. Another option is to code here a new version of driverthread.
Master *mst;
void *MonitorMaster(void* arg);
pthread_mutex_t * stateMutex;   // Needed to avoid a possible problem when switching status, not 
				// related to the connection (that has its own mutex)
TransfList *TransfLists[maxSheet];  // Another option is to give signiValue via define, so storage
				    // could be allocated only when the value is >0. csh is used as 
				    // index somewhere below, unluckily even this one must be 
				    // global since using the existing structures I can't pass 
				    // the value..
int csh;                        

// Starting with functions to handle stored transfer informations
// They're very similar to the ones of cellList.cc, but different

TransfList::TransfList():recordList(NULL)
{
#if debug > 4
    LOG1(ThisModule "TransfList::TransfList()\n");
    FLUSH;
#endif
}

int TransfList::insert(base *item) {
    TransfRecord * it = (TransfRecord *) item;
    it->nextTR = recordList;
    recordList = it;
#if debug > 5
    LOG1(ThisModule "before search.\n");
    FLUSH;
#endif
    int i;
    if (search(item, i) == 0)
	atInsert(i, item);
#if debug > 5
    LOG1(ThisModule "insert done.\n");
    FLUSH;
#endif
    return i;
}

TransfRecord *TransfList::find(coord k)
{
#if debug >4
    LOG2(ThisModule "enter find: %s\n", cellN(k));
    FLUSH;
#endif
    int l = 0;
    int h = list::count - 1;
    while (l <= h) {
	int mid = (l + h) >> 1;
	int c = ((TransfRecord *) (items[mid]))->c->compare(k);
	if (c < 0)
	    l = mid + 1;
	else {
	    h = mid - 1;
	    if (c == 0) {
		return ((TransfRecord *) (items[mid]));
	    }
	}
    }
    return (NULL);
}

TransfRecord::TransfRecord():LastValSent(0), nextTR(NULL)
{
#if debug > 4
    LOG1(ThisModule "TransfRecord::TransfRecord()\n");
    FLUSH;
#endif
}

// Still functions to define...

void SetDirtyFlag(void *ce, void *)
{
    cell *c = (cell *) ce;
    if (c->getSourceType()==soExternal) {
       c->setState( c->getState() | stDirty);
#if debug >4
    LOG2(ThisModule "Set dirty flag for cell: %s\n",
	   cellName(c).c_str());
#endif
    }
}

void SetDirtyStates()    // It is used before assuming Master Status, to point that values might be not up to date
{   			// yet, before they're actually used by the new Master. 
    for (int sh = 1; sh <= maxSheet; sh++) {
       if (sheets[sh] != NULL) {
#if debug >3
	 LOG2(ThisModule "setting dirty flags for sheet no.: %d !\n", sh);
#endif
       sheets[sh]->forEach(SetDirtyFlag, NULL);
       }
    }
}

/* Function definitions for the Master's side */

void markSingleCell(void *ce, void *mst_p)
{   
    cell *c = (cell *) ce;
    if((c->isNumber()) && 	// suppose boolean and is meant. T.H.
    ((c->getSourceType())!=soRDNmodule)) { // Selecting stuff for Callbacks. Another option is
	int mask = ((Master *) mst_p)->Mask;                   // a check right in the Process function. Worse.
        c->setCflags( c->getCflags() | mask);
#if debug >4
        LOG3(ThisModule "Marked cell: %s my bit mask: %x\n", cellName(c).c_str(), mask);
#endif
        if ((mst->signiValue!=0) && // suppose boolean and is meant. T.H.
	(mst->TRexist==false)) { // Configuring it at 0 disables the mechanism at the root,
           TransfRecord *tr = new TransfRecord();          // so that useless computations are avoided.
           tr->c = c;
#if debug >4
           int pos=
#endif                                   // The current sheet (csh) should be passed by MarkAllCell() instead of being
           TransfLists[csh]->insert(tr); // a global. It could be known from the inside of the function forEach, but
#if debug >4                             // we're not modifying things when it's possible to do otherwise.
           LOG2(ThisModule "Insert done at position: %d\n", pos);
#endif
        }
     }
}


void Master::markAllCells()
{
    for (csh = 1; csh <= maxSheet; csh++) {
	if (sheets[csh] != NULL) {
#if debug >3
	    LOG2(ThisModule "marking cells of sheet no.: %d !\n", csh);
#endif
	    sheets[csh]->forEach(markSingleCell, (void *) this);
	}
    }
    TRexist=true; // signiValue's value doesn't matter, the "attempt" to create the lists must be done only once.
}

int Master::OpenSocket()      // Called by setStatus or by SignalToSpare
{   
    int fd;
    sockaddr_in addr;
    socklen_t addrlen;
    LOG1(ThisModule "enter OpenSocket\n");
    FLUSH;
    addr.sin_family = AF_INET;
    addr.sin_port =
	((port_trg & 0xff) << 8) | ((port_trg & 0xff00) >> 8);
#ifndef DONT_USE_GETHOSTBYNAME
    hostent *he = gethostbyname(peer.c_str());
    memcpy(&addr.sin_addr, he->h_addr_list[0], sizeof(addr.sin_addr));
#else
    inet_aton(peer.c_str(), &addr.sin_addr);
#endif
    fd = socket(AF_INET, SOCK_STREAM, 0);
    LOG2(ThisModule "OpenSocket: socket is %d\n", fd);
    addrlen = sizeof(addr);
    if (connect(fd, (sockaddr *) & addr, addrlen)) {
	LOG2(ThisModule "Socket error for Master: %s \n", strerror(errno));
	close(fd);
	fd = 0;
    } else
	LOG2(ThisModule "Connected to host: %s \n", peer.c_str());
    FLUSH;
    return(fd);
}

string formatItoa(int i) {      //It seems there is an "itoa" function, but it's not a standard.
   char st[4];
   sprintf(st, "%d", i);
   return (st);
}

string formatDb(float v) {        // In place of formatDouble
   char st[264];
   sprintf(st, "%g", v);
   return (st);
}

void sendSingleValue(void *p, void *fd)
{   
    string answ;
    char buffer[bSize];
    int wpos = 0, rpos = 0, ifd;

    cell *c = (cell *) p;
    if((c->isNumber())&&	// suppose boolean and is meant. T.H.
	((c->getSourceType())!=soRDNmodule)) {// Another option: check the mask, so only in MarkAllCells
       ifd = (int) fd;                                       // the actual values to send are chosen.. 
       float val = c->getVal(); // A new variable ensures the correct value is stored in the proper transfer record
       string s = "Master sending a value::" + formatItoa(csh) + ":" +cellName(c->k) + "=" +
           formatDb(val) + ":" + formatDb(c->getState()) + "\n";
       write(ifd, s.c_str(), s.length());
#if debug >4
       LOG2(ThisModule "Sent single cell information:: %s", s.c_str());
       FLUSH;
#endif
       usleep(10000 * 5);    // Original client module uses this setting as reasonable time for an answer
       readline(ifd, &buffer[0], answ, wpos, rpos);
       if (answ.empty()) {
	    LOG1(ThisModule "No answer from the spare instance!\n");
            FLUSH;
       }
       else if (answ == "Information received"){
#if debug >4
	  LOG1(ThisModule "Value has been sent successfully.\n");
#endif
          if(mst->signiValue!=0) {
              TransfRecord *tr = TransfLists[csh]->find(c->k);
#if debug >4
              LOG5(ThisModule "Value sent for cell %d:%s Found Transfer Record %d:%s\n",
                 csh, cellName(c->k).c_str(), csh, cellName(tr->c->k).c_str());
#endif
              tr->LastValSent = val;
          }
       }
    }
}

void Master::sendAllValues()
{   
    /*
	Why can't csh be local here ? T.H.
    */
    for (csh = 1; csh <= maxSheet; csh++) {      // Again the global csh is needed to pass current sheet number to
        if ((sheets[csh]!=NULL)) {                // the function above
#if debug >3
	    LOG2(ThisModule "Sending all values of sheet no.: %d !\n", csh);
#endif
	sheets[csh]->forEach(sendSingleValue, (void *) permfd);
        }
    }
}


// I needed to redefine these two functions
void Master::recalcCallBack(ClientInfo * ci)
{
#if debug >0
    LOG2(ThisModule "calling process(%p, 0)\n",this);
    FLUSH;
#endif
    process(ci);		// cyclic call from recalc
}

void Master::enterCallBack(ClientInfo * ci)
{
#if debug >0
    LOG2(ThisModule "calling process(%p, 0)\n",this);
    FLUSH;
#endif
    process(ci);		// out-of-cycle call from setCell
}


void Master::process(ClientInfo * ci) {
    string answ;
    char buffer[bSize];
    int wpos = 0, rpos = 0;        // Maybe I could just send 0 values...
    bool doSend = true;
    TransfRecord *tr;  // If defined later, space is saved when signiValue is 0, but defining it here avoids to
                      // repeat the find(k) operation when signiValue is not 0

    float val = ci->c->getVal(); // A new variable ensures the correct value is stored in the proper transfer record
    if(signiValue!=0) {      // Again this allows to disable the mechanism without useless computations
	tr = TransfLists[ci->sheet]->find(ci->c->k);
#if debug >4
	LOG5(ThisModule "Callback for cell %d:%s Found Transfer Record %d:%s\n",
    	    ci->sheet, cellName(ci->c->k).c_str(), ci->sheet, cellName(tr->c->k).c_str());
	LOG2(ThisModule "Last value sent: %g\n", tr->LastValSent);
#endif
	if (tr==NULL) LOG1(ThisModule "Can't happen!! Every Callback cell should have its transfer record!\n");
	FLUSH;
	double ref = tr->LastValSent;              // These operations follow the same logic used by loggers
	if (ref == 0.0)
	    ref = val;
	if (ref == 0.0) { //...then both are zero
	    ref = 1.0;     // So the division works...
	    doSend=false; // ... but send operation is blocked from here
        }
	double vdiff = (val - ref) / ref / signiValue;
	double distance = sqrt(vdiff * vdiff);             // In this case it simply makes it positive
	if (distance < 1.0) doSend = false;
    }
#if debug >2
    LOG4(ThisModule " %d:%s has now value: %g\n", ci->sheet, cellName(ci->c->k).c_str(), val);
#endif
    if (doSend) {// May need to list some more cases through AND conditions. Might move the check for sourcetype here.
#if debug >2
	LOG1(ThisModule "about to enter callBack and lock connectMutex.\n");
#endif
	pthread_mutex_lock (connectMutex);	// Must be placed here otherwise SignalToSpare might kill
						// permfd while from this function it could still 
						// appear >0. Probably not a great issue anyway.
	if (permfd <= 0) { 			// With the new settings this case should never 
    						// happen. But it's safer to leave the check.
    	    LOG1(ThisModule "callBack failed, connection to the spare unavailable.\n");
	    FLUSH;
	    if ((Mask)!=0) {
		if (unRegisterModule(Mask)==0) {
	    	    LOG1(ThisModule "Unregistered to stop callbacks.\n");
		    Mask=0;
	        }
	        else LOG1(ThisModule "I couldn't unregister!!!!!!\n");
	    }
	    pthread_mutex_unlock (connectMutex);
        }
        else {
    	    string sh = formatItoa(ci->sheet);
            string s = "Master sending a value::" + sh + ":" + cellName(ci->c->k) + "=" +
		     formatDb(val) + ":" + formatDb(ci->c->getState()) + "\n";
            write(permfd, s.c_str(), s.length());
#if debug >3
            LOG4(ThisModule "Callback's report to the spare. %d:%s changed to: %g\n",
		ci->sheet, cellName(ci->c->k).c_str(), val);
#endif
	    usleep(10000 * 5);       // Original client module uses this setting as reasonable time for an answer
            readline(permfd, &buffer[0], answ, wpos, rpos);
#if debug >3
	    LOG2(ThisModule "Answer received: %s\n", answ.c_str());   //check answ.lenght()?
	    FLUSH;
#endif
            if (answ.empty()) {
		LOG1(ThisModule "No answer from the spare instance!\n");
                FLUSH;
	    }
	    else if (answ == "Information received"){
		if (signiValue!=0)
	    	    tr->LastValSent = val;
	        gettimeofday(&LastTransf, NULL);	// Might reposition it. Not important. Anyway placed here is good,
						    // as it is set only if the update is answered.
#if debug >4                                    
	        LOG1(ThisModule "Value has been sent successfully.\n");
#endif
	    }
            else if (answ != "Information received"){
	        LOG1(ThisModule "Unknown format of the answer from the spare instance.\n");
	        FLUSH; // For now we are content with just this answer
	    }
	    pthread_mutex_unlock (connectMutex); //Might be placed earlier, but avoids possible confusion with debug info
       }
    }
#if debug >4
    else LOG1(ThisModule "callBack. No relevant change to send.\n");
#endif
  }



void Master::SignalToSpare()     // Might want to alter this function using somehow "break" instead of "goto",
{ 
    timeval current;             // seems fine anyway.
    double diff;
    char buffer[bSize];
    int wpos = 0, rpos = 0;        //maybe I could just send 0 values...
    string s, answ;
    bool sendAll = false;

    cell *c = ((sheets[priorSheet])->find(priorCell));
    while(IAmCurrentMaster) {
       connect_retry:
       while (permfd<=0) {
#if debug >2
	      LOG1(ThisModule "Waiting to (re)connect to the spare.\n");
#endif
	      usleep(WaitRetry);        //wait to retry, define value better
	      if (!IAmCurrentMaster) break;
	      permfd = OpenSocket();
	      priority = (int) c->getVal(); //While the spare is absent, priority can be changed with no re-negotiation,
	      setCell(priorSheet, priorCell, priority, soRDNmodule);//so we re-read it here and we reset the sourceType
	      sendAll=true;
       }
       if (!IAmCurrentMaster) break;
       if (sendAll) { // Complete values update at start and after a re-connection of the Spare:
	   pthread_mutex_lock (connectMutex); // In theory the mutex is not necessary, since Callbacks are stopped,
	   sendAllValues();                   // but it does no harm, and maybe some pathological cases are still
	   sendAll=false;                    // possible.
	   pthread_mutex_unlock (connectMutex);
        }
        // Program's flow can trap IAmCurrentMaster false here, in the LAST cycle, after portServer orders to leave
        // Master status. Can't lock a mutex in the starting while check, so it's repeated here. The mutex is always
        pthread_mutex_lock (stateMutex);                                 // found free, except possibly in that case.
	if (IAmCurrentMaster) {
// The following check ideally is triggered only if a variation is detected when the spare is already present,
// thanx to the above operations.
// This is only to avoid a doubling of the setStatus phase, not a great problem if it somehow happens anyway.
               if (priority != ((int) c->getVal())) {
                   LOG1(ThisModule "My priority has changed. Forcing a re-negotiation of the status.\n");
                   priority = (int) c->getVal();
		   setCell(priorSheet, priorCell, priority, soRDNmodule); //Confirms the value and resets the new
	           pthread_mutex_lock (connectMutex);                     //sourceType!
	           if ((Mask)!=0) {
	             if (unRegisterModule(Mask)==0) {   // Better to unregister before forced renegotiation, so we start
	               LOG1(ThisModule "Unregistered to stop callbacks.\n"); // again the normal sequence of operations.
	               Mask=0;
	             }
	             else LOG1(ThisModule "I couldn't unregister!!\n");
                   }
		   pthread_mutex_unlock (connectMutex);
		   system(fakestop.c_str());
                   setStatus();
		   pthread_mutex_unlock (stateMutex);
	           break;
               }
	      else {
	       priority = (int) c->getVal();
	       setCell(priorSheet, priorCell, priority, soRDNmodule); // To be ABSOLUTELY sure, we confirm it even here.
	      }
              if (Mask==0) {
                  Mask = registerModule("redundancy", this);   //other clients send the tfi or dri name...
                  LOG2(ThisModule "after registerModule with mask: %x\n", Mask);
                  markAllCells();
	       }
               gettimeofday(&current, NULL);
               diff = 1e6 * (current.tv_sec - LastTransf.tv_sec) + current.tv_usec - LastTransf.tv_usec;
               if (diff>Interval) {
                 pthread_mutex_lock (connectMutex);
	         s = "Master active\n";
#if debug >3
	         int res =
#endif
	             write(permfd, s.c_str(), s.length());
#if debug >3
	         LOG2(ThisModule "Signal sent: %d\n", res);
#endif
	         usleep(10000 * 5);       //original client module uses this setting as reasonable time for an answer
                 readline(permfd, &buffer[0], answ, wpos, rpos);
#if debug >3
	         LOG2(ThisModule "Answer received: %s\n", answ.c_str());
	         FLUSH;
#endif
	         gettimeofday(&LastTransf, NULL);
                 if (answ.empty()) {
		    LOG1(ThisModule "No answer from the spare instance!\n");
		    FLUSH;
		    close(permfd);
	            permfd=0;    // Since the connection is closed from here, we must handle the case that the spare is
		                 // running still but not answering for other reasons. The spare itself will wait a
		                 // while for other connections from the Master before taking over.
		    if ((Mask)!=0) {  // Now unregistering here is a must, otherwise, in case a new connection
	               if (unRegisterModule(Mask)==0) { // is created vary fast, some callbacks might create a queue
	                  LOG1(ThisModule "Unregistered to stop callbacks.\n"); // of slightly wrong values sent
	                  Mask=0;                                               // after the sendAllValues operation.
	               }
	               else LOG1(ThisModule "I couldn't unregister!!!!!!\n");
	            }   // The mutexes here are freed until an answer from the spare is received:
                    pthread_mutex_unlock (stateMutex);
	            pthread_mutex_unlock (connectMutex);
	            goto connect_retry;
	         }
                else if (answ != "Signal received") {
	            LOG1(ThisModule "Unknown format of the answer from the spare instance.\n");
	            FLUSH;
	        }
	        //for now we are content with just this answer
	        pthread_mutex_unlock (connectMutex);  //again could be placed earlier, but it's better for debug info
            }
            usleep(Interval); // The actual interval will be slightly longer than Interval :)
         }                 // particularly true if LastTransf has been very recently updated by Callback
                           // Checks can be made more often, but maximum delay is in the order of Interval*2 anyway.
	 pthread_mutex_unlock (stateMutex);
    }
   //returns to the end of SignalActivity when this instance loses the Master status
}


void *SignalActivity(void* arg) {
    ((Master *)arg)->SignalToSpare();
#if debug >3
    LOG1(ThisModule "SignalToSpare's cicle terminated.\n");
#endif
//  pthread_exit(NULL);
//  thread should terminate automatically when it reaches this point, if I'm not mistaken...
}


void *SpoofAddress(void* arg) {
    system(mst->fakestart.c_str()); //Starts to spoof the IP for the external HMIs and such
#if debug >3
   LOG1(ThisModule "SpoofAddress thread terminated.\n");
#endif
//  pthread_exit(NULL);
//  thread should terminate automatically when it reaches this point, if I'm not mistaken...
}


void Master::setStatus () {
   string s, answ;
   int wpos = 0, rpos = 0;    // Again wpos and rpos are not really needed so far
   char buffer[bSize];

   if (permfd <= 0) permfd = OpenSocket();   // If no connection problems arise, it's the first and last assignment to
   if (permfd <= 0) {                    // permfd, until the status remains master
       LOG1(ThisModule "Unable to connect to the other instance. Assuming Master status.\n");
       FLUSH;
       IAmCurrentMaster = true;
   }
   else {                  // No need for the mutex yet, the other activities have still to start
       string pr = formatItoa(priority);
       s = "Calling instance. My priority is:" + pr + "\n";
       write(permfd, s.c_str(), s.length());
#if debug >2
       LOG2(ThisModule "Negotiating priority. Sent message::%s\n",s.c_str());
#endif
       usleep(10000 * 5);       // Original client module uses this setting as reasonable time for an answer
       readline(permfd, &buffer[0], answ, wpos, rpos);
#if debug >3
       LOG2(ThisModule "Answer received: %s\n", answ.c_str());   //Check answ.lenght()?
       FLUSH;
#endif
       if (answ.empty()) {
          LOG1(ThisModule "No answer from remote instance! Assuming Master status.\n");
          FLUSH;
	  close(permfd);
	  permfd=0;                   //connection will be retried in SignalToSpare
	  IAmCurrentMaster = true;
       }
       else {
            int p = answ.Find("Replying instance. My priority is:");
            if (p >= 0) {
               answ = answ.from(p + 34);
               int remotePriority = atoi(answ.c_str());
	       if (priority > remotePriority)  { //if the values are equal the older instance "wins"
	          IAmCurrentMaster = true;
		  LOG1(ThisModule "My priority is higher. Assuming Master status.\n");
	       }
	       else {
	          LOG1(ThisModule "Continuing as spare instance.\n");
	          IAmCurrentMaster = false;
		  close(permfd);
	          permfd=0;         // *this* connection is not used when spare
	       }
	    }
	    else {
	       LOG1(ThisModule "Unknown format of the answer from remote instance!!!\n");
	       FLUSH;
	    }
       }
   }
   if (IAmCurrentMaster) { // a new Master starts with dirty flags, since it may come from Spare state and unknown
       SetDirtyStates();   // happenings...
       usleep(coordtime);//Gives more time to the new spare's drivers to stop, since they shouldn't ever run together.
       setCell(flagSheet, flagCell, 1, soRDNmodule);   //drivers and modules are made active
       usleep(coordtime);//Tries to give time to drivers, so sheets will load updated values before sending them.
       LOG1(ThisModule "Back from sleep. Starting Master activities\n");// To check drivers' status before and after this
       pthread_t spoof;                                                 // point in the log.
#if debug > 0
	int res =
#endif
          pthread_create(&spoof, NULL, SpoofAddress, NULL);
#if debug > 0
        LOG2(ThisModule "created IP spoofing thread: %d\n", res);
        if(res) {
	   LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	   FLUSH;
        }
#endif
       if (permfd > 0) { // If connection is not available, registration will be done later, by SignalToSpare
          sendAllValues();
          if (Mask == 0) {
              Mask = registerModule("redundancy", this);   // Other clients send the tfi or dri name...
              LOG2(ThisModule "after registerModule with mask: %x\n", Mask);
              markAllCells();  // Callbacks now will start on their own, below we start the additional signaling
	  }
	  else {
              LOG1(ThisModule "Should never happen!!! Mask was not 0 before registration!\n");
	      FLUSH;
          }
        }
        pthread_t sign;
#if debug > 0
	res =
#endif
          pthread_create(&sign, NULL, SignalActivity, this);
#if debug > 0
        LOG2(ThisModule "created Master's signal thread: %d\n", res);
        if(res) {
	   LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	   FLUSH;
        }
#endif
   }
   else {
      setCell(flagSheet, flagCell, 0, soRDNmodule);  // Stops drivers and modules (as long as they're configured
      usleep(2*coordtime);                           // for it, of course)
      LOG1(ThisModule "Back from sleep. Starting Spare activities\n");
      gettimeofday(&LastRec, NULL);//Waits the same time as the master, to avoid another take over for Master's absence!!
      pthread_t monitor;
#if debug > 0
	int res =
#endif
      pthread_create(&monitor, NULL, MonitorMaster, this);
#if debug > 0
      LOG2(ThisModule "created Spare's monitoring thread: %d\n", res);
      if(res) {
	  LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	  FLUSH;
      }
#endif
  }
}


/* Now the definitions for the Spare's side */


void Master::MonitorActivity() {
  timeval current;
  double diff;

  cell *c = ((sheets[priorSheet])->find(priorCell));
  gettimeofday(&LastRec, NULL);
  while (!IAmCurrentMaster) {
     pthread_mutex_lock (stateMutex); // Similarly to SignalToSpare, the mutex and the redundant check are necessary
     if(!IAmCurrentMaster) {          // to handle "pathological" cases
        gettimeofday(&current, NULL);
        diff = 1e6 * (current.tv_sec - LastRec.tv_sec) + current.tv_usec - LastRec.tv_usec;
        if (diff>=WaitMaster) {
           LOG1(ThisModule "Master not detected. Spare taking over.\n");
	   priority = (int) c->getVal();  // Better to load the up to date value even in this case.
	   setCell(priorSheet, priorCell, priority, soRDNmodule); //Confirms the value and resets the new sourceType
//         SetDirtyStates(); // Still to consider better. Here's used since we're not received updates for some time.
                             // Now commented since it is also called by SetStatus.
	   setStatus();   // Almost certainly takes over, unless the master returns to life very quickly, I figure :)
           pthread_mutex_unlock (stateMutex);
           break;        //... in that case another MonitorActivity thread will be started (this "break" seems
        }                 // mandatory, to end this one in any case)
        else if (priority != ((int) c->getVal())) {
           LOG1(ThisModule "My priority has changed. Forcing a re-negotiation of the status.\n");
           priority = (int) c->getVal();
	   setCell(priorSheet, priorCell, priority, soRDNmodule); //Confirms the value and resets the new sourceType
           setStatus();
	   pthread_mutex_unlock (stateMutex);
	   break;
        }
        else if (diff<WaitMaster){   // Might turn this in a simple "else" :)
#if debug >2
	    LOG1(ThisModule "Master's activity confirmed.\n");
#endif
	   priority = (int) c->getVal();
	   setCell(priorSheet, priorCell, priority, soRDNmodule); // To be ABSOLUTELY sure, we confirm it even here.
	   IAmCurrentMaster = false;  // It was necessary to confirm it here, at first... now removable?
           pthread_mutex_unlock (stateMutex);
        }
     }
     usleep(WaitMaster);    // Again in some cases the actual delay may be in the order of 2*WaitMaster
  }
  //returns to the end of MonitorMaster when this instance gains (even if temporarily) the Master status
}


void *MonitorMaster(void* arg) {
   ((Master *)arg)->MonitorActivity();
#if debug >3
   LOG1(ThisModule "MonitorActivity's cicle terminated.\n");
#endif
//  pthread_exit(NULL);
//  thread should terminate automatically when it reaches this point, if I'm not mistaken...
}



Master::Master() {
    permfd = 0;
    Mask = 0;
    IAmCurrentMaster = false;
    TRexist=false;
    connectMutex = new(pthread_mutex_t);    // Could be created elsewhere, but it's handier now
    pthread_mutex_init(connectMutex,NULL);
    gettimeofday(&LastTransf, NULL);
    //something missing?
}


#include "accepter.cc"


void parseLine(int fd, string & line) {
    int p;
    string s, readstring;

    p = line.Find("Master active");
    if (p >= 0) {
#if debug >3
	 LOG1(ThisModule "Signal received\n");
#endif
        s = "Signal received\n";
#if debug >3
	 int res =
#endif
	write(fd, s.c_str(), s.length());
#if debug >3
        LOG2(ThisModule "Answer sent: %d\n", res);
#endif
	gettimeofday(&(mst->LastRec), NULL);
	return;
    }
    p = line.Find("Master sending a value::");
    if (p >= 0) {
	pthread_mutex_lock (stateMutex);
        if (!(mst->IAmCurrentMaster)) {
           s = line.from(p + 24);
	   p = s.Find(":");
	   readstring = s.before(p);
	   s = s.after(p);
           int sheet = atoi(readstring.c_str());
	   p = s.Find("=");
	   readstring = s.before(p);
	   s = s.after(p);
	   coord cell = getPos(readstring);
	   p = s.Find(":");
	   readstring = s.before(p);
	   s = s.after(p);
	   double v = atof(readstring.c_str());
           int st = atol(s.c_str());

#if debug >2
	   LOG4(ThisModule "I will set Cell %d:%s to %0.7f\n", sheet, cellN(cell), v);
	      FLUSH;
#endif
	   setCell(sheet, cell, v, soExternal);     //ok, must check all of these details, source, state etc.
	   setState(sheet, cell, st, true);
	   gettimeofday(&(mst->LastRec), NULL);
	   s = "Information received\n";
#if debug >3
	   int res =
#endif
             write(fd, s.c_str(), s.length());
#if debug >3
           LOG2(ThisModule "Answer sent: %d\n", res);
#endif
        }
#if debug >2
	else LOG1(ThisModule "I am the Master now. Value from remote instance is not accepted.\n");
#endif
        pthread_mutex_unlock (stateMutex);
        return;
    }
    p = line.Find("Calling instance. My priority is:");
    if (p >= 0) {
        s = line.from(p + 33);
	int remotePriority = atoi(s.c_str());
#if debug >2
        LOG2(ThisModule "Remote instance's priority: %d\n", remotePriority);

#endif
        readstring = formatItoa(mst->priority);  //using readstring just to avoid creating a new variable
        s = "Replying instance. My priority is:" + readstring + "\n";
#if debug >3
	 int res =
#endif
	write(fd, s.c_str(), s.length());
#if debug >2
        LOG2(ThisModule "Negotiating priority. Sent message::%s\n",s.c_str());
#endif
        if (remotePriority > (mst->priority)) {    // Must BECOME or REMAIN Spare
           setCell(mst->flagSheet, mst->flagCell, 0, soRDNmodule);//Might be already 0, but it's better here for timings.
	   LOG1(ThisModule "Back from sleep. Starting Spare activities\n");
	   gettimeofday(&(mst->LastRec), NULL); // Again sleeps as much as the new Master, then marks time.
           pthread_mutex_lock (stateMutex); // Avoids problems with SignalToSpare
	   if (mst->IAmCurrentMaster) {
	       mst->IAmCurrentMaster = false;  // From here all the rest of the settings to leave Master status
	       system(mst->fakestop.c_str());
	       if ((mst->Mask)!=0) {
	          if (unRegisterModule(mst->Mask)==0) {
	             LOG1(ThisModule "Unregistered to stop callbacks.\n");
	             mst->Mask=0;
	          }
	          else LOG1(ThisModule "I couldn't unregister!!\n");
               }
	       pthread_mutex_unlock (stateMutex);
	       close(mst->permfd);
	       mst->permfd = 0;
	       usleep(2*coordtime);
	       pthread_t monitor;
#if debug > 0
	       res =
#endif
                  pthread_create(&monitor, NULL, MonitorMaster, mst);
#if debug > 0
               LOG2(ThisModule "created Spare's monitoring thread: %d\n", res);
               if(res) {
	           LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	           FLUSH;
               }
#endif
	    }
	    else {
	        pthread_mutex_unlock (stateMutex);
	    	usleep(2*coordtime);
		}
	    LOG1(ThisModule "Continuing as spare instance.\n"); // Common output for both cases.
        }
	else {              // Higher or equal priority. Must BECOME or REMAIN Master
	   pthread_mutex_lock (stateMutex);     // In this case is used in to coordinate with MonitorActivity
           if (!(mst->IAmCurrentMaster)) {
	       mst->IAmCurrentMaster = true;
               SetDirtyStates();
               usleep(coordtime);//Gives more time to the new spare's drivers to stop, since they shouldn't ever run
               setCell(mst->flagSheet, mst->flagCell, 1, soRDNmodule); // together. Drivers and modules are then made
	                                                               // active.
               usleep(coordtime);//Tries to give time to drivers, so sheets will load updated values before sending them.
	       LOG1(ThisModule "Back from sleep. Starting Master activities\n");
	       pthread_t spoof;
#if debug > 0
	       res =
#endif
                  pthread_create(&spoof, NULL, SpoofAddress, NULL);
#if debug > 0
               LOG2(ThisModule "created IP spoofing thread: %d\n", res);
               if(res) {
	            LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	            FLUSH;
               }
#endif
	       pthread_mutex_unlock (stateMutex);
	       pthread_t sign;
#if debug > 0
	       res =
#endif
                  pthread_create(&sign, NULL, SignalActivity, mst);
#if debug > 0
               LOG2(ThisModule "created Master's signaling thread: %d\n", res);
               if(res) {
	           LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	           FLUSH;
               }
#endif
	   }
	   else pthread_mutex_unlock (stateMutex);
	   LOG1(ThisModule "Continuing as Master instance.\n"); // Common output for both cases.
	 }
	return;
    }
    LOG1(ThisModule "Unknown or no signal received\n");
}


void *portServer(void *arg) {    // Altered from server.cc. Later might be adapted better, if needed.
    int fd;
    int rp = 0, wp = 0;		// buffer positions
    char buffer[bSize];
    string line;
    fd = *(int *) arg;
    LOG2(ThisModule "portServer: My fd is:%d\n", fd);
    FLUSH;
    int waitCount = 0;
    while (waitCount < 300) {
	while (readline(fd, &buffer[0], line, wp, rp) && (fd > 0)) {
	    LOG2(ThisModule "portServer (Spare instance): request:%s\n", line.c_str());
	    FLUSH;
	    parseLine(fd, line);
	    usleep(10000);
	}
	waitCount++;
    }
    LOG1(ThisModule "portServer: I closed my fd.\n");
    FLUSH;
    return NULL;
}


#include "driverThread.cc" //For now I'm using this one



void * setup(void * arg) {
    string params, readpar;

    DriverInfo *di = (DriverInfo *) arg;
    LOG2(ThisModule "enter setup. Parameters in: %s\n", di->s1.c_str());
    fstream f((di->s1).c_str(), ios::in);
    bool b = f.ios::good();
    if (b) {
      mst = new Master();
      if (!mst) {
	  LOG1(ThisModule "no Master instance generated!\n");
          FLUSH;
      }
      nextline:
        getline(f, params);
      if (params.Find("#") == 0) goto nextline;  //ok, ugly.. But hey, the file is simple and not that better :)
      int p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
#if debug >3
      LOG2(ThisModule "peer:%s\n", readpar.c_str());
#endif
      mst->peer = readpar;
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
#if debug >3
      LOG2(ThisModule "target port:%s\n", readpar.c_str());
#endif
      mst->port_trg = atoi(readpar.c_str());
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
#if debug >3
      LOG2(ThisModule "IP address to spoof:%s\n", readpar.c_str());
#endif
//      string command = "fake " + readpar + " >& /dev/null";	// Paolo's original
      string command = "ifconfig eth0:0 " + readpar + " up 1>&2";
      mst->fakestart = command;
//      command = "fake remove " + readpar + " >& /dev/null";	// Paolo's original
      command = "ifconfig eth0:0 down 1>&2";
      mst->fakestop = command;
      
      LOG2(ThisModule "Command is %s\n", mst->fakestart.c_str());
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
#if debug >3
      LOG2(ThisModule "local port:%s\n", readpar.c_str());
#endif
      di->s = "port=" + readpar;        //for compatibility with the existing driverthread
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find(":"));
      readpar = params.before(p1);
      params = params.after(p1);
      mst->flagSheet = atoi(readpar.c_str());
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
      mst->flagCell = getPos(readpar.c_str());
#if debug >3
      LOG3(ThisModule "CurrentMaster flag is in sheet:%d cell:%s\n", mst->flagSheet, readpar.c_str());
#endif
      cell *c = ((sheets[mst->flagSheet])->find(mst->flagCell));
      if (c==NULL) {
         LOG1(ThisModule "External CurrentMaster flag cell NOT FOUND!!!!\n");
         FLUSH;
         }
      else setCell(mst->flagSheet, mst->flagCell, 0, soRDNmodule);  //At start current master is false.
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find(":"));
      readpar = params.before(p1);
      params = params.after(p1);
      mst->priorSheet = atoi(readpar.c_str());
      p1 = (params.find("'"));
      readpar = params.before(p1);
      params = params.after(p1);
      mst->priorCell = getPos(readpar.c_str());
#if debug >3
      LOG3(ThisModule "Priority value is in sheet:%d cell:%s\n", mst->priorSheet, readpar.c_str());
#endif
      c = ((sheets[mst->priorSheet])->find(mst->priorCell));
      if (c==NULL) {
         LOG1(ThisModule "External priority cell NOT FOUND!!!!\n");
         FLUSH;
         }
      else {
         mst->priority = (int) c->getVal();
#if debug >3
      LOG2(ThisModule "My priority is:%d\n", mst->priority);
#endif
         setCell(mst->priorSheet, mst->priorCell, mst->priority, soRDNmodule); //Confirms the value and ensures the
      }                                                                        // new sourceType is set.
      p1 = (params.find("'"));
      params = params.after(p1);
      p1 = (params.find("'"));
      readpar = params.before(p1);
#if debug >3
      LOG2(ThisModule "Significant value:%s\n", readpar.c_str());
#endif
      mst->signiValue = atof(readpar.c_str());
      for (p1 = 1; p1 <= maxSheet; p1++) {
          if ((sheets[p1]!=NULL)&& 	// suppose boolean and is meant. T.H. 
	    ((mst->signiValue)!=0))  // signiValue == 0 stops the mechanism and prevents
	      TransfLists[p1] = new TransfList();         // useless operations as much as possible
          else  TransfLists[p1] = NULL;
      }
      if ((mst->signiValue)!=0)
         mst->markAllCells();
      // "fake" use of this function (Mask is still 0!). Serves to create the Transfer lists, since SendAllValues
      // must be launched before Callbacks are activated by the real registration.
      stateMutex = new(pthread_mutex_t);
      pthread_mutex_init(stateMutex,NULL);
      mst->setStatus();   // First and possibly even last call of this function.
      pthread_t serv;     // The server thread is started in any case!!! When in master status, it's used to receive
#if debug > 0             // communications from a new instance and negotiate the status. When in spare status, the
	int res =         // server is used to receive the "Still Alive" signal and the updates from Callbacks
#endif                    // mixed cases may happen if priority is changed via HMI
	     pthread_create(&serv, NULL, DriverThread, di);
#if debug > 0
	LOG2(ThisModule "created server thread: %d\n", res);
        if(res)
		LOG2(ThisModule "Thread creation error:%s\n", strerror(res));
	FLUSH;
#endif
  }
  else {
        LOG2(ThisModule "Can't open file: %s\n", di->s1.c_str());
	FLUSH;
  }
}



