/*
 This module stores short time trend data in memory and calculates 
 statistical values.
 
 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"
#include "globals.h"

#include "vi_string.h"

#include "driverinfo.h"

#define debug 0
#define ThisModule "mod_stat  : "

#include "module.h"
#include "plclog.h"
#include <math.h>


extern "C" void setup(void *);

struct trendData {
    int number;
    double time;	// this is seconds since Jan 1st, 1970 with fractional
    double v;		// The value
};

class trendRecordInst:public moduleInstruction {
    DECLARE_PERSISTENCE(trendRecordInst)
  public:
    int number;
    int numEntries;	// number of entries
    coord AvrCell;	// cell for Average
    coord RMSCell;	// cell for Average
    coord countCell;	// cell for N(umber of samples)
    coord LLimCell;	// cell lower limit
    coord ULimCell;	// cell upper limit    
    coord numClasses;	// number of histogram classes
    coord clearCond;	// clear if this cell is non zero
    coord classStart;	// first cell to store histogram class member counts
    int clearMode;	// different actions on clear
    trendData * trendMem;	// the Memory
    int nextEntry;	// entry to use, a ring buffer index
    int usedEntries;	// how many entries are yet used
    trendRecordInst(){};
    virtual bool subscribe(Module * mod);	// subscribe to a cell
    virtual void Read(Engine & e);
    void process(ClientInfo * ci, Module * mod, int flags);
};

class aModule:public Module {
  public:
    virtual ~aModule() {};	//pro forma destructor
    aModule (DriverInfo* di) {
	LOG1(ThisModule "aModule(DriverInfo)\n");
	init(di);
    }
    virtual void addInstructionType(){
	LOG1(ThisModule "add Instruction Type\n");
        TypeManager::Add("STAT", CreateNewtrendRecordInst);
    }
};

void trendRecordInst::process(ClientInfo * ci, Module * mod, int flags) {
/* 
    Test if we are called in valueChange. Note if so. Then we ignore calls from recalc from now.
*/					
    if (flags & 1) {
	calledInEnter=true; 
#if debug >5
	LOG1(ThisModule "called in setCell\n");
	FLUSH;
#endif	
    } else {
#if debug >5
	LOG1(ThisModule "called in recalc.\n");
	FLUSH;
#endif	    
/* 
    if not, test if we were ever called in enter (setCell()). If so, we ignore this call.
*/					
	if (calledInEnter) return;    
    }	

    int classCounts [1000];
    bool ok;
    int doClear=(int)numContents(sheet, clearCond, ok);
    if (doClear!=0) {
	usedEntries=0;
	nextEntry=0;
	setCell(sheet, clearCond, 0, soUDFmodule);
	return;
    }
    trendMem[nextEntry].v=ci->c->getVal();
    struct timeval tv;
//    gettimeofday(&tv, NULL);
//    trendMem[nextEntry].time = tv.tv_sec + 1e-6 * tv.tv_usec;
//    trendMem[nextEntry].number=number;	
    int ne=nextEntry+1;
    ne%=numEntries;
    nextEntry=ne;
    number++;
    if (usedEntries < numEntries) usedEntries++;
#if debug >5
    LOG2(ThisModule "%d entries\n", numEntries);
    FLUSH;
#endif	    
    int n=usedEntries,ii;
    double avr,sqsum,ugw,ogw,xw,delta;	
    int classes,inClass;
    classes=(int)numContents(sheet, numClasses, ok);
#if debug >5
    LOG3(ThisModule "%d classes %d used entries\n", classes,n);
    FLUSH;
#endif	        
    ugw=numContents(sheet, LLimCell, ok);
    ogw=numContents(sheet, ULimCell, ok);
    xw=(ogw+ugw)/2;
    avr=0;
    sqsum=0;	
    for (ii=0;ii<classes+2;ii++){
	classCounts[ii]=0;
    }
#if debug >5
	LOG1(ThisModule "class counts cleared.\n");
	FLUSH;
#endif	    
#if debug >5
    LOG3(ThisModule "low limit: %g high limit: %g.\n",ugw,ogw);
    FLUSH;
#endif    
    for (ii=0;ii<n;ii++){
	double x=trendMem[ii].v;
	avr+=x;
	sqsum+=x*x;
	if (x<ugw)classCounts[0]++; else 
	if (x>ogw) {
	    classCounts[classes+1]++;
	} else {
	    delta=x-ugw;
	    inClass=1+(int)(classes*delta/(ogw-ugw));
	    classCounts[inClass]++;
	}
    }
#if debug >5
    LOG1(ThisModule "calculation done.\n");
    FLUSH;
#endif
    setCell(sheet, countCell, n, soUDFmodule);
    setCell(sheet, AvrCell, avr/n, soUDFmodule);
    setCell(sheet, RMSCell, sqrt(sqsum/n), soUDFmodule);
    coord ccc=classStart;
    setCell(sheet, ccc, classes, soUDFmodule);
    for (ii=0;ii<classes+2;ii++){
	setCell(sheet, ccc, classCounts[ii], soUDFmodule);
	ccc.row++;
    }
#if debug >5
    LOG1(ThisModule "process done.\n");
    FLUSH;
#endif    
}

void setup(void *arg)
{
    DriverInfo *di = (DriverInfo *) arg;
    LOG3(ThisModule "enter setup. list: %s path: %s\n",
	   di->s1.c_str(), di->s.c_str());
    Module *drv = new aModule(di);
    if (!drv)
	LOG1(ThisModule "no instance generated!\n");
    FLUSH;
}


IMPLEMENT_PERSISTENCE(trendRecordInst, "TREND")

void trendRecordInst::Read(Engine & e)
{
    string stc,stAverage,stRMS,stCount,
    stUGW, stOGW, stNumClasses,stClearCond,stClassStart;
    e >> comment >> sheet >> stc>> numEntries >> stCount >> 
	stAverage>> stRMS >>stUGW >> stOGW >>stNumClasses >>stClearCond >> stClassStart >>clearMode;
    Cell = getPos(stc);
    countCell = getPos(stCount);
    AvrCell = getPos(stAverage);
    RMSCell = getPos(stRMS);
    LLimCell=getPos(stUGW);
    ULimCell=getPos(stOGW);
    numClasses=getPos(stNumClasses);
    clearCond=getPos(stClearCond);
    classStart=getPos(stClassStart);
    trendMem=NULL;
    nextEntry=0;
    usedEntries=0;
    number=0;
}

bool trendRecordInst::subscribe(Module * mod)	// subscribe to a cell
{
    if (moduleInstruction::subscribe(mod)) {
	trendMem=(trendData*)malloc(numEntries*sizeof(trendData));
#if debug >3
	LOG3(ThisModule "got %d bytes at %p.\n", numEntries*sizeof(trendData), trendMem);
	FLUSH;
#endif	
	return true;
    } else return false;
}

/*
    Changes:
    
    04/14/2003 	Moved common module code to module.cc
    05/20/2003 	Bug fixes, removed members variables which are already defined in module.h
    08/03/2003 	Bug fixes.
*/
