/*
 Implementation of a PLC driver for the BECKHOFF BC9000 controller. 

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include "lib0.h"
#include <stdio.h>
#include "plclog.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include "nethelp.h"

#include "vi_string.h"

#define debug 10

#include <byteswap.h>

#include "plcdrivers/modbus.h"

class modbusTI:public transferInstruction {
  public:
    DECLARE_PERSISTENCE(modbusTI)
    string slave;
    int unitID;
    int slfd;
    coord errCell;
    void modbusTI::Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(modbusTI, "MODBUS")

void modbusTI::Read(Engine & e)
{
    string stc, ctc, erc;
    e >> comment >> DBNumber >> slave >> unitID >> startMW >> endMW >> direction >>
	sheet >> stc >> ctc >> erc;
    slfd=0;	
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
    errCell = getPos(erc);
#if debug >3
    LOG2("Comment       :%s\n", comment.c_str());
    LOG2("Function Code :%d\n", DBNumber);
    LOG2("Slave         :%s\n", slave.c_str());
    LOG2("Unit          :%d\n", unitID);
    LOG2("Start Adr     :%d\n", startMW);
    LOG2("Item Count    :%d\n", endMW);
    LOG2("Type Code     :%d\n", direction);
    LOG2("Sheet         :%d\n", sheet);
    LOG2("To Cell       :%s\n", cellN(startCell));
    LOG2("Condition     :%s\n", cellN(ctrlCell));
    LOG2("Error Cell    :%s\n", cellN(errCell));
#endif
}

#pragma pack (1)

class modbusDriver:public PLCdrv {
  public:
    word DriverNr;
    modbusTI *pa;
    word DriverState;
    modbusMessage msg;
    int transaction_id;
    uc ibuf[300];
    int tmo_read  ;
    modbusDriver() {
    };
    virtual void execTI(transferInstruction * p);
    int readChars(int fd, int size);
    void sendmodbus(int fd, int len);
//    void watchdog_reset();
    bool checkAnswer(int i, int count, int func);
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    execList();
	    execCount++;
	}
    };
    int openSocket(string peer);
};

int modbusDriver::openSocket(string peer)
{
    sockaddr_in addr;
    socklen_t addrlen;
    peer = noSpace(peer);
    tmo_read = 416291  ;
    LOG2("enter OpenSocket for peer (%s)\n", peer.c_str());
    FLUSH;
    addr.sin_family = AF_INET;
    
    addr.sin_port =htons(502);        
    inet_aton(peer.c_str(), &addr.sin_addr);
    int slfd = socket(AF_INET, SOCK_STREAM, 0);
    LOG2("OpenSocket: socket is %d\n", slfd);
    addrlen = sizeof(addr);
    if (connect(slfd, (sockaddr *) & addr, addrlen)) {
	LOG2(" Socket error: %s ", strerror(errno));
	close(slfd);
	slfd = 0;
	usleep(100000);
    } else
	LOG2(" Connected to host: %s \n", peer.c_str());
    return slfd;	
    FLUSH;
}


void modbusDriver::intSetup(DriverInfo * i)
{
    TypeManager::Add("MODBUS", CreateNewmodbusTI);
//    openSocket(i->s);
    PLCdrv::load(i);
    changedMask = registerDriver("MODBUS_M", this);
    errorMask = changedMask;
};

int modbusDriver::readChars(int fd, int size)
{
    fd_set FDS;
    bool charsAvail;
    unsigned char *b;
    int i, res=0;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = tmo_read;
    usleep(tmo_read);
    FD_ZERO(&FDS);
    FD_SET(fd, &FDS);
    charsAvail = (select(fd + 1, &FDS, NULL, NULL, &t) > 0);
    t.tv_sec = 0;
    t.tv_usec = tmo_read;
    LOG3("readChars %d %d\n", fd, charsAvail);
    if (!charsAvail) {
	LOG3(" timeout %d %d\n", fd, charsAvail);
    }
    if (charsAvail) {
	LOG2("readChars reading up to %d chars.\n", size);
	res = read(fd, &ibuf, size);
	b = (unsigned char *) &ibuf;
#include "dumpchars.cc"
	LOG2("readChars %d chars read.\n", res);
	FLUSH;
	return (res);
    } else
	return (0);
 }

void modbusDriver::sendmodbus(int fd, int len) {
     int res;
     res=modbusSend(len, &transaction_id, tmo_read, &msg, wfd);    
}


bool modbusDriver::checkAnswer(int i, int num_regs, int func)
{
    if (i < 9) {
	 if (i == 0) {
	     LOG1("unexpected close of connection at remote end\n");
	 } else {
	     LOG2("response was too short - %d chars\n", i);
	 }
     } else if (0 != (ibuf[7] & 0x80)) {
	 LOG2("MODBUS exception response - type %d\n",
		ibuf[8]);
	 if (ibuf[8] == 4) {
//	    if (num_regs)
//		watchdog_reset();
	}
    } else if (ibuf[7] != func) {
	 LOG3("wrong response %d for function %d\n", func,
		ibuf[7]);
	 tmo_read = 10 + (3 * tmo_read) / 2;
	return false;
    } else if (i != (9 + num_regs)) {
	 LOG3("incorrect response size is %d expected%d\n",
		i, (9 + num_regs));
     } else {
	for (i = 0; i < num_regs; i++) {
	     int w = (ibuf[9 + i]);	// + (ibuf[10+i+i]& 0x00FF); 
	    LOG3("byte %d = %x\n", i, w);
	}
//	tmo_read=(500*tmo_read ) /501;
//	if(!tmo_read) tmo_read=1;
//	tmo_read-=1;
	    if (!tmo_read)
	    tmo_read = 1;
	return true;
    }
    return false;
}

void modbusDriver::execTI(transferInstruction * p)
{
    double rvalue;
    coord cell,firstCell;
    int i, ii, count, mask, byte, bc;
    bool ok, isMessageValid=false;
    short * nextValue;
    char a[30];
    sprintf(a, "tmo: %d", tmo_read);
    params = a;
    LOG1("execTI\n");
    FLUSH;
    if (numContents(p->sheet, p->ctrlCell, ok) == 0){
	return;
    }
//    if (rfd == 0) {
//	usleep(20000);
//	return;
//    }
    execError = false;
    pa = (modbusTI *) p;
    cell = p->startCell;
    ii = p->startMW;
    if (pa->slfd<=0)
        pa->slfd=openSocket(pa->slave);
    if(pa->slfd<=0) return;
    LOG2("execTI %s\n", p->comment.c_str());
    execError = false;
    memset(&msg, 0, sizeof(modbusMessage));
    msg.function = p->DBNumber;
    switch (p->DBNumber) {
	case 3: //Read Multiple Registers
	    
//	    break;
	case 2:
	    count = p->startMW;
	    msg.ddd[0] = p->startMW / 0x100;
	    msg.ddd[1] = p->startMW % 0x100;
	    msg.ddd[2] = count / 0x100;
	    msg.ddd[3] = count % 0x100;
	    sendmodbus(pa->slfd,12);
	    i=readChars(pa->slfd,12);
	    checkAnswer(i, (count + 7) / 8, p->DBNumber);
	    mask = 1;
	    byte = 0;
	    bc = 0;
	    for (i = 0; i < count; i++) {
		rvalue = (ibuf[9 + byte]) & mask;
		setCell(p->sheet, cell, rvalue, soExternal);
		LOG4("set cell %d:%s =%g ",
		       p->sheet, cellName(cell).c_str(), rvalue
		);
		LOG3("%x %x\n",ibuf[9 + byte], mask);       
		mask <<= 1;
		cell.row++;
		bc++;
		if (bc == 8) {
		    byte++;
		    mask = 1;
		    bc = 0;
		}
	    }
	    LOG2("execError after read from device: %d\n", execError);
	    break;
	case 5:		//write single coil
	    while (ii <= p->endMW) {
		memset(&msg, 0, sizeof(modbusMessage));

		msg.function = p->DBNumber;
		msg.ddd[0] = ii / 0x100;
		msg.ddd[1] = ii % 0x100;
		if (numContents(p->sheet, cell, ok) != 0)
		    msg.ddd[2] = 255;
		else
		    msg.ddd[2] = 0;
		msg.ddd[3] = 0;
		sendmodbus(pa->slfd, 12);
		i=readChars(pa->slfd, 261);
		checkAnswer(i, 3, p->DBNumber);
		ii++;
		cell.row++;
	    }
	case 15:		//write multiple coils
	    while (ii <= p->endMW) {
		memset(&msg, 0, sizeof(modbusMessage));
		msg.function = p->DBNumber;
		msg.ddd[0] = ii / 0x100;
		msg.ddd[1] = ii % 0x100;
		if (numContents(p->sheet, cell, ok) != 0)
		    msg.ddd[2] = 255;
		else
		    msg.ddd[2] = 0;
		msg.ddd[3] = 0;
		sendmodbus(pa->slfd, 12);
		i=readChars(pa->slfd, 261);
		checkAnswer(i, 3, p->DBNumber);
		ii++;
		cell.row++;
	    }
	    LOG2("execError after read from device: %d\n",
		   execError);
	    break;
	case 16:		//write multiple registers
/*
    Here we want to send only the values that have changed. So, when we encounter the first
    change, we set up a message packet and store the values into it. If the cell is also
    changed, we add it to the existing message. If there was no change, we send the message 
    and say it's invalid.
*/	
	    int MB_Adr=p->startMW; 	// the Modbus Address
	    int MB_Words;
	    for (ii = 0; ii<p->endMW;ii++) {
		int flags = getFlags(p->sheet, cell, ok);
#if debug >3		    
		LOG4("cell %s flags %x mask %x\n", cellN(cell),flags,changedMask);
		FLUSH;
#endif	
		if (flags & changedMask) {
		    if (!isMessageValid) {
			memset(&msg, 0, sizeof(modbusMessage));
			msg.function = p->DBNumber;
			msg.ddd[0] = MB_Adr / 0x100;
			msg.ddd[1] = MB_Adr % 0x100;
			nextValue=(short *) &msg.ddd[2];
			isMessageValid=true;
			firstCell=cell;
			MB_Words=0;		// Modbus word count	
		    }	
	    	    float d=numContents(pa->sheet,cell,ok);	    
		    short ss; 
		    us usv;
		    int si; 
		    ui uiv;
		    switch (pa->direction) {
			case 1: ss=(short) d; 
				*nextValue=ss;    
				nextValue++;
				break;
			case 2: ss=(short) d; 
				*nextValue=bswap_16(ss);    
				nextValue++;    
				break;	
			case 3: usv=(us) d;
				*nextValue=usv;    
				nextValue++;
				break;
			case 4: usv=(us) d; 
				*nextValue=bswap_16(usv);
				nextValue++;    
				break;		
			case 5: si=(int) d; 
				* ((int *) nextValue)=si;    
				nextValue+=2;    
				break;
			case 6: si=(int) d; 
				* ((int *) nextValue)=bswap_32(si);
				nextValue+=2;    
				break;	
			case 7: uiv=(ui) d; 
				* ((ui *) nextValue)=uiv;
				nextValue+=2;    
				break;
			case 8: uiv=(ui) d; 
				* ((ui *) nextValue)=bswap_32(uiv);
				nextValue+=2;    
				break;	
			case 9: * ((float *) nextValue)=d;
				nextValue+=2;    
				break;		
			case 10:si = * ((int *) &d); 
				si=bswap_32(si);
				* (int *) nextValue=si;
				nextValue+=2;    
				break;
		    }
//		} else {
		    sendmodbus(pa->slfd, 10+2*MB_Words);
		    i=readChars(pa->slfd, 261);
		    isMessageValid=false;
		}	
//		int itemCount=p->endMW;
		if (numContents(p->sheet, cell, ok) != 0)
		    msg.ddd[2] = 255;
		else
		    msg.ddd[2] = 0;
		msg.ddd[3] = 0;
		checkAnswer(i, 3, p->DBNumber);
		ii++;
		cell.row++;
		switch (pa->direction) {
		    case 1: 
		    case 2: 
		    case 3:
		    case 4: MB_Adr++; break;
		    case 5: 
		    case 6: 
		    case 7:
		    case 8: 
		    case 9: 
		    case 10:MB_Adr+=2; break;
		}
	    }
	    LOG2("execError after read from device: %d\n",
		   execError);
	    break;    
	}
}

#include "thread_makro.cc"
#if debug>0
DTDebug(modbusDriver)
#else
DT(modbusDriver)
#endif
/*
    Changes:
    
    04/30/2002	written
    06/29/2002	got rid of some warnings
    09/23/2003	changed variable name "us->usv,ui->uiv" to avoid conflict
		with type "us,ui"
*/
