/*
 A basic module. Does nothing.

 Module API:
 1. A real module defines a setup() function. This is called by the loader on program start.
 2. setup() calls the module specific constructor myModule(driverInfo di);
 2a. Constructor myModule(driverInfo di) leaves most of initialization to Module.init(di).
 2b. Module.init(di) registers Module with the cellList. 
     Pointers to methods recalcCallBack() and enterCallBack() are stored
     by cellList. The cellList returns a unique bitmask for the module.
 2c. Module.init(di) registers specific instruction type with the TypeManager via specific 
     addInstructionType();
 2d. Module.init(di) loads the instructions from path given in di.
 2e. If load succeeded, the trigger cells for each instruction are marked via subscribe().
 2f. Module.init(di) calls specific setupAfterListLoad();
 
 From now on, the method recalcCallBack() is called whenever a value change is detected after 
 recalculation. Call frequency is less or equal recalculation frequency.

 The method enterCallBack() is called whenever a numeric cell's value is changed
 via it's setCell() method. Call frequency may be above recalculation frequency, if a driver
 reads external values faster. This is used in trends and statistics to store all values
 even if they change to fast to be processed by recalculation.
 
 Both methods call process(instruction, module, flags).
 If called from enterCallBack(), process should set the calledInEnter flag of the
 respective instruction and ignore further calls from recalcCallBack().
 Thus it avoids to double process those value changes also detected in recalculation.
 
 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "plclog.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <fstream>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"
#include "persist.h"
#include "globals.h"

#include "vi_string.h"

#include "driverinfo.h"

#include "module.h"
#include "lilo.h"

#define debug 10

extern "C" void setup(void *);
using std::fstream;

void markCellsRI(void *ri, void *mod)
{
    ((Module *) mod)->markCell(((moduleInstruction *) ri));
}

Module::Module(DriverInfo * di) {
    DI = di;
    init (di);
}

void Module::init(DriverInfo * di)
{
//    _PID=getpid();
    LOG1( "enter Module::init()\n");
    FLUSH;
    changedMask = registerModule(di->s1, this);
    LOG2( "after registerModule with mask: %x\n", changedMask);
    string s = di->s;
    path = noSpace(s);
    LOG2( "extra information or filename: %s\n", path.c_str());
    LOG2( "load from %s\n", di->s1.c_str());
    FLUSH;
    addInstructionType();
    il=loadlist((di->s1).c_str());
/* * /    
    fstream f((di->s1).c_str(), ios::in);
    Engine *e = new Engine(f);
    il = new list();
    base *c;
    do {
	*e >> c;
	if (c != 0)
	    il->insert(c);
    } while (c != NULL);
    delete e;
    LOG1( "instruction list loaded.\n");
    FLUSH;
    f.close();
/ * */    
    if (il) {
	il->forEach(markCellsRI, this);
	setupAfterListLoad();
    } else {
	LOG1( "instruction list empty!\n");
    }
}

moduleInstruction * Module::findInstruction(ClientInfo * ci) {
    moduleInstruction *ent=NULL;
    string sx;
#if debug >0
    LOG1( "enter findInstruction.\n");
    FLUSH;
#endif
#if debug >3
    LOG4( "%d:%s changed to: %g\n",
	   ci->sheet, cellName(ci->c->k).c_str(), ci->c->getVal());
    FLUSH;
#endif
#if debug >4
    LOG1( "search corresponding tfi line.\n");
    LOG3( "List: %p %d entries\n", il, il->count);
    FLUSH;
#endif
    if (il) {
	for (int i = 0; i < il->count; i++) {
	    ent = (moduleInstruction *) (il->items[i]);
#if debug >4
	    LOG3( "entry: %d %s \n", ent->sheet, cellN(ent->Cell) );
#endif
	    if ((ent->sheet == ci->sheet)
		&& (ent->Cell.l == ci->c->k.l)) {
#if debug >0		
		LOG2( "entry: %p is the right statement\n",
		       ent);
		FLUSH;
#endif		
		break;
	    }
	}
	if (ent==NULL) {
	    LOG1( "cant happen: no instruction for client info. Who subscribed to this?\n");
    	    FLUSH;
	}
    }
    return (ent);
}

void Module::recalcCallBack(ClientInfo * ci)
{
    moduleInstruction *ent= findInstruction(ci);
#if debug >0    
    LOG3( "recalcCallBack calling process(%p, %p, 0)\n",ent,this);
    FLUSH;
#endif    
    if (ent) ent->process(ci, this, 0);		// this is the cyclic call from recalc
}

void Module::enterCallBack(ClientInfo * ci)
{
    moduleInstruction *ent= findInstruction(ci);
#if debug >0        
    LOG3( "enterCallBack calling process(%p, %p, 1)\n",ent,this);
    FLUSH;
#endif        
    if (ent) ent->process(ci, this, 1);		// this is the out-of-cycle call from setCell
}

bool moduleInstruction::subscribe(Module * mod)
{
    calledInEnter=false;
    if (sheets[sheet] == NULL) {
	LOG2( "there is no sheet no.: %d !\n", sheet);
	return false;
    }
#if debug >3    
    LOG3("I mark: %s my bit mask: %x\n",  cellN(Cell), mod->changedMask);
#endif		   

    cell *c = sheets[sheet]->find(Cell);
    if (c != NULL) {
	c->setCflags( c->getCflags() | mod->changedMask);
    } else {
	LOG2( "no cell %s!\n", cellN(Cell));
    }

    return true;
}

bool Module::markCell(moduleInstruction * ri)
{
    return ri->subscribe(this);
}

void moduleInstruction::Read(Engine & e)
{
    string cellPos;
    e >> comment >> sheet >> cellPos;
    Cell = getPos(cellPos);
}

void moduleInstruction::process(ClientInfo * ci, Module * mod, int flags) {
    LOG1("process() is abstract !\n");
}
/*
    Changes:
    04/14/2003 	Moved common module code to this file.
    02/18/2004 	Log which callback has been invoked (setCell or recalc).
    11/04/2004 	Modules keep their configuration information
*/
