/*
 Definition of a basic module. Does nothing.
 
 Modules are used to do something on data change.
 1. data logging
 2. calculations which are to complex for the spreadsheet.

 Module API:
 1. A real module defines a setup() function. This is called by the loader on program start.
 2. setup() calls the module specific constructor myModule(driverInfo di);
 2a. Constructor myModule(driverInfo di) leaves most of initialization to Module.init(di).
 2b. Module.init(di) registers Module with the cellList. 
     Pointers to methods recalcCallBack() and enterCallBack() are stored
     by cellList. The cellList returns a unique bitmask for the module.
 2c. Module.init(di) registers specific instruction type with the TypeManager via specific 
     addInstructionType();
 2d. Module.init(di) loads the instructions from path given in di.
 2e. If load succeeded, the trigger cells for each instruction are marked via subscribe().
 2f. Module.init(di) calls specific setupAfterListLoad();
 
 From now on, the method recalcCallBack() is called whenever a value change is detected after 
 recalculation. Call frequency is less or equal recalculation frequency.

 The method enterCallBack() is called whenever a numeric cell's value is changed
 via it's setCell() method. Call frequency may be above recalculation frequency, if a driver
 reads external values faster. This is used in trends and statistics to store all values
 even if they change to fast to be processed by recalculation.
 
 Both methods call process(instruction, module, flags).
 If called from anterCallBack(), process should set the calledInEnter flag of the
 respective instruction and ignore further calls from recalcCallBack().
 Thus it avoids to double process those value changes also detected in recalculation.
 
 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#ifndef module__
#define module__

#include <sys/time.h>
#include "persist.h"
#include "plclog.h"
#include "lib0.h"
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <fstream>
#include "cellList.h"
#include "mytypes.h"
#include "list.h"
//#include "persist.h"
#include "globals.h"

#include "vi_string.h"

#include "driverinfo.h"

using std::fstream;

class Module;

class moduleInstruction: public base {
  public:
    string comment;		// Every instruction can have a comment
    coord Cell;			// every instruction is triggered when a cell's value changes.
    int sheet;			// the sheet where the is in

    bool calledInEnter;		// a module may mark an instruction with this flag when processed
				// first time in enterCallBack. If this flag is set,
				// it will not be processed in further calls to recalcCallBack
				
    moduleInstruction() {} 	// make the compiler happy
    virtual ~moduleInstruction() {}	// make the compiler happy
    virtual void Read(Engine & e);	// read an Instruction from file
					// Most modules will overload this

    virtual bool subscribe(Module * mod);	// subscribe to a cell
						// overload this if you need a per-instruction setup
						// Most modules don't overload this

    virtual void process(ClientInfo * ci, Module * mod, int flags);
					// Modules MUST overload this
					// pointer to the executing module is the most 
					// universal information here
};

class Module:public client {
  public:
//    int _PID;
    DriverInfo * DI;		// The information passed from configuration file
    int changedMask;		// bit mask to extract drivers "new value flag" 
				// set by calculation thread
    string path;		// some path
    Module(){};
    Module(DriverInfo * di);	// Modules MUST define their own constructor.
				// The constructor MUST call init().
    
    virtual void init(DriverInfo * di);	// Modules may overload this. Overloaded version
					// should call original module::init()
    
    virtual ~ Module() {};	// pro forma destructor
    list *il;			// list of instructions
    
    virtual void addInstructionType(){};	// Modules should overload this	to register a
						// their specific instruction type
						
    virtual void setupAfterListLoad(){};	// Modules can overload this to perform some.
						// some special task after instruction list
						// is loaded and before first call to process()
    
    virtual void recalcCallBack(ClientInfo * ci);	// Modules may overload this. 
    virtual void enterCallBack(ClientInfo * ci);	// Modules may overload this. 
	
    moduleInstruction * findInstruction(ClientInfo * ci);  	// cannot be overridden
    bool markCell(moduleInstruction *);			
    /*
	cannot be overridden. Overload Instruction::subscribe() instead. There you can cast mod
	to a pointer to your module type.
    */	
};

#endif

