/*
 Implementation of a serial line master driver for OMRON PLCs using
 hostlink protocol.
 This should make visual behave more or less like an OMRON PLC. The first
 gola of this is to communicate with SOMETHING because i've no OMRON
 PLC to test. As with the Allen Bradley slave, one can also use it
 as a kind of bridge:
 Disconnect an existing SCAD/HMI or other communication device from
 the PLC, connect VISUAL to the PLC, let VISUAL get all the data from 
 the PLC, including what the other device needs and connect the other 
 device to a serial interface handled by this slave driver.


 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include "plclog.h"

#include "vi_string.h"

#define debug 4

#include <stdio.h>

#define tmo_normal 999999
#define ThisModule "OMRON_MAS: "

#include "omron_lib.h"

//#define sim_broken_transport

class Omron_Master_TI:public transferInstruction {
  public:
    int unitID, itemCount;		//      Omron's numerical unit ID
    string headerCode;
    coord errCell;
    DECLARE_PERSISTENCE(Omron_Master_TI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(Omron_Master_TI, "OMRON")

void Omron_Master_TI::Read(Engine & e)
{
    string ctc, stc, erc;
    e >> comment >> unitID >> headerCode >> startMW >> itemCount >> sheet >> stc >> erc >> ctc;
    startCell = getPos(stc);
    errCell = getPos(erc);
    ctrlCell = getPos(ctc);
#if debug >3
    LOG2("Comment        :%s\n", comment.c_str());
    LOG2("Unit ID        :%d\n", unitID);
    LOG2("Header         :%s\n", headerCode.c_str());
    LOG2("first word     :%d\n", startMW);
    LOG2("number of words:%d\n", itemCount);
    LOG2("sheet          :%d\n", sheet);
    LOG2("to Cell        :%s\n", cellN(startCell));
    LOG2("Errors to Cell :%s\n", cellN(errCell));
    LOG2("Condition      :%s\n", cellN(ctrlCell));
#endif
}

class OmronMaster:public OmronDriver {
  public:
    OmronMaster() { };
    virtual void intSetup(DriverInfo * i);
    virtual void execTI(transferInstruction * p);
    void execute() {
	while (1) {
	    execList();
	    execCount++;
	}
    }
};

void OmronMaster::intSetup(DriverInfo * i)
{
    rfd= setPort(i->s, &wfd);
//    fd= setPort(i->s);    
    LOG2("intSetup write fd: %d\n",wfd);
    TypeManager::Add("OMRON", CreateNewOmron_Master_TI);
    PLCdrv::load(i);
    changedMask = registerDriver("OmronMaster", this);
    errorMask = changedMask;
};

/*
void ascii_encode(uc * b,int x, int digits) {
    if (digits>3) {
	*b=0x30 + (x/1000); b++; x=x%1000;
    }
    if (digits>2) {
	*b=0x30 + (x/100); b++; x=x%100;
    }
    if (digits>1) {
	*b=0x30 + (x/10); b++; x=x%10;
    }
    *b=0x30 + x;
}
*/
void OmronMaster::execTI(transferInstruction * p)
{
    LOG1("execTI() starts.\n");
    FLUSH;
    bool ok;
    if (numContents(p->sheet, p->ctrlCell, ok) == 0.0) {
#if debug >2
	LOG3("execTI condition %d:%s not true, nothing to do.\n", p->sheet, cellN(p->ctrlCell));
	FLUSH;
#endif	
	return;
    }
/*
    Ok to execute.
*/	    
    Omron_Master_TI * pa =(Omron_Master_TI*) p;
    double rvalue;
    coord cell=pa->startCell;
    int itemLen=4, i, errorState=0, res, endCode, pos;
    uc b[256];
    if (pa->headerCode[0]=='W') goto write;
    if (b[4]=='G') {
	itemLen=1;
    }
    b[1]=0x30 + (pa->unitID/10);
    b[2]=0x30 + (pa->unitID%10);
    i=pa->headerCode.length();
    memcpy(b+3,pa->headerCode.c_str(),i);
    ascii_encode(b+i+3,pa->startMW,4);
    ascii_encode(b+i+3+4,pa->itemCount,4);
    omron_send(b,i+3+4+4);
    res=readAnswer(b);
    if (res<0) {
	errorState=-res;
    }	
    if (memcmp(b+3,pa->headerCode.c_str(),i)) {
	errorState=256;
    }
    endCode=ascii_decode(b+6,2);
    if (endCode!=0) {
	errorState=endCode;
    }
    if (errorState!=0) {
	setCell(pa->sheet, pa->errCell, errorState, soIoState);
	LOG3("set error cell %s to %d\n",cellN(pa->errCell),errorState);
	return;
    }
    pos=6;
    
    for (i=0;i<pa->itemCount;i++) {
	rvalue=ascii_decode(b+pos,itemLen);
	pos+=itemLen;
	setCell(pa->sheet, cell, rvalue, soExternal);
#if debug >2	
	LOG3("set cell %s to %g\n",cellN(cell), rvalue);
#endif	
	cell.row++;
    }
    return;	// from read commands
    
write:  
    int PLCadr=pa->startMW;
    for (i = 0; i < pa->itemCount; i++) {
	int flags = getFlags(p->sheet, cell, ok);
#if debug >3		    
	LOG5("write: testing cell %s flags %x mask %x ok:%d\n",cellN(cell),flags,changedMask,ok);
	FLUSH;
#endif		    
	if (!ok) goto next;
	if (flags & changedMask) {
	    float d=numContents(pa->sheet,cell,ok);
#if debug >3		    
            LOG3("cell %s changed to %g \n",cellN(cell),d);
            FLUSH;
#endif		    	    
	    b[1]=0x30 + (pa->unitID/10);
	    b[2]=0x30 + (pa->unitID%10);
	    i=pa->headerCode.length();
	    memcpy(b+3,pa->headerCode.c_str(),i);
	    ascii_encode(b+i+3, PLCadr, 4);
	    ascii_encode(b+i+3+4,(int)d,4);
	    omron_send(b,i+3+4+4);
#if debug >3		    			    
#endif			    
	    res = readAnswer(b);
	    if (res>0) {
#if debug >3		    
		LOG3("clear flags of cell %s mask %x\n",cellN(cell),changedMask);
		FLUSH;
#endif		    		    		    
		clrPushButton(p->sheet, cell); // change transmitted, now clear the flag
	    } else {
		if (res<0) {
		    setCell(pa->sheet, pa->errCell, errorState, soIoState);
		    LOG3("set error cell %s to %d\n",cellN(pa->errCell),-res);
		}
	    }	
	}
next:	
	cell.row++;	// Next cell
	PLCadr++;	// Next PLC variable
    }
}

#include "thread_makro.cc"
#if debug>0
int fd;
DTDebug(OmronMaster)
#else
DT(OmronMaster)
#endif

