/*
 Common definitions for a serial line driver for OMRON PLCs

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#ifndef OMRON_lib__
#define OMRON_lib__

#define STX 0x02
#define ETX 0x03
#define ENQ 0x05
#define ACK 0x06
#define CR  0x0D
#define DLE 0x10
#define NAK 0x15

#define uc unsigned char

static const uc hexChars[16]={'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};

class OmronDriver: public PLCdrv {
  public:
    OmronDriver() {}
    
    void ascii_encode(uc * b,int x, int digits) {
	if (digits>3) {
	    *b=0x30 + (x/1000); b++; x=x%1000;
	}
	if (digits>2) {
	    *b=0x30 + (x/100); b++; x=x%100;
	}
	if (digits>1) {
	    *b=0x30 + (x/10); b++; x=x%10;
	}
	*b=0x30 + x;
    };
    
    int ascii_decode(uc * b, int digits) {
	int x=0;
	if (digits>3) {
	    x=x+1000* ((*b) & 0x0F); b++;
	}
	if (digits>2) {
	    x=x+100* ((*b) & 0x0F); b++;
	}
	if (digits>1) {
	    x=x+10* ((*b) & 0x0F); b++;
	}
        x=x+ ((*b) & 0x0F); b++;
	return x;
    };
/*
    Checksum calculation copied from Tomeu Capu's omrom library
*/
/* 
    Calcula el FCS d'una cadena de carcters 
    per a fer el checksum.
*/
    unsigned short  calc_CRC(unsigned char * b, int size) {
	int i,r=0;
	for(i=0;i<size;i++) {
	    r=(r^b[i]);
#if debug >10
	    LOG3("CRC %c %d\n",b[i],r);
#endif	    
	}
	return r;
    }
    
    int readChars(uc *b, int tmo) {
	fd_set FDS;
	bool charsAvail;
	int res;
	timeval t;
	t.tv_sec = 0;
	t.tv_usec = tmo;
	FD_ZERO(&FDS);
	FD_SET(rfd, &FDS);
	charsAvail = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
	t.tv_sec = 0;
	t.tv_usec = tmo;
#if debug >2
	LOG1("readChars (here we are) \n");
#endif
	if (!charsAvail) {
	    LOG1("timeout !\n");
	    return (0);
	} else {
#if debug >2
	    LOG1("readChars reading up to 1024 chars.\n");
#endif
	    res = read(rfd, b, 1024);
#if debug >2
	    int i;
#include "dumpchars6.cc"
	    LOG2("readChars %d chars read.\n", res);
	    FLUSH;
#endif
	    return (res);
	} 
    }    
    
    int readAnswer(uc * b1) {
	int res = 0, r;
	do {    
	    r=readChars(b1+res, tmo_normal);
	    res += r;
	    if (r == 0) {
    		return -1;
	    } else {
//do we have the final * CR ? 
		if (res<8) return -2; // answer too short
		if ((b1[res - 2] == '*') && (b1[res - 1] == CR)) {
		    int FCS=calc_CRC(b1,res-4);
#if debug >2
    		    LOG4("readAnswer() found *CR, check FCS %X %c %c\n",FCS,b1[res-4],b1[res-3]);
#endif
		    if ( (b1[res-4]!=hexChars[FCS/16]) || (b1[res-3]!=hexChars[FCS%16]) ) return -4;
		    return res;
		} 
	    }
	} while (1);	
    }
    

    void omron_send(uc * b, int res){
	b[0]='@';
	unsigned short FCS=calc_CRC(b,res);
	b[res]=hexChars[FCS/16];
	b[res+1]=hexChars[FCS%16];
	b[res+2]='*';
	b[res+3]=13;
	res+=4;
#if debug >3
	int i;
#include "dumpchars5.cc"
#endif
	write(wfd, b, res);
    };

};

#endif //OMRON_lib__
