/*
 Implementation of a PLC driver for the SIEMENS S5 series via its programmer
 port. See www.runmode.com for information about the underlying AS511
 protocol and a proposal of circuit to connect the current loop programmer
 port to a common V24.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "plcdrv.h"
#include "s5.h"
#include "s5l.h"
#include "lib0.h"
#include "setport.h"

#include "vi_string.h"

#include "plclog.h"
#define debug 5
#define ThisModule "S5Driver : "
//
// Define this to doublecheck the addresses of flags etc. in PLC memory.
//
//#define paranoid
//
// Define the state of the copies of the PLCs memory we hold:
//
#define tnotUsed 0		// we dont use this
#define tUpToDate 1		// copy has been read ok
#define tError 2		// read error, can't use

class S5ti:public transferInstruction {
    DECLARE_PERSISTENCE(S5ti)
    virtual void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(S5ti, "S5")

void S5ti::Read(Engine & e)
{
    string stc, ctc;
    e >> comment >> sheet >> startMW >> endMW >> direction >> DBNumber >>
	stc >> ctc;
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
#if debug >3
    LOG2("comment        :%s\n", comment.c_str());
    LOG2("sheet          :%d\n", sheet);
    LOG2("first word     :%d\n", startMW);
    LOG2("last word:     :%d\n", endMW);
    LOG2("direction:     :%d\n", direction);
    LOG2("DB (data area) :%d\n", DBNumber);
    LOG2("to Cell        :%s\n", cellN(startCell));
    LOG2("condition      :%s\n", cellN(ctrlCell));
#endif
}

class S5drv:public PLCdrv {
  public:
    bool sysParsOk;
    char flagsOk;
    char inputsOk;
    char outputsOk;
    int firstFlag;		//the first byte of flags we copy
    int zFlags;			//number of flag bytes we copy
    int firstOutput;		//the first byte of outputs we copy
    int zOutputs;		//number of outputs bytes we copy
    int firstInput;		//the first byte of inputs we copy
    int zInputs;		//number of inputs bytes we copy
    sysPars sp;
    S5drv();
    unsigned char MMM[512];
    unsigned char PAA[64];
    unsigned char PAE[64];
    unsigned char DB_buffer[1024];
    virtual void execute();
    virtual void execTI(transferInstruction * p);
    virtual void execTI2(transferInstruction * p);
    virtual void
	coreOfSetLoop(int &i, transferInstruction * p, coord & Cell,
		      char step);
    virtual void errorSet(transferInstruction * p);
    virtual void intSetup(DriverInfo * i);
};

void S5drv::errorSet(transferInstruction * p)
{
    coord cell;
    int i;
    int step;

#if debug > 0
    LOG1("enter errorSet\n");
    FLUSH;
#endif
    if (p == NULL)
	return;
    if (p->direction == 0) {
				//read
	    cell = p->startCell;
	i = p->startMW;
	if ((p->DBNumber == -7)  ||(p->DBNumber == -4) ||(p->DBNumber == -1))
	    step = 2;
	else
	    step = 1;
	while (i <= p->endMW) {
#if debug > 0
	    LOG2("errorSet: I set Cell: %s to COM error.\n",
		   cellName(cell).c_str());
	    FLUSH;
#endif				/*  */
	    setState(p->sheet, cell, errorMask, true);
	    cell.row++;
	    i += step;
	}
	usleep(500000);
    }
}

void S5drv::coreOfSetLoop(int &i, transferInstruction * p,
			  coord & cell, char step)
{
    word mVal;
#if debug > 0
    LOG2("coreOfSetLoop %s \n", cellN(cell));
    FLUSH;
#endif				/*  */    
    while (i <= p->endMW) {
	switch (p->DBNumber) {
	case -9:
	    mVal = PAE[i / 8] & (1 << (i % 8));
	    break;
	case -8:
	    mVal = PAE[i];
	    break;
	case -7:
	    mVal = 256 * PAE[i] + PAE[i + 1];
	    break;
	case -6:
	    mVal = PAA[i / 8] & (1 << (i % 8));
	    break;
	case -5:
	    mVal = PAA[i];
	    break;
	case -4:
	    mVal = 256 * PAA[i] + PAA[i + 1];
	    break;
	case -3:
	    mVal = MMM[i / 8] & (1 << (i % 8));
	    break;
	case -2:
	    mVal = MMM[i];
	    break;
	case -1:
	    mVal = 256 * MMM[i] + MMM[i + 1];
	    break;
	default:
	    mVal = DB_buffer[2 * (i - p->startMW)] << 8 
		|DB_buffer[2 * (i - p->startMW) + 1];
	}
	setCell(p->sheet, cell, mVal, soExternal);
#if debug >0
	LOG4("I set %d:%s to %d\n",
	       p->sheet, cellName(cell).c_str(), mVal);
#endif
	cell.row++;
	i += step;
    }
}

void S5drv::execTI2(transferInstruction * p)
{
    int i;
    coord cell;
    int step;
    blockInfo BI;
    word workAdr;
    int flags;
    bool ok;
    unsigned char workCopy;
    double rvalue;
    word rWord;
#if debug > 0
    LOG3("execTI2 DBNumber %d direction %d\n", p->DBNumber, p->direction);
#endif    
    execError = false;
    cell = p->startCell;
    i = p->startMW;
    if ((p->DBNumber == -7)  ||(p->DBNumber == -4) 
 ||(p->DBNumber == -1))
	step = 2;
    else
	step = 1;
    if (p->direction == 0) {	// read 
	switch (p->DBNumber) {
	case -9:
	case -8:
	case -7:
	    if (inputsOk == tUpToDate)
		coreOfSetLoop(i, p, cell, step);
	    break;
	case -6:
	case -5:
	case -4:
	    if (outputsOk == tUpToDate)
		coreOfSetLoop(i, p, cell, step);
	    break;
	case -3:
	case -2:
	case -1:
	    if (flagsOk == tUpToDate)
		coreOfSetLoop(i, p, cell, step);
#if debug > 0
		else LOG1("flagsOk<>tUpToDate\n");
#endif    		
	    break;
	default:		// Data block area 
	    if (getBlockInfo(rfd,wfd, bt_DB, p->DBNumber, &BI)) {
		if (readData(rfd,wfd, BI.address + 2 * p->startMW,
			      2 * (1 + p->endMW - p->startMW),
			      &DB_buffer[0])) {
		    coreOfSetLoop(i, p, cell, step);
		}
		
		else {
		    execError = true;
		    sBreak(rfd,wfd, DBReadFault);
		}
	    } else {
		execError = true;
		sBreak(rfd,wfd, DBNotPresent);
	    };
	    
		// Data block
	}			// switchr }
    } else {
				// dir. read ends ,now for write
	    switch (p->DBNumber) {
	case -3:		// Flag Bits
	    workAdr = 0xffff;	// can never be valid
	    while (i <= p->endMW) {
		flags = getFlags(p->sheet, cell, ok);
		if (ok) {
		    if (flags & changedMask) {
			if (workAdr != (i / 8)) {	// is flag is in 
			    // another byte
			    if (workAdr != 0xffff) {	// write copy back
				// to PLC 
				ok = setPLCvar(opdMB, workAdr, workCopy,
					       &sp, rfd, wfd);
			    }
			    ok = readData(rfd,wfd, sp.Merker + (i / 8), 1,
					   &workCopy);
			    workAdr = i / 8;
			}
			if (numContents(p->sheet, cell, ok) == 0) {
			    workCopy = workCopy & (~(1 << (i % 8))) ;
			} else
			     {
			    workCopy |= (1 << (i % 8));
			    }
			clrPushButton(p->sheet, cell);
		    }		// flags changed 
		}		//  if ok..
		cell.row++;
		i++;
	    }			// while i..
	    if (workAdr != 0xffff) {	// write the copy back into PLC 
		ok = setPLCvar(opdMB, workAdr, workCopy, &sp, rfd, wfd);
	    }
	    break;		// Flag Bits
	case -2:		// Flag Bytes 
	    while (i <= p->endMW) {
		flags = getFlags(p->sheet, cell, ok);
		if (flags & changedMask) {
		    rvalue = numContents(p->sheet, cell, ok);
		    if (rvalue <= 0xff) {
			ok = setPLCvar(opdMB, i, int (rvalue), &sp, rfd, wfd);
			clrPushButton(p->sheet, cell);
		    }
		}		// flags changed 
		cell.row++;
		i++;
	    }			// while i
	    break;
	case -1:		// Flag words
	    while (i <= p->endMW) {
#if debug > 0	    
		LOG2("write: %d\n", i);
		FLUSH;
#endif		
		flags = getFlags(p->sheet, cell, ok);
#if debug > 0		
		LOG3("flags: %x changedMask %x\n", flags,
		       changedMask);
		FLUSH;
#endif		
		if (flags & changedMask) {
		    rvalue = numContents(p->sheet, cell, ok);
		    if (rvalue <= 0xffff) {
			rWord = int (rvalue);
			rWord = _swap(rWord);
			ok = setPLCvar(opdMW, i, rWord, &sp, rfd, wfd);
			clrPushButton(p->sheet, cell);
		    }
		}		// flags changed 
		cell.row++;
		i += step;
	    }			// while i
	    break;
	default:		// data area
	    if (getBlockInfo(rfd, wfd, bt_DB, p->DBNumber, &BI)) {
		while (i <= p->endMW) {
		    flags = getFlags(p->sheet, cell, ok);
		    if (flags & changedMask) {
			rvalue = numContents(p->sheet, cell, ok);
			if (rvalue <= 0xffff) {
			    rWord = int (rvalue);
			    ok =
				setPLCvar(opdDW, BI.address + 2 * i,
				    _swap(rWord), &sp, rfd, wfd);
			    clrPushButton(p->sheet, cell);
			}
		    }		// flags changed 
		    cell.row++;
		    i++;
		}		// while i..
	    } else
		 {		// BlockInfo failed
		sBreak(rfd, wfd, DBNotPresent);
		};		// data blocks
	}			// switch
    }				// direction write
}

void S5drv::execTI(transferInstruction * p)
{
    bool ok;
    execError = false;
#if debug > 0    
    LOG2("condition: %s\n", cellName(p->ctrlCell).c_str());
#endif    
    if (numContents(p->sheet, p->ctrlCell, ok) != 0)
	execTI2(p);
    if (execError)
	errorSet(p);
}

void globalErrorSet(void *ti, void *drv)
{
    ((S5drv *) drv)->errorSet(((transferInstruction *) ti));
}

void S5drv::execute()
{
    bool sysParsOk;
    int errorCode;
//    PID=getpid();
    do {
#if debug > 0
	LOG1("before getSysPars\n");
	FLUSH;
#endif				/*  */
	sysParsOk = getSysPars(rfd, wfd, &sp, false);
#if debug > 0
	LOG1("before getSysPars(verify)\n");
	FLUSH;
#endif	
#ifdef paranoid
	if (sysParsOk)
	    sysParsOk = getSysPars(rfd, wfd, &sp, true);
#endif				/*  */
	sp.Inputs = _swap(sp.Inputs);
	sp.Outputs = _swap(sp.Outputs);
	sp.Merker = _swap(sp.Merker);
	sp.Zeiten = _swap(sp.Zeiten);
	sp.Zaehler = _swap(sp.Zaehler);
	if (!sysParsOk) {
#if debug > 0
	    LOG1("sysParsError!\n");
	    FLUSH;
#endif				/*  */
	    sBreak(rfd, wfd, errorCode + 0x100);
#if debug > 0
	    LOG1("SysParsError , sBreak done.\n");
	    FLUSH;
#endif				/*  */
	    il->forEach(globalErrorSet, this);
	    tcsendbreak(wfd, 410);
#if debug > 0
	    LOG1("TCSendBreak after global error done.\n");
	    FLUSH;
#endif
	    usleep(500 * 1000);
	    execError = true;
	} else {		// System parameters read ok
#if debug > 0
	    LOG1("SysPars Ok.\n");
	    FLUSH;
#endif				/*  */
	    if (flagsOk != tnotUsed) {	//flags to read? }
#if debug > 0
		LOG3("read Flags from FY %d to FY %d\n",
		       firstFlag, firstFlag + zFlags);
		FLUSH;
#endif
		if (readData(rfd, wfd, sp.Merker + firstFlag, zFlags, &MMM[firstFlag])) {
		    flagsOk = tUpToDate;
#if debug > 0
		    LOG1("flags read ok.\n");
		    FLUSH;
#endif				/*  */		    		    
		} else {
		    flagsOk = tError;
#if debug > 0
		    LOG1("flags read error!.\n");
		    FLUSH;
#endif				/*  */		    
		}    
		
	    }
	    if (outputsOk != tnotUsed) {	// outputs to read? }
		if (readData
		    (rfd, wfd, sp.Outputs + firstOutput, zOutputs,
		     &PAA[firstOutput]) )
		    outputsOk = tUpToDate;
		else
		    outputsOk = tError;
	    }
	    if (inputsOk != tnotUsed) {	// inputs to read? }
		if (readData
		    (rfd, wfd, sp.Inputs + firstInput, zInputs,
		     &PAE[firstInput]) )
		    inputsOk = tUpToDate;
		else
		    inputsOk = tError;
	    }
	    
#if debug > 0
	    LOG2("before execList() %d\n", firstFlag + zFlags);
	    FLUSH;
#endif
	    execList();
	}
	execCount++;
    }
    while (1);
}


void setReadRange(void *pv, void *sv)
{
    transferInstruction *p = (transferInstruction *) pv;
    S5drv *s = (S5drv *) sv;
    word b, e;
    if (p->direction == 0)
	switch (p->DBNumber) {
	case -9:
	case -6:
	case -3:
	    b = p->startMW / 8;
	    e = p->endMW / 8;
	    break;
	case -8:
	case -5:
	case -2:
	    b = p->startMW;
	    e = p->endMW;
	    break;
	case -7:
	case -4:
	case -1:
	    b = p->startMW;
	    e = p->endMW + 1;
	    break;
	}
    switch (p->DBNumber) {
    case -9:
    case -8:
    case -7:
	if (b < s->firstInput)
	    s->firstInput = b;
	if (e > s->zInputs)
	    s->zInputs = e;
	break;
    case -6:
    case -5:
    case -4:
	if (b < s->firstOutput)
	    s->firstOutput = b;
	if (e > s->zOutputs)
	    s->zOutputs = e;
	break;
    case -3:
    case -2:
    case -1:
	if (b < s->firstFlag)
	    s->firstFlag = b;
	if (e > s->zFlags)
	    s->zFlags = e;
	break;
    }
}


S5drv::S5drv()
{
}

void S5drv::intSetup(DriverInfo * i)
{
    rfd = setPort(i->s, &wfd);
    LOG2("intSetup wfd: %d\n", wfd);
    TypeManager::Add("S5", CreateNewS5ti);
    PLCdrv::load(i);
    changedMask = registerDriver("Simatic S5", this);
    errorMask = changedMask;
    flagsOk = tError;
    inputsOk = tError;
    outputsOk = tError;
    firstInput = 63;
    firstOutput = 63;
    firstFlag = 511;
    zInputs = 0;
    zOutputs = 0;
    zFlags = 0;
    
    il->forEach(setReadRange, this);
    if (zInputs < firstInput) inputsOk = tnotUsed;
    if (zOutputs < firstOutput) outputsOk = tnotUsed ;
    if (zFlags < firstFlag) flagsOk = tnotUsed;
    zInputs = zInputs - firstInput + 1;
    zOutputs = zOutputs - firstOutput + 1;
    zFlags = zFlags - firstFlag + 1;
#if debug>0    
    if (inputsOk == tnotUsed) {
	LOG1("E-block: no inputs used.\n");
    } else {
	LOG3("E-block: %d %d byte.\n",firstInput,zInputs);    
    }	
    if (outputsOk == tnotUsed) {
	LOG1("A-block: no outputs used.\n");
    } else {
	LOG3("A-block: %d %d byte.\n",firstOutput,zOutputs);
    }	
    if (flagsOk == tnotUsed) {
	LOG1("M-block: no flags used.\n");
    } else {
	LOG3("M-block: %d %d byte.\n",firstFlag,zFlags);
    }
#endif    
}

#include "thread_makro.cc"
#if debug>0
DTDebug(S5drv)
#else
DT(S5drv)
#endif
/*
    Changes:
    
    02/12/2002 change output statements to printf, this gives less code size.
*/
