/*
 Implementation of basic functions for a PLC driver for the SIEMENS S5 
 series via its programmer port.
 See www.runmode.com for information about the underlying AS511
 protocol and a proposal of circuit to connect the current loop programmer
 port to a common V24.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
//#include "s5.h"
#include "s5l.h"
#include "plclog.h"
#define debug 5

#define ThisModule "S5l     : "

#define uc unsigned char

word _swap(word w)
{
    return (256 * lo(w) + hi(w));
}

void sendChar(int fd, unsigned char c)	// 11/04/04
{
#if debug >4
    int res=
#endif	
    write(fd, &c, 1);
#if debug >4
    int rfd=fd;
    LOG3(ThisModule "write to %d : %d\n", fd, res);
    FLUSH;
#endif				/*  */
}
void sendData(int fd, unsigned char n)
{
    sendChar(fd, n);
    if (n == 0x10)
	sendChar(fd, n);	// repeat it, if 0x10
}
void s1006(int fd)
{				// sends the frequently used sequence 0x10 0x06
    sendChar(fd, 0x10);
    sendChar(fd, 0x06);
}

void s1004(int fd)
{				// sends the frequently used sequence 0x10 0x04
    sendChar(fd, 0x10);
    sendChar(fd, 0x04);
}

/*
   Send a 16bit word. SIEMENS S5 uses big endian.
*/ 
void sendWordData(int fd, word w)
{
    sendData(fd, w / 256);
    sendData(fd, w % 256);
}

bool readChars(int rfd, uc *target, size_t size)
{
    bool b;
    unsigned char a;
    word e1;
    fd_set FDS;
    e1 = 0;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
    do {
	FD_ZERO(&FDS);
	FD_SET(rfd, &FDS);
	b = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
	t.tv_sec = 0;
	t.tv_usec = 1000 * tmo_read2;
	if (b) {
	    read(rfd, &a, 1);
	    
#if debug >4
	    LOG2(ThisModule "got %d\n", (unsigned int) a);
#endif
	    *((char *) target + e1) = a;
	    if (a == 0x10) {
		FD_ZERO(&FDS);
		FD_SET(rfd, &FDS);
		b = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
		t.tv_sec = 0;
		t.tv_usec = 1000 * tmo_read2;
		if (b)
		    read(rfd, &a, 1);
	    }
	}
	e1++;
    }
    while ((e1 != size) && b);	//long as we get chars in time and have not enough
    return (b);
}

bool testChars(int rfd, uc *target, int len, bool cntrl)
{
    bool b, res;
    uc a;
    word e1;
    fd_set FDS;
    e1 = 0;
    timeval t;
    res = true;
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
#if debug >3
    LOG2(ThisModule "enter testChars %d chars.\n", len);
#endif
    do {
	FD_ZERO(&FDS);
	FD_SET(rfd, &FDS);
	b = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
#if debug >3
	LOG2(ThisModule "testChars select has chars? %d\n", b);
#endif
	t.tv_sec = 0;
	t.tv_usec = 1000 * tmo_read2;
	if (b) {
#if debug >3
	    int r =
#endif
		read(rfd, &a, 1);
	    
#if debug >3
		LOG2(ThisModule "read: %d\n", r);
#endif
	    if (*(target + e1) != a) {
		res = false;
	    }
#if debug >3
	    LOG3(ThisModule "wanted char %d, got %d\n",
		  (int) *(target + e1), (int) a);
#endif
	    if ((a == 0x10) && (!cntrl)) {
		FD_ZERO(&FDS);
		FD_SET(rfd, &FDS);
		b = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0); 
		t.tv_sec = 0;
		t.tv_usec = 1000 * tmo_read2; 
		if (b) read(rfd, &a, 1);
	    }
	}
	e1++; 
    }	while ((e1 != len) && b);	//the longest block is 45 char
    return (b);
}; 

/*
  sBreak interrupts the cummunication in progress, waits 500 ms
  if the PLC might stil send something and empties the receive
  buffer then.
*/

void sBreak(int rfd,int wfd, word lev) {
    bool charsAvail;
    fd_set FDS;
    int n;
    char waste[256];
    timeval t;
    t.tv_sec = 0; 
    t.tv_usec = 10000;
#if debug > 0
    LOG5(ThisModule"sBreak %d %d  %d %d\n",
	lev / 256, lev % 256, hi(lev),
	lo(lev)); 
    FLUSH;
#endif				/*  */
/*
    tdp->lastError.level1=lo(lev);
    tdp->lastError.level2=hi(lev);
    tdp->lastError.time=time(NULL);
*/
    if (hi(lev) == 1) {
	sendChar(wfd, 0x10);
	sendChar(wfd, 0x04);
	usleep(500 * 1000); 
    } else {
	if (hi(lev) == 2) {
	    tcsendbreak(wfd, 2 /* 410 */);
	    usleep(300 * 1000);}
	}
	do {
	    FD_ZERO(&FDS);
	    FD_SET(rfd, &FDS);
	    charsAvail =  (select(rfd + 1, &FDS, NULL, NULL, &t) > 0); 
	    if (charsAvail) {
		n = read(rfd, &waste, 256);
#if debug > 0
	    LOG2(ThisModule "sBreak wasting %d chars\n", n);
	    FLUSH;
#endif
	}
    }
    while (charsAvail);
}

/*
  Prolog to start function number n on a S5 PLC:
  
  PC to PLC: 0x02                                                      
  PLC to PC: 0x10 0x06                                                   
  PC to PLC: function number
  PLC to PC: 0x02           
  PC to PLC: 0x10 0x06
  PLC to PC: 0x16 0x10 0x03
  PC to PLC: 0x10 0x06
  
  ...execution of function n...
  
  Epilog after execution:
    
  PLC to PC: 0x10 0x03
  PC to PLC: 0x10 0x06
  PLC to PC: 0x02
  PC to PLC: 0x10 0x06
  PLC to PC: 0x12 0x10 0x03
  PC to PLC: 0x10 0x06

*/ 
bool prolog(int rfd,int wfd, unsigned char n) {
	uc x1[] = {0x10, 0x06}; 
	uc x2[] = {0x02}; 
	uc x3[] = {0x16, 0x10, 0x03};
#if debug > 0
	LOG1(ThisModule "begin prolog\n");
	FLUSH;
#endif
	sendChar(wfd, 0x02);
#if debug > 0	
	LOG1(ThisModule "before testChars\n");
#endif	
	if (!testChars(rfd, x1, sizeof(x1), true)) {
#if debug > 0		
	    LOG1(ThisModule"after testChars failed\n");
#endif		 
	    sBreak(rfd, wfd, SF_T1006 + 0x100);
#if debug > 0		    
	    LOG1(ThisModule"after sBreak. will return.\n");
#endif
	    return (false);
	}
#if debug > 0
	LOG1(ThisModule"after testChars succeeded\n");
#endif	
	sendChar(wfd, n);
	if (!testChars(rfd, x2, sizeof(x2), true)) {
	        sBreak(rfd, wfd, SF_T02 + 0x100);
		return (false);
	};
	s1006(wfd);
	if (!testChars(rfd, x3, sizeof(x3), true)) {
	    sBreak(rfd, wfd, SF_T161003 + 0x100);
	    return (false);
	};
#if debug > 0	
	LOG1(ThisModule"after testChars succeeded\n");
#endif	
	s1006(wfd);
#if debug > 0	
	LOG1(ThisModule "testChars ok\n");
#endif		
	return (true);
}

bool epilog(int rfd, int wfd) {
    uc x1[] = {0x10, 0x03}; 
    uc x2[] = {0x02}; 
    uc x4[] = {0x12, 0x10, 0x03};
    if (!testChars(rfd, x1, sizeof(x1), true)) {
	sBreak(rfd, wfd, EF_T1003 + 0x100);
	return (false);
    }
    s1006(wfd);
    if (!testChars(rfd, x2, sizeof(x2), true)) {
	sBreak(rfd, wfd, EF_T02 + 0x100);
	return (false);
    }
    s1006(wfd);
    if (!testChars(rfd, x4, sizeof(x4), true)) {
	sBreak(rfd, wfd, EF_T121003 + 0x100);
	return (false);
    }
    s1006(wfd); 
    return (true);
}


bool getSysPars(int rfd, int wfd, sysPars * sp, bool comp) {
    int errorCode; errorCode = 0; 
    uc x1[] = {0x10, 0x06, 0x02}; 
    if (!prolog(rfd, wfd, 0x18)) {
	errorCode = GSP_SF_1;
	goto SysParsError; 
    }
    s1004(wfd);
    if (!testChars(rfd, x1, sizeof(x1), true)) {
	errorCode = GSP_T100602_1;
	goto SysParsError; 
    }
    s1006(wfd); 
    if (comp) {
#if debug > 0	
	LOG1(ThisModule "comparing SysPars\n");
	FLUSH;
#endif		
	if (!testChars(rfd, (uc*)sp, 45, false)) {
	    errorCode = GSP_ReadCh;
	    goto SysParsError; 
	}
    } else {
	if (!readChars(rfd, (uc *)sp, 45)) {
	    errorCode = GSP_ReadCh;
	    goto SysParsError; 
	}
    }
    if (!epilog(rfd, wfd)) {
	errorCode = GSP_EF_1;
	goto SysParsError; 
    }
    return (true); 
SysParsError:
    return (false);
}

bool getBlockInfo(int rfd, int wfd, unsigned char type,
       unsigned char n,
       blockInfo * BI) {
    uc x5[] = {0x10, 0x06, 0x02}; 
    uc x2[] = {0x02}; 
    uc x4[] = {0x12, 0x10, 0x03}; 
    if (!prolog(rfd,wfd, 0x1A)) {
	sBreak(rfd,wfd, GBI_SF + 0x200);
	return (false);
    }
#if debug > 0	    
    LOG1(ThisModule"getBlockInfo: prolog done.\n");
#endif    
    sendData(wfd, type); 
    sendData(wfd, n);
    s1004(wfd);
    if (!testChars(rfd, x5, sizeof(x5), true)) {
	sBreak(rfd, wfd, GBI_T100602 + 0x100);
	return (false);}
	s1006(wfd);
	if (!readChars(rfd, (uc*)BI, sizeof(blockInfo))) {
	    s1006(wfd);
	    if (!testChars(rfd, x2, sizeof(x2), true)) {
		sBreak(rfd,wfd, GBI_ntrdch_Test02 + 0x100);
		return (false);
	    }
	    s1006(wfd);
	    if (!testChars(rfd, x4, sizeof(x4), true)) {
		sBreak(rfd,wfd, GBI_ntrdch_Test121003 + 0x100);
		return (false);
	    }
	    s1006(wfd); 
	}
	LOG1(ThisModule"getBlockInfo: readChars done.\n");
	BI->address = _swap(BI->address);
	if (!epilog(rfd,wfd)) {
	    sBreak(rfd,wfd, GBI_rdch_EF + 0x200);
	    return (false);
	}
	LOG2(ThisModule"getBlockInfo: success. %x\n",BI->address); 
	return (true);
    }

    bool setPLCvar(char ot, word address, word value, sysPars * sp, int rfd, int wfd, int _PID) {
	word absadr; 
	absadr = address;
	uc x6[] = {0x10, 0x06, 0x02}; 
	uc x7[] = {0x12, 0x10, 0x03}; 
	switch (ot) {
	    case opdAB:
	    case opdAW:
		absadr = address + sp->Outputs; break; 
	    case opdEB:
	    case opdEW:
		absadr = address + sp->Inputs; break; 
	    case opdMB:
	    case opdMW:
		absadr = address + sp->Merker; break;
	}
	if (!prolog(rfd,wfd, 0x03)) {
	    sBreak(rfd,wfd, SVAR_SF + 0x200);
	    return (false); 
	}
	sendWordData(wfd, absadr);
	sendData(wfd, lo(value));
	if ((ot == opdAW) || (ot == opdEW) || (ot == opdMW) || (ot == opdDW))
	    sendData(wfd, hi(value)); 
	    s1004(wfd);
	    if (!testChars(rfd, x6, sizeof(x6), true)) {
		sBreak(rfd,wfd, SVAR_T100602 + 0x200);
	        return (false); 
	    }
	    s1006(wfd);
	    if (!testChars(rfd, x7, sizeof(x7), true)) {
		sBreak(rfd,wfd, SVAR_T121003 + 0x100);
	    	return (false); 
	    }
	    s1006(wfd); return (true); 
	}
	
	bool readData(int rfd,int wfd, word sa, word count, unsigned char *target, int _PID) {
	    uc x8[] = {0x10, 0x06, 0x02}; 
	    uc x9[] = {0, 0, 0, 0, 0}; 
	    if (!prolog(rfd,wfd, 4)) {
		sBreak(rfd,wfd, read_SF + 0x200);
		return (false);
	    }
//
//  send start address:
//
	    sendWordData(wfd, sa);
//
//  send end address:
//    
	    sendWordData(wfd, sa + count - 1);
	    s1004(wfd);
	    if (!testChars(rfd, x8, sizeof(x8), true)) {
		sBreak(rfd,wfd, read_T100602 + 0x100);
	        return (false);
	    }
	    s1006(wfd);
	    if (!testChars(rfd, x9, sizeof(x9), true)) {
		sBreak(rfd,wfd, read_T00000 + 0x100);
	        return (false);
	    }
	    if (!readChars(rfd, target, count)) {
	        sBreak(rfd, wfd, read_readCh + 0x100);
	        return (false);
	    }
	    if (!epilog(rfd,wfd)) {
		sBreak(rfd,wfd, read_EndF + 0x100);
	        return (false);
	    }
	    return (true); 
	}
