/*
 ediCellList.cc * adds insert/delete functionality to cellList.
 this is not in cellList, to keep minimum runtime code small for
 embedded systems. Could be  moved into a shared library for load
 on demand ?.

 Part of VISUAL, a human machine interface and data acquisition program

 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001. 

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include "cell.h"
#include "cellList.h"
#include "ediCellList.h"
#include "parser.h"
#include "lib0.h"
#include "shift.h"
#include <ctype.h>
#include <stdio.h>

#include "vi_string.h"

#include "plclog.h"
#include "refCell.h"

#define debug 10

#define ThisModule "ediCellList: "

void
 ediCellList::deleteCell(coord c)
{
#if debug >0
    LOG2(ThisModule "delete cell %s\n", cellN(c));
    FLUSH;
#endif
    cell *cl = find(c);
    if (cl == NULL) {
#if debug >0
	LOG2(ThisModule "there is no cell %s\n", cellN(c));
#endif
	return;
    }
    if (cl->isNumber()) {
	nCell *c2;
#if debug >0
	LOG2(ThisModule "searching numList for %s\n", cellN(c));
	FLUSH;
#endif
	c2 = (nCell *) numList;
	if (c2 == NULL) {
	    LOG1(ThisModule
		   "can't happen: numeric cell, but numList empty !!!\n");
	    FLUSH;
	} else {
	    if (c2->k.l == c.l) {
#if debug >0
		LOG2(ThisModule "%s is first.\n", cellN(c2));
		FLUSH;
#endif
		numList = c2->nextNum;
#if debug >0
		LOG2(ThisModule "%s removed.\n", cellN(c2));
		FLUSH;
#endif
	    } else {
		while ((c2->nextNum->k.l != c.l) && (c2 != NULL)) {
		    c2 = c2->nextNum;
		}
		if (c2 == NULL) {
		    LOG1(ThisModule
			   "can't happen: numeric cell not in numList!!!\n");
		    FLUSH;
		} else {
#if debug >0
		    LOG3(ThisModule "%s found. It points to:%s\n",
			   cellN(c2), cellN(c2->nextNum));
		    FLUSH;
#endif
		    c2->nextNum = c2->nextNum->nextNum;
#if debug >0
		    if (c2->nextNum)
			LOG3(ThisModule "%s now points to: %s\n",
			       cellN(c2), cellN(c2->nextNum));
		    else
			LOG2(ThisModule "%s now points nowhere.\n",
			       cellN(c2));
		    FLUSH;
#endif
		}
	    }
	}
    }
    if (cl->isFormula()) {
	fCell *c3;
#if debug >0
	LOG2(ThisModule "searching calcList for %s\n", cellN(c));
	FLUSH;
#endif
	c3 = (fCell *) calcList;
	if (c3 == NULL) {
	    LOG1(ThisModule
		   "can't happen: formula cell, but calcList empty !!!\n");
	} else {
	    if (c3->k.l == c.l) {
#if debug >0
		LOG2(ThisModule "%s is first.\n", cellN(c3));
		FLUSH;
#endif
		calcList = c3->nextFormula;
#if debug >0
		LOG2(ThisModule "%s removed.\n", cellN(c3));
		FLUSH;
#endif
	    } else {
		while ((c3->nextFormula->k.l != c.l) && (c3 != NULL)) {
		    c3 = c3->nextFormula;
		}
		if (c3 == NULL) {
		    LOG1(ThisModule
			   "can't happen: formula cell not in calcList!!!\n");
		    FLUSH;
		} else {
#if debug >0
		    LOG3(ThisModule "%s found. It points to:%s\n",
			   cellN(c3), cellN(c3->nextFormula));
		    if (c3->nextFormula->nextFormula) {
			LOG3(ThisModule "%s points to:%s\n",
			       cellN(c3->nextFormula),
			       cellN(c3->nextFormula->nextFormula));
		    } else {
			LOG2(ThisModule "%s is last.\n",
			       cellN(c3->nextFormula));
		    }
		    FLUSH;
#endif
		    c3->nextFormula = c3->nextFormula->nextFormula;
#if debug >0
		    if (c3->nextFormula)
			LOG3(ThisModule "%s now points to: %s\n",
			       cellN(c3), cellN(c3->nextFormula));
		    else
			LOG2(ThisModule "%s now points nowhere.\n",
			       cellN(c3));
		    FLUSH;
#endif

		}
	    }
	}
    }

    list::free(cl);
    changed = true;
#if debug >0
    LOG2(ThisModule "cell %s deleted.\n", cellN(c));
#endif

}

/*
    create a cell from a text string:
*/
void ediCellList::createCell(coord c, string s, int sourceT)
{
    if (s.empty())
	return;
    char s0;
#ifndef use_string_2
    s0 = s.at(0);
#else
    s0 = s.elem(0);
#endif
    bool isTimed = false, isFormula = false;
    cell *ce;
    char *p;
    char **sp = &p;
#if debug >0
    LOG2(ThisModule "1st char: %c\n", s0);
#endif
    if (s0 == '*') {		// prefix for time recording
#ifndef use_string_2
	s = s.erase(0, 1);
#else
	s.del(0, 1);
#endif
#if debug >0
	LOG2(ThisModule "rest: %s\n", s.c_str());
#endif
	isTimed = true;
#ifndef use_string_2
	s0 = s.at(0);
#else
	s0 = s.elem(0);
#endif
#if debug >0
	LOG2(ThisModule "1st char: %c\n", s0);
#endif
    }
    if ((s0 == '=') || (s0 == '+')) {	// prefix for formula
#ifndef use_string_2
	s = s.erase(0, 1);
#else
	s.del(0, 1);
#endif
#if debug >0
	LOG2(ThisModule "rest: %s\n", s.c_str());
#endif
	isFormula = true;
    }
    if (isFormula) {
	if (isTimed) {
	    ce = new fTimeCell();
	} else {
	    ce = new fCell(s);
	}
//	((fCell *) ce)->text = s;
#if debug >0
	LOG2(ThisModule "before: %d\n", ((fCell *) ce)->TPN.length());
#endif
//	((fCell *) ce)->TPN = parse(s);
#if debug >0
//	LOG2(ThisModule "after: %d\n", ((fCell *) ce)->TPN.length());
#endif
    } else {
	double d = strtod(s.c_str(), sp);
#if debug >0
//      printf ("strtod: p:%s *p:%d\n",p,*p);
#endif
	if ((*p) == 0) {
	    if (isTimed) {
		ce = new nTimeCell();
	    } else {
		ce = new nCell();
	    }
	    ce->setVal(d);
	    ce->setSourceType(sourceT);
	} else {
	    if (s.substr(0,2)=="->") {
		ce = new RefCell();
		((RefCell *) ce)->text = s.from(2);
		((RefCell *) ce)->setupFromString();
#if debug > -1
		LOG1(ThisModule "create done.\n");
		FLUSH;
#endif		
	    } else {
		ce = new tCell();
		((tCell *) ce)->text = s;
	    }	
	}

    }
    ce->k = c;
#if debug > -1
    LOG1(ThisModule "before insert.\n");
    FLUSH;
#endif    
    insert(ce);
#if debug > -1
    LOG1(ThisModule "insert done.\n");
    FLUSH;
#endif    
    changed = true;
}

void ediCellList::insertLabel(coord c, string s)
{
    if (s.empty())
	return;
    cell *ce = find(c);
    if (ce != NULL)
	ce->Label = s;
    changed = true;
}

void ediCellList::insertLabelAndCFlags(coord c, string s, int &cFlags)
{
//    if (s.empty())
//      return;
    cell *ce = find(c);
    if (ce != NULL) {
	if (!s.empty())
	    ce->Label = s;
	ce->setCflags( cFlags);
    }
    changed = true;
}

void ediCellList::mark(coord c, int b)
{
    cell *ce = find(c);
    if (ce != NULL) {
	if (b == markToggle)
	    ce->setFlags( ce->getFlags() ^ flMarked);
	else if (b == markTrue)
	    ce->setFlags( ce->getFlags() | flMarked);
	else if (b == markFalse)
	    ce->setFlags( ce->getFlags() & (~flMarked));
    }
}

void ediCellList::withArea(int what, coord st, coord en, string fn,
			   bool & valid)
{
    coord c;
    cell *f;
    string Contents, symName;
    for (c.col = st.col; c.col <= en.col; c.col++) {
	for (c.row = st.row; c.row <= en.row; c.row++) {
	    switch (what) {
	    case 1:
		deleteCell(c);
		break;
	    case 2:
		mark(c, true);
		break;
/*	
        case srSAVE: f=find(c);
            if (!=NULL) F->Write(s);
             break;
*/
	    case 4:
		f = find(c);
		if (f != NULL) {
		    Contents = f->stringContents(2, 7);
		    symName = symbolName(c, symOnly);
		    deleteCell(c);
		    createCell(c, "*" + Contents, soDirect);
		    if (!symName.empty()) {
			insertLabel(c, symName);
		    }		// empty
		}		// if not NULL
		break;
	    }
	    			// switch
	}			//for
    }				//for
    valid = false;
}

void ediCellList::insertShiftedCell(coord c, string s, int sourceT, int sp,
				    int ze)
{
    if (s[0] == '=') {		//is it a formula ? }
	shift(s, sp, ze);
    }
    createCell(c, s, sourceT);
}

void shiftCellRefs(string & s, int sp, int ze)
{
    unsigned int i = 0;
    enum pstate { s_text, s_col, s_row } state = s_text;
    coord c;
    string s2 = "", cName;
    s = s + "a";		/* mark the end so we end in state s_text */
    do
	 {
/*      LOG3(ThisModule"%d %d\n",s.at(i),state);
*/
	switch (state) {
	case s_text:
	    if (s.at(i) != '$')
		s2 += s[i] ;
	    else {
		state = s_col;
		cName = "";
	    }
	    break;
	case s_col:
#ifndef use_string_2
	    if ((toupper(s.at(i)) >= 'A') && (toupper(s.at(i)) <= 'Z'))
		cName += s.at(i) ;
#else
	    if ((toupper(s.elem(i)) >= 'A') && (toupper(s.elem(i)) <= 'Z'))
		cName += s.elem(i) ;
#endif

	    else {
		state = s_row;
		i--;
	    }
	    break;
	case s_row:
	    if ((s.at(i) >= '0') && (s.at(i) <= '9'))
		cName += s.at(i) ;
	    else {
		state = s_text;
		LOG2(ThisModule "cName: %s", cName.c_str());
		c = getPos(cName);
		FLUSH;
		c.row += ze;
		c.col += sp;
		s2 = s2 + "$";
		s2 = s2 + cellName(c) + s[i];
		
//         LOG2(ThisModule"result: %s\n",s2.c_str());
	    }
	    break;
	}			// switch
	i++;
	}
    while (i < s.length());
#ifndef use_string_2
    s2.erase(s2.length() - 1, 1);
#else
    s2.del((int) s2.length() - 1, 1);
#endif
    s = s2;
}

void ediCellList::copyCells(coord st, coord en, coord cur, bool doShift)
{
    cell *ce;
    coord p1, p2;
    word sp, ze;
    string symName;
    if (doShift) {
    	ze = cur.row - st.row;
    	sp = cur.col - st.col;
    } else {
	ze=sp=0;
    }	
    for (p1.col = st.col; p1.col <= en.col; p1.col++) {
	p2.col = p1.col - st.col + cur.col;
	for (p1.row = st.row; p1.row <= en.row; p1.row++) {
	    ce = find(p1);
	    p2.row = p1.row - st.row + cur.row;
	    if (ce != NULL) {
		insertShiftedCell(p2, ce->stringContents(7, 2), soDirect,
				  sp, ze);
		symName = symbolName(p1, symOnly);
		if (!symName.empty()) {
		    shiftCellRefs(symName, sp, ze);
		    insertLabel(p2, symName);
		}
	    }
	}
    }
}

/*
    Changes:
    05/13/2002 Object Read & write functions return void now. cout<<... replaced
		by printf..
    
    26/05/2002 preserve module registration flags when editing a cell
    
    25/06/2002 flmarked-> flMarked
    
    21/09/2002 added doShift to copyCells
    
*/
