/*
 Implements a function to evaluate a formula given in tokenized postfix
 notation (TPN).

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001. 

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include "cell.h"
#include "lib0.h"
#include "pString.h"
#include "cellList.h"
#include <math.h>
#include <stdio.h>
#include "globals.h"
#include "ptokens.h"
#include "plclog.h"
#define debug 0
#define ThisModule "evalTPN  : "
#define tpnStackLen 100

double evalTPN(pString & s, cellList & cl, int &error, unsigned int &state)
{
#if debug > 2
    LOG1(ThisModule"this is evalTPN\n");
    FLUSH;
#endif
    double d[tpnStackLen];
    double *p, *q;
    p = &d[0];
    coord c;
    cell *ce;
    int sheet, i = s.length(), j = 0, j5;
    if (i == 0) {
#if debug > 0
	LOG1(ThisModule"TPN is empty!!\n");
	FLUSH;
#endif
	return (0);
    }
    error = 0;
    state = 0;
    char token;
    while (j < i) {
	token = s.at(j);
	j++;
#if debug > 1
	LOG3(ThisModule"%d at index: %d\n", (int) token, j);
	FLUSH;
#endif
	switch (token) {
	case ptADD:
#if debug >3
	    LOG2(ThisModule"ADD %g", *p);
#endif
	    q = p;
	    p--;
#if debug >3
	    LOG2(ThisModule"+%g", *p);
#endif
	    *p = *p + *q;
#if debug >3
	    LOG2(ThisModule"=%g\n", *p);
#endif
	    break;
	case ptMUL:
#if debug > 1
	    LOG2(ThisModule"MUL %g", *p);
#endif
	    q = p;
	    p--;
#if debug >1
	    LOG2(ThisModule"*%g", *p);
#endif
	    *p = *p * (*q);
#if debug > 1
	    LOG2(ThisModule"=%g\n", *p);
#endif
	    break;
	case ptMINUS:
	    *p = -*p;
#if debug >3
	    LOG1(ThisModule"NEG\n");
#endif
	    break;
	case ptRECIP:
#if debug >2
	    LOG1(ThisModule"before 1/x\n");
#endif
	    if (*p == 0) {
		error |= ceDiv0;
#if debug >2
		LOG1(ThisModule"x=0, error flag set\n");
#endif
	    } else {
		*p = 1 / (*p);
#if debug >2
		LOG1(ThisModule"1/x done.\n");
#endif
	    }
	    break;
	case ptLESS:
#if debug >2
	    LOG2(ThisModule"LT %g", *p);
#endif
	    q = p;
	    p--;
#if debug >2
	    LOG2(ThisModule"< %g", *p);
#endif
	    if ((*q) < (*p))
		*p = 1;
	    else
		*p = 0;
#if debug >2
	    LOG2(ThisModule"= %g\n", *p);
#endif
//              abort();
	    break;

	case ptEQUAL:
#if debug >2
	    LOG2(ThisModule"EQU %g", *p);
#endif
	    q = p;
	    p--;
#if debug >2
	    LOG2(ThisModule"= %g", *p);
#endif
	    if ((*q) == (*p))
		*p = 1;
	    else
		*p = 0;
#if debug >2
	    LOG2(ThisModule"= %g\n", *p);
#endif
//              abort();
	    break;
	case ptGREATER:
#if debug >2
	    LOG2(ThisModule"GT %g", *p);
#endif
	    q = p;
	    p--;
#if debug >2
	    LOG2(ThisModule"> %g", *p);
#endif
	    if ((*q) > (*p))
		*p = 1;
	    else
		*p = 0;
#if debug >2
	    LOG2(ThisModule"= %g\n", *p);
#endif
//              abort();
	    break;
#ifdef handleBorlandReal	    
	case ptDOSNumber:	// this case was left for compatibility with an earlier
	    // DOS version
	    p++;
	    *p = s.getDosReal(j);
#if debug > 1
	    LOG2(ThisModule"TURBO PASCAL real %g\n", *p);
#endif
	    break;
#endif	    
	case ptRAND:
	    p++;
	    *p = rand();
	    *p /= (RAND_MAX + 1);
	    *p += 0.5;
#if debug > 1
	    LOG2(ThisModule"random %g\n", *p);
#endif
	    break;
	case ptDOUBLE:
	    p++;
	    *p = s.getDouble(j);
#if debug > 1
	    LOG2(ThisModule"double: %g\n", *p);
#endif
	    break;
	case ptSTATE:
//          LOG2(ThisModule"state\n");
	    j5 = j - 5;		//This should be the position of the cell coord
	    c = s.getCoord(j5);
	    ce = cl.find(c);
	    if (ce != NULL) {
		p--;
		*p = ce->getState();
		error |= ce->getError();
#if debug >0
		LOG3(ThisModule"state %s=%g\n", cellN(c), *p);
#endif
//              p++;
	    } else
		error |= ceMissing;
	    break;
	case ptERR:
	    j5 = j - 5;		//This should be the position of the cell coord
	    c = s.getCoord(j5);
	    ce = cl.find(c);
	    if (ce != NULL) {
		error = 0;
/*		
		if (ce->typeID() == tfCell) {
		    *p = ((fCell *) ce)->mathError;
//                  error |= ce->getError();
		    state |= ce->state;
		} else {
		    if (ce->typeID() == tfTimeCell) {
			*p = ((fTimeCell *) ce)->mathError;
//                      error |= ce->getError();
			state |= ce->state;
*/		
		if (ce->isFormula()) {
//			*p = ((fTimeCell *) ce)->mathError;
			*p = ce->getError();
//                      error |= ce->getError();
			state |= ce->getState();
		    } else
			error |= ceWrongType;
//		}
#if debug >0
		LOG3(ThisModule"error %s=%g\n", cellN(c), *p);
#endif
	    } else
		error |= ceMissing;
	    break;
	case ptTIME:
	    j5 = j - 5;		//This should be the position of the cell coord
	    c = s.getCoord(j5);
	    ce = cl.find(c);
	    if (ce != NULL) {
		if (ce->typeID() == tnTimeCell) {
		    *p = ((nTimeCell *) ce)->time;
		    error |= ce->getError();
		    state |= ce->getState();
		} else {
		    if (ce->typeID() == tfTimeCell) {
			*p = ((fTimeCell *) ce)->time;
			error |= ce->getError();
			state |= ce->getState();
		    } else
			error |= ceWrongType;
		}
#if debug >0
		LOG3(ThisModule"time %s=%g\n", cellN(c), *p);
#endif
//              p++;
	    } else
		error |= ceMissing;
	    break;
	case ptPREV:
	    j5 = j - 5;		//This should be the position of the cell coord
	    c = s.getCoord(j5);
	    ce = cl.find(c);
	    if (ce != NULL) {
		*p = ce->getPreVal();
		error |= ce->getError();
		state |= ce->getState();
	    } else {
		error |= ceMissing;
	    }
#if debug >0
	    LOG3(ThisModule"prev %s=%g\n", cellN(c), *p);
#endif
//              p++;
	    break;
	case ptDIFF:
	    j5 = j - 5;		//This should be the position of the cell coord
	    c = s.getCoord(j5);
	    ce = cl.find(c);
	    if (ce != NULL) {
		*p = ce->getVal()-ce->getPreVal();
		error |= ce->getError();
		state |= ce->getState();
	    } else {
		error |= ceMissing;
	    }
#if debug >0
	    LOG3(ThisModule"prev %s=%g\n", cellN(c), *p);
#endif
//              p++;
	    break;    
	case ptSQRT:
	    if (*p >= 0) {
		*p = sqrt(*p);
#if debug >2
		LOG2(ThisModule"SQR %g\n", *p);
#endif
	    } else {
		error = error | ceRootNeg;
	    }
	    break;
	case ptLN:
	    if (*p > 0) {
		*p = log(*p);
#if debug >2
		LOG2(ThisModule"LOG %g\n", *p);
#endif
	    } else {
		error = error | ceLogNeg;
	    }
	    break;
	case ptEXP:
	    *p = exp(*p);
#if debug >2
	    LOG2(ThisModule"EXP %g\n", *p);
#endif
	    break;
	case ptLOG10:
	    if (*p > 0) {
		*p = log10(*p);
#if debug >2
		LOG2(ThisModule"LOG10 %g\n", *p);
#endif
	    } else {
		error = error | ceLogNeg;
	    }
	    break;
	case ptSIN:
	    *p = sin(*p);
#if debug >2
	    LOG2(ThisModule"SIN %g\n", *p);
#endif
	    break;
	case ptCOS:
	    *p = cos(*p);
#if debug >2
	    LOG2(ThisModule"COS %g\n", *p);
#endif
	    break;
	case ptTAN:
	    *p = tan(*p);
#if debug >2
	    LOG2(ThisModule"TAN %g\n", *p);
#endif
	    break;
	case ptARCSIN:
	    *p = asin(*p);
#if debug >2
	    LOG2(ThisModule"ASIN %g\n", *p);
#endif
	    break;
	case ptARCCOS:
	    *p = acos(*p);
#if debug >2
	    LOG2(ThisModule"ACOS %g\n", *p);
#endif
	    break;
	case ptARCTAN:
	    *p = atan(*p);
#if debug >2
	    LOG2(ThisModule"ATAN %g\n", *p);
#endif
	    break;
	case ptABS:
	    if (*p < 0)
		*p = -(*p);
#if debug >2
	    LOG2(ThisModule"ABS %g\n", *p);
#endif
	    break;
	case ptINT:
	    *p = (int) *p;
#if debug > 1
	    LOG2(ThisModule"INT %g\n", *p);
#endif
	    break;
	case ptImpC:
	    sheet = s.getWord(j);
	    j++;
	    c = s.getCoord(j);
	    if (sheets[sheet] != NULL) {
		ce = sheets[sheet]->find(c);
		goto nonIndexed2;
	    } else {
		error |= ceMissingSheet;
	    }
#if debug > 0
	    LOG3(ThisModule"import cell %s from sheet %d\n", cellN(c), sheet);
#endif
	    break;
	case ptIndC:		// continues with code for case 'I' fetch cell value!
#if debug > 1
	    LOG1(ThisModule"indexed cell\n");
#endif
	    c = s.getCoord(j);
#if debug > 1
	    LOG3(ThisModule"base %s index %g\n", cellN(c), *p);
#endif
	    c.row += (int) (*p);
#if debug > 1
	    LOG2(ThisModule"real coord. %s\n", cellN(c));
#endif
	    goto nonIndexed;
	case ptCELL:
	    p++;
	    c = s.getCoord(j);
	  nonIndexed:
	    ce = cl.find(c);
	  nonIndexed2:
	    if (ce != NULL) {
		if (ce->typeID() == ttCell)
		    error |= ceWrongType;
		else {
		    *p = ce->getVal();
		    error |= ce->getError();
		    state |= ce->getState();
#if debug > 1
		    LOG3(ThisModule"value (%s): %g\n", cellN(c), *p);
#endif
		}
	    } else {
		error |= ceMissing;
#if debug > 1
		LOG2(ThisModule"cell %s not found!\n", cellN(c));
#endif
	    }
//              abort();
//              if (token=='Y') abort();
	    break;
	case ptNOTEQUAL:
	    q = p;
	    p--;
	    if (*p != *q)
		*p = 1;
	    else
		*p = 0;
	    break;
	case ptAND:
	    q = p;
	    p--;
#if debug >2
	    LOG3(ThisModule"%g AND %g", *p, *q);
#endif
	    *p = ((int) *p & (int) (*q));
#if debug >2
	    LOG2(ThisModule"=%g\n", *p);
#endif
	    break;
	case ptOR:
	    q = p;
	    p--;
#if debug >2
	    LOG3(ThisModule"%g OR %g", *p, *q);
#endif
	    *p = ((int) *p | (int) (*q));
#if debug >2
	    LOG2(ThisModule"=%g\n", *p);
#endif
	    break;
	case ptXOR:
	    q = p;
	    p--;
#if debug >2
	    LOG3(ThisModule"%g XOR %g", *p, *q);
#endif
	    *p = ((int) *p ^ (int) (*q));
#if debug >2
	    LOG2(ThisModule"=%g\n", *p);
#endif
	    break;
	case ptHOURS:
	    p++;
	    *p = glHours;
#if debug > 1
	    LOG1(ThisModule"HOURS\n");

#endif
	    break;
	case ptMINUTES:
	    p++;
	    *p = glMinutes;
#if debug > 1
	    LOG1(ThisModule"MINUTES\n");

#endif
	    break;
	case ptSECONDS:
	    p++;
	    *p = glSeconds;
#if debug > 1
	    LOG1(ThisModule"SECONDS\n");
#endif
	    break;
	case ptUSECS:
	    p++;
	    *p = glUSeconds;
#if debug > 1
	    LOG1(ThisModule"USECONDS\n");
#endif
	    break;	    
	case ptYEAR:
	    p++;
	    *p = glYear;
#if debug > 1
	    LOG1(ThisModule"YEAR\n");
#endif
	    break;
	case ptMONTH:
	    p++;
	    *p = glMonth;
#if debug > 1
	    LOG1(ThisModule"MONTH\n");
#endif
	    break;
	case ptDAY:
	    p++;
	    *p = glDay;
#if debug > 1
	    LOG1(ThisModule"DAY\n");
#endif
	    break;
	case ptDOW:
	    p++;
	    *p = glDayOfWeek;
#if debug > 1
	    LOG1(ThisModule"DayOfWeek\n");
#endif
	    break;
	case ptMIN:
	    d[0] = *p;
	    while (p != &d[0]) {
#if debug > 1
		LOG3(ThisModule"MIN %g %g\n", *p, d[0]);
#endif
		p--;
		if (*p < d[0])
		    d[0] = *p;
#if debug > 1
		LOG3(ThisModule"MIN %g, %g\n", *p, d[0]);
#endif
	    }
	    p = &d[0];
#if debug > 1
	    LOG3(ThisModule"MIN %g %g\n", *p, d[0]);
#endif
	    break;
	case ptMAX:
	    d[0] = *p;
	    while (p != &d[0]) {
#if debug > 1
		LOG3(ThisModule"MAX %g %g\n", *p, d[0]);
#endif
		p--;
		if (*p > d[0])
		    d[0] = *p;
	    }
	    p = &d[0];
#if debug > 1
	    LOG3(ThisModule"MAX %g %g\n", *p, d[0]);
#endif
	    break;
	case 'z':
#if debug >2
	    LOG1(ThisModule"before doing MOD\n");
#endif
#if debug > 1
	    LOG2(ThisModule"%g MOD", *p);
#endif
	    q = p;
	    p--;
	    if (*p == 0) {
		error |= ceDiv0;
#if debug > 1
		LOG2(ThisModule"x=0, error flag set%g\n", *p);
#endif
	    } else {
#if debug > 1
		LOG2(ThisModule"%g", *p);
#endif
		*p = (int) (*q) % (int) (*p);
#if debug > 1
		LOG2(ThisModule"=%g\n", *p);
#endif
	    }
//              abort();
	    break;
	default:
	    LOG2(ThisModule"unknown token %d !!\n", (int) token);
	    abort();
	}
    }
#if debug > 1
    LOG2(ThisModule"final result: %g\n\n", *p);
#endif
    return (*p);
}

/*
    Changes:
    04/25/2002  handling of state flags.
		STATE() function added
    05/13/2002  cout<<... replaced by printf(..
    06/05/2002  TIME() function added.
    
    03/05/2003  fixed debug logging
    07/16/2003  added USeconds system var
    08/18/2004  made DIFF work (again?)
*/
