/*
 Declares a type Manager, a basic object and macros for derived
 objects which registers themselves with the type Manager.

 Part of VISUAL, a human machine interface and data acquisition program

 Most of this code is stolen from the sources of the CommonC++ library.
 
 CommonC++ is Copyright (C) 1999-2001 Open Source Telecom Corporation.
 Daniel Silverstone designed original Common C++ classes (persistence
	engine, math libs, etc...)
 
 The rest: (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#ifndef	__XX_PERSIST_H__
#define	__XX_PERSIST_H__

typedef char int8;
typedef unsigned char uint8;
typedef short int16;
typedef unsigned short uint16;
typedef int int32;
typedef unsigned int uint32;

#ifdef need_long_long
typedef long long int64;
typedef unsigned long long uint64;
#endif

#include <iostream.h>

#include <map>
#include "mytypes.h"

#include "vi_string.h"

using std::map;

class Engine;

// This typedef allows us to declare NewBaseObjectFunction now
typedef class base *(*NewbaseFunction) (void);

/**
 * This class manages the types for generation of the persistent objects.
 * Its data structures are managed automatically by the system. They are
 * implicitly filled by the constructors who declare classes to the system.
 *
 * @author Daniel Silverstone
 * @short Type manager for persistence engine.
 */
class TypeManager {
  public:

	/**
	 * This manages a registration to the typemanager - attempting to
	 * remove problems with the optimisers
	 */
    class Registration {
      public:
	Registration(const char *name,
		     NewbaseFunction func):myName(name) {
	    TypeManager::Add(name, func);
	}
	~Registration() {
	    TypeManager::Remove(myName.c_str());
	}
      private:
	string myName;
    };

	/**
	 * This adds a new construction function to the type manager
	 */
    static void
     Add(const char *name, NewbaseFunction construction);

	/**
	 * And this one removes a type from the managers lists
	 */
    static void
     Remove(const char *name);

	/**
	 * This function creates a new object of the required type and
	 * returns a pointer to it. NULL is returned if we couldn't find 
	 * the type
	 */
    static base *CreateInstanceOf(const char *name);

    typedef
     map < string, NewbaseFunction > StringFunctionMap;
};


/*
 * The following defines are used to declare and define the relevant code
 * to allow a class to use the Persistence::Engine code.
 */

#define DECLARE_PERSISTENCE(ClassType)					\
  public:								\
    friend Engine& operator>>(Engine& ar, ClassType *&ob);		\
    friend Engine& operator<<(Engine& ar, ClassType const *&ob);	\
    friend base *CreateNew##ClassType();				\
    virtual const char* GetPersistenceID() const;			\
    static TypeManager::Registration RegistrationFor##ClassType;

#define IMPLEMENT_PERSISTENCE(ClassType,FullyQualifiedName)		      \
  base *CreateNew##ClassType() { return new ClassType; }		      \
  const char* ClassType::GetPersistenceID()const {return FullyQualifiedName;} \
  Engine& operator>>(Engine& ar, ClassType *&ob)			      \
    { ar >> (base *&) ob; return ar; }				      \
  Engine& operator<<(Engine& ar, ClassType const *ob)			      \
    { ar << (base const *)ob; return ar; }			      \
  TypeManager::Registration						      \
        ClassType::RegistrationFor##ClassType(FullyQualifiedName,	      \
					      CreateNew##ClassType);

/**
 * base
 *
 * This object is the base for all Persistent data which is not
 * natively serialised by the Persistence::Engine
 *
 * It registers itself with the Persistence::TypeManager
 * using a global constructor function. A matching deregister call
 * is made in a global destructor, to allow DLL's to use the
 * Persistence::Engine in a main executable.
 *
 * @author Daniel Silverstone
 * @short Base class for classes that will be persistent.
 */
class base {
  public:
	/**
	 * This returns the ID of the persistent object (Its type)
	 */
    virtual const char *GetPersistenceID() const;

	/**
	 * This constructor is used in serialisation processes.
	 * It is called in CreateNewInstance in order to create
	 * an instance of the class to have Read() called on it.
	 */
    base();

	/**
	 * Default destructor
	 */
    virtual ~ base();

	/**
	 * This method is used to write to the Persistence::Engine
	 * It is not equivalent to the << operator as it writes only the data
	 * and not the object type etc.
	 */
//    virtual bool Write(Engine & archive) const;
    virtual void Write(Engine & archive) const;

	/**
	 * This method is used to read from a Persistence::Engine
	 * It is not equivalent to the >> operator as it does no 
	 * typesafety or anything.
	 */
    virtual void Read(Engine & archive);
};


/**
 * Engine
 *
 * This class constructs on a standard C++ STL stream and then
 * operates in the mode specified. 
 *
 * @author Daniel Silverstone
 * @short stream serialization of persistent classes.
 */

class Engine {
  public:
	/**
	 * These are the modes the Persistence::Engine can work in
	 */
	/**
	 * Constructs a Persistence::Engine with the specified stream in
	 * the given mode. The stream must be initialised properly prior
	 * to this call or problems will ensue.
	 */
    Engine(iostream & stream);

	/**
	 * This Flushes the buffers and closes the Persistence::Engine
	 * this must happen before the underlying stream is shut down
	 */
    ~Engine();


    // Write operations
    void
     Write(const base * object);
    /*
     * shortcut, to make the following more readable
     */
//#define _ENGINEWRITE_REF(valref) WriteBinary((const uint8*)&valref,sizeof(valref))
#define _ENGINEWRITE_REF(valref) myUnderlyingStream<<valref
    void
     Write(int8 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(uint8 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(int16 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(uint16 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(int32 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(uint32 i) {
	_ENGINEWRITE_REF(i);
    }
#ifdef need_long_long    
    void
     Write(int64 i) {
	_ENGINEWRITE_REF(i);
    }
    void
     Write(uint64 i) {
	_ENGINEWRITE_REF(i);
    }
#endif    
    void
     Write(float i) {
	_ENGINEWRITE_REF(i);
    }
//    void
//     Write(double i) {
//	_ENGINEWRITE_REF(i);
//    }
#undef _ENGINEWRITE_REF

    void  Write(const string & str);
    // Every write operation boils down to one or more of theses
    void  WriteBinary(const uint8 * data, const uint32 size);

    // Read Operations

    void  Read(base * &object);
#define _ENGINEREAD_REF(valref) myUnderlyingStream>>valref
    void  Read(int8 & i) {
	_ENGINEREAD_REF(i);
    }
    void
     Read(uint8 & i) {
	_ENGINEREAD_REF(i);
    }
    void Read(int16 & i) {
	_ENGINEREAD_REF(i);
    }
    void Read(uint16 & i) {
	_ENGINEREAD_REF(i);
    }
    void Read(int32 & i) {
	_ENGINEREAD_REF(i);
    }
    void
     Read(uint32 & i) {
	_ENGINEREAD_REF(i);
    }
#ifdef need_long_long        
    void
     Read(int64 & i) {
	_ENGINEREAD_REF(i);
    }
    void
     Read(uint64 & i)
     {
	_ENGINEREAD_REF(i);
    }
#endif    
    void
     Read(float &i)
     {
	_ENGINEREAD_REF(i);
    }
    void
     Read(double &i)
     {
	_ENGINEREAD_REF(i);
    }
#undef _ENGINEREAD_REF

    void
     Read(string & str);
    void
     Write(const double & str); 
    // Every read operation boild down to one or more of these
    void   ReadBinary(uint8 * data, uint32 size);

  public:
	/**
	 * The underlying stream
	 */
    iostream & myUnderlyingStream;

};

// Standard >> and << stream operators for base
Engine & operator >> (Engine & ar,
		      base * &ob);
Engine & operator << (Engine & ar,
		      base const *ob) ;

Engine & operator >> (Engine & ar, int8 & ob) ;
Engine & operator << (Engine & ar, int8 ob) ;

Engine & operator >> (Engine & ar, uint8 & ob);
Engine & operator << (Engine & ar, uint8 ob) ;

Engine & operator >> (Engine & ar, int16 & ob);
Engine & operator << (Engine & ar, int16 ob) ;

Engine & operator >> (Engine & ar, uint16 & ob);
Engine & operator << (Engine & ar, uint16 ob) ;

Engine & operator >> (Engine & ar, int32 & ob);
Engine & operator << (Engine & ar, int32 ob) ;

Engine & operator >> (Engine & ar, uint32 & ob);
Engine & operator << (Engine & ar, uint32 ob) ;

#ifdef need_long_long    
Engine & operator >> (Engine & ar, int64 & ob) ;
Engine & operator << (Engine & ar, int64 ob) ;

Engine & operator >> (Engine & ar, uint64 & ob);
Engine & operator << (Engine & ar, uint64 ob) ;
#endif
Engine & operator >> (Engine & ar, float &ob) ;
Engine & operator << (Engine & ar, float ob) ;

Engine & operator >> (Engine & ar, double &ob);
Engine & operator << (Engine & ar, double ob) ;

Engine & operator >> (Engine & ar, string & ob);
Engine & operator << (Engine & ar, string ob) ;

Engine & operator >> (Engine & ar, bool & ob) ;
Engine & operator << (Engine & ar, bool ob) ;

//Engine& operator <<( Engine& ar, coord ob);

#endif
/*
    Changes:
    
    02/19/2003 	special write method for doubles allows to control precision.
    03/31/2003 	Engine::mode is not needed
*/
