/*
 Implementation of a driver for GE Fanuc PLCs via Ethernet.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2005.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"

#include "vi_string.h"

#define debug 6

#include <stdio.h>

extern "C" {
#define LINUX
#include <openSRTP.h>
}

#include "plclog.h"
#include "GE_Core.h"

char * srtpAreaName(int area) {
    return "something";
}

IMPLEMENT_PERSISTENCE(GEF_TI, "GEFANUC")

siEntry areat[]={
    {"%SA", srtpAreaSAByte},
    {"%SB", srtpAreaSBByte},
    {"%SC", srtpAreaSCByte},
    {"%M", srtpAreaFlagByte},
    {"%T", srtpAreaTimerByte},
    {"%Q", srtpAreaOutputByte},
    {"%I", srtpAreaInputByte},
    {"%G", srtpAreaGlobalByte},
    {"%R", srtpAreaRegisters},
    {"%L", srtpAreaSubblockLocal},
    {"%P", srtpAreaProgramBlockData},
    {"%AI", srtpAreaAnalogInput},
    {"%AQ", srtpAreaAnalogOutput},
};

void GEF_TI::Read(Engine & e)
{
    TFI::Read(e);
#if debug >13   
    show();
#endif    
    dataArea=translateByTable(area.c_str(), areat, sizeof(areat)/sizeof(siEntry));
    if (dataArea==-1)
	LOG1("Warning: data area code must be in %%R, %%M, %%I, %%Q, %%T, %%AI, %%AQ, %%G, %%L, %%P, %%SA, %%SB or %%SC");
    
#if debug >13
    LOG3("Data area      :%d (%s)\n", dataArea, srtpAreaName(dataArea));
    LOG2("Data Block     :%d\n", areaNumber);
    LOG2("first word     :%d\n", startAddr);
    LOG2("Sheet          :%d\n", sheet);
    LOG2("Bytes          :%d\n", byteCount);
#endif
}

    GEFcoreDriver::GEFcoreDriver() {
	firstBitNumber=1;
    };
    uc * GEFcoreDriver::getProtocolData(){
	return (uc*) dc->resultPointer;
    }
    char * GEFcoreDriver::getProtocolErrorMessage(int code) {
	return srtpStrerror(code);
    }

    const char * GEFcoreDriver::getPLCAreaName(TFI * p) {
	return p->area.c_str();
    }    
    
    int GEFcoreDriver::getPLCBitsPerItem(TFI * p) {
	GEF_TI *pa=(GEF_TI *)p;
	return srtpBitsPerItem(pa->dataArea);
    }
    int GEFcoreDriver::getPLCFirstBitNumber() {
	return 1;
    }

void GEFcoreDriver::intSetup(DriverInfo * ddi) {
    PLCdriver::intSetup(ddi);
    srtpSetDebug(libraryDebugLevel);
    rfd = setPort(portParams, &wfd);
    fd.rfd=rfd;
    fd.wfd=wfd;
    LOG3("intSetup write fd: %d libraryDebugLevel:%d\n", wfd, libraryDebugLevel);
    if(wfd>0) {
        di_error=false;
        openCount=0;
    } else {
        LOG2("execute: write fd: %d after connect\n", wfd);
        di_error=true;
    }
    retryCount=0;
    openCount=0;
    dc=srtpNewConnection(fd);
    TypeManager::Add("GEF", CreateNewGEF_TI);
    loadList(DI);
    bitMask = registerDriver("GE Fanuc", this);
    LOG3("intSetup done fd: %d errorMask %x\n", wfd, bitMask);
    FLUSH;
};

bool GEFcoreDriver::openSRTP() {
    int res;
    LOG2("srtpConnect trial %d\n",openCount);
    if (openCount>2) { 
	di_error=true;
	return false;
    }
    res=srtpConnectPLC(dc);
    if (res==0) {
	openCount=0;
        return true;
    } else {
	openCount++;
	return false;
    }
}    


int GEFcoreDriver::protocolRead(TFI * p) {
    GEF_TI *pa=(GEF_TI *)p;
    int bpa=srtpBitsPerItem(pa->dataArea);
    int itemCount=(8*(p->byteCount))/bpa;
    LOG3("one item has:%d bits. Ttem count is:%d\n",bpa, itemCount);
    int res=srtpReadBytes(dc, 
	pa->dataArea, p->startAddr, itemCount, NULL);
    LOG2("res from PLC library read %d\n",res);	
//    res=dc->AnswLen;
    protOK=true;
    if (res<0) protOK=false;
//    if(p->byteCount!=dc->AnswLen) protOK=false;
#if debug >5
    LOG4("protocolRead: need:%d got:%d error:%d\n",p->byteCount, dc->AnswLen, res); 
#endif	    	
    return res;
}

int GEFcoreDriver::protocolWrite(TFI * p, int PLCadr, uc*data, int length) {
    LOG3("protocolWrite(dat:%p, len:%d)\n",data,length);
    GEF_TI *pa=(GEF_TI *)p;
    int bpa=srtpBitsPerItem(pa->dataArea);
    int itemCount=8*length/bpa;
    LOG3("one item has:%d bits. Ttem count is:%d\n",bpa, itemCount);
    int res=srtpWriteBytes(dc, pa->dataArea, PLCadr, itemCount, data);
    LOG2("res from PLC library write %d\n",res);
//    if (res>0) res=0;	// this version of openSRTP returns <0 for errors;
//    else {
//	protOK=false;
//    }
    return res;
}

void GEFcoreDriver::invalidateConnection() {
    protOK=false;
    LOG1("Hello from invalidateConnection!\n");
}


