/*
 Implementation of a skeleton PLC driver. Real drivers will overwrite most
 or some methods.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#define debug 0

#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "PLCdriver.h"
#include "lib0.h"
#include "cellList.h"
#include "persist.h"
#include <fstream>
#include <stdio.h>
#include <unistd.h>
#include "plclog.h"
#include "byteswap.h"
#include "lilo.h"
#include "debuglevels.h"

using std::fstream;
/*
    Definition of a transport instruction for all PLC drivers.
    Each instruction describes a block of data in PLC which is copied
    to or from a column of the spreadsheet beginning at coordinates
    startCell. Transport instructions are held in a list, which is 
    cyclicly executed by the driver. Conditional execution may be 
    achieved using controlCell. PLC drivers may extend this class.
*/

TFI::TFI() {};

//IMPLEMENT_PERSISTENCE(transferInstruction, "TRANSFERINST")

void TFI::Read(Engine & e)
{
    string stc, ctc, erc;		// temporary strings for coordinates
    e >> comment >> ctc >> erc >> 
    deviceAddress >> subAddress >> area >> areaNumber >> startAddr >> method >>
    sheet >> stc >> format;
    startCell = getPos(stc);
    ctrlCell = getPos(ctc);
    errorCell = getPos(erc);
    expandFormat();
    byteCount=formatBytes();	// normally
    isWrite=(method=="w")||(method=="W");	
    // For most drivers ok. This is not quite OOP conform: If you need another way to tell
    // read from write instructions, overwrite the result in your Read method. See beckhoff
    // or modbus drivers for example.
};

void TFI::expandFormat() {
#if debug>8
    LOG2("enter expandFormat %s\n",format.c_str());
#endif    
    string f2="";
    char * lChar;
    const char * cform = format.c_str();
    while (*cform!=0) {
#if debug>9    
	LOG2("expandFormat char:%c\n",*cform);
#endif	
	if (('0'<=*cform) && (*cform<='9')) {
	    int rep=strtol(cform, &lChar, 0);
#if debug>9	    
	    LOG3("rep:%d lastChar:%p\n", rep,lChar);
#endif	    
	    cform=lChar;
	    for (int j=0;j<rep-1;j++) f2+=*cform; 
	    // 1.char less, specifier is added in else part	
	} else {
	    f2+=*cform;
	    cform++;    
	}
#if debug>9	
    	LOG2("expandFormat part:%s\n",f2.c_str());
#endif	
    }
#if debug>8    
    LOG2(" = %s\n",f2.c_str());
#endif    
    format=f2;
}

void TFI::show() {    
    LOG2("comment        :%s\n", comment.c_str());
    LOG2("Errors to Cell :%s\n", cellN(errorCell));
    LOG2("Condition      :%s\n", cellN(ctrlCell));
    LOG1("Address of and within PLC:\n");
    LOG2("Device address :%s\n", deviceAddress.c_str());
    LOG2("Sub address	 :%s\n", subAddress.c_str());
    LOG2("Data area      :%s\n", area.c_str());
    LOG2("Area number    :%d\n", areaNumber);
    LOG2("first word     :%d\n", startAddr);
    LOG2("method         :%s\n", method.c_str());
    LOG1("Address in Spreadsheet:\n");
    LOG2("Sheet          :%d\n", sheet);
    LOG2("to Cell        :%s\n", cellN(startCell));
    LOG2("Format         :%s\n", format.c_str());
}

int TFI::formatBytes() {
    int j=0,bits=0;
    for (unsigned int i=0;i<format.length();i++) {
	char c=format[i]; //char c=f.at(i);
	switch(c) {
	    case 'x':
	    case 'X': bits+=1; break;
	    case 'c':
	    case 'C':
	    case 'b':
	    case 'B': j+=1; break;
	    case 'i':
	    case 'I': 
	    case 'w':
	    case 'W': j+=2; break;
	    case 'f':
	    case 'F':
	    case 'R': 
	    case 'r':
	    case 'u':
	    case 'U':
	    case 'l':
	    case 'Z':
	    case 'L': j+=4; break;
	    case 'd': 
	    case 'D': j+=8; break;
	}	    
    }
    j=j+(bits +7)/8;
    return j;	
}

bool TFI::isDelay(){
    return false;
}


const char * PLCdriver::getPLCAreaName(TFI * p) {
    return "DB";
}
int PLCdriver::getPLCBitsPerItem(TFI * p) {
    return 8;
}

char * PLCdriver::getProtocolErrorMessage(int code) {
    return "your driver should override getProtocolErrorMessage!";
}


PLCdriver::PLCdriver() {
    pid=getpid();
    execCount=0;
    consecutiveErrors=0;
    driverDebugLevel=0;		// the debug level for driver code
    libraryDebugLevel=0;	// the debug level for an external communication library
    firstBitNumber=0;		// the number of the lowest bit. Standard is 0.

    doRun=true;
    DI=NULL;
    TypeManager::Add("Delay", CreateNewDLY);	// Make this instruction known to all drivers.
#if debug>8
    LOG1("TypeManager::Add(Delay, CreateNewDLY) done.\n");
#endif    
}

/*
    load the instruction list:
*/
void PLCdriver::loadList(DriverInfo * i) {
    LOG2("load from %s\n", i->s1.c_str());
    FLUSH;
    handle = i->th;
    il=loadlist((i->s1).c_str());
}

/*
    Execute a single instruction:
*/
int PLCdriver::execTI(TFI * p) {
//#if debug >0
    LOG2("execTI %s\n", p->comment.c_str());
    FLUSH;
//#endif    
    coord cell;
    int flags;
    bool ok;
/*
    Test whether the condition for conditional execution is true:
*/    
    double d=numContents(p->sheet, p->ctrlCell, ok);
    if (!ok) {
	LOG3("execTI: cell for conditon (%d:%s) not found!\n", 
	    p->sheet, cellN(p->ctrlCell)); // hint in LOG, if it is a config error...
	return plcExecOk;   // ...but return ok
    }
    if (d==0) {
	LOG1("execTI: conditon 0, nothing to do.\n"); 
	return plcExecOk;
    }
/*
    If it is a delay instructions, just do the delay and exit:
*/    
    if (p->isDelay()) {
	usleep( ((DLY*)p)->msecs);
	return plcExecOk;
    } 	    
//    LOG1("execTI after delay\n");
/*
    Is the file descriptor still valid?
*/    
    if(wfd==failed_fd) {
	LOG2("execTI invalid fd %d\n", wfd);
	FLUSH;
	usleep(100000);
	return plcExecFileClosed;
    }	
/*
    Here starts normal operation
*/    
    cell=p->startCell;    
    if (p->isWrite) {
	int PLCadr=p->startAddr;
	int bitCount=0;
	ui bitContainer=0;	// used to agglomerate bits up to one byte
				
	int fpos=0;			// current position in format string
	int formatLen=0;		// number of valid bytes in wbuffer, 
					// resulting from conversion
	bool doWrite;			// we must send changed value
	bool doWriteBits=false;		// we must send the byte conteinig changed bits
	union {
	    uc ucv;
	    char cv;
	    short sv;
	    us usv;
	    int iv;
	    ui uiv;
	    float f;
	    us Z[2];
	    uc b[4];
	} data;
	int i=0;			// counting bytes
	while (i<p->byteCount) {
	    char specifier=p->format[fpos];
#if debug >0	    
	    LOG4("check for write i:%d, specifier:%c cell:%s\n",i,specifier,cellN(cell));
#endif	    
	    doWrite=false;		// assume: "no need to really write"
	    
	    flags = getFlags(p->sheet, cell, ok);// test whether value has changed						
	    if (ok) {				// silently ignore missing cells
						// an application may leave
						// gaps in a block.
	        if (flags & bitMask) {
		    doWrite=true;		// we must write a new value
		    d=numContents(p->sheet, cell, ok);
		    /* convert the value of d to bytes according to format */
		}	
	    }
	    char ic[10];
	    sprintf(ic,"%d",PLCadr);
	    string ss=ic;
	    string ss1=getPLCAreaName(p);
	    int bitsPerItem = getPLCBitsPerItem(p);
	    switch(specifier) {
		case 'x':
		    if (d!=0.0) bitContainer|=(0x80>>bitCount);
		    goto continueWithBits;
	        case 'X': 
		    if (d!=0.0) bitContainer|=(1<<bitCount);
continueWithBits:		    
		    formatLen=0;
		    ss="DBX"+ss+".";
		    sprintf(ic,"%d", bitCount+firstBitNumber);
//		    ss=ss+(char)('0'+bitCount); 
		    ss+=ic;
		    if (doWrite) doWriteBits=true;
		    bitCount+=1;
		    if (bitCount==bitsPerItem) {
			formatLen=1;
			bitCount=0;
			data.ucv=bitContainer;
			bitContainer=0;
			doWrite=doWriteBits; // write byte if at least one bit changed
			doWriteBits=false;
		    }	
		    break;
		case 'b':    
		case 'B':
		    data.ucv=(ui)d; formatLen=1; ss=ss1+"B"+ss; break;
		case 'c':
		case 'C':
		    data.cv=(int)d; formatLen=1; ss=ss1+"B"+ss; break;
		case 'i':    
		    data.sv=(int)d;
		    formatLen=2; ss=ss1+"W"+ss; break;    
		case 'I':    
		    data.sv=(int)d; 
		    data.sv=bswap_16(data.sv);
		    formatLen=2; ss=ss1+"W"+ss; break;
		case 'w':    
		    data.usv=(ui)d;
		    formatLen=2; ss=ss1+"I"+ss; break;    
		case 'W':    
		    data.usv=(ui)d; 
		    data.usv=bswap_16(data.usv);
		    formatLen=2; ss=ss1+"I"+ss; break;    
		case 'l':    
		    data.iv=(int)d;
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'L':    
		    data.iv=(int)d;
		    data.iv=bswap_32(data.iv);
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'u':    
		    data.uiv=(ui)d;
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'U':    
		    data.uiv=(ui)d;
		    data.uiv=bswap_32(data.uiv);
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'f':
		case 'r':
		    data.f=d;
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'F':
		case 'R':    
		    data.f=d;
		    data.iv=bswap_32(data.iv);
		    formatLen=4; ss=ss1+"D"+ss; break;
		case 'Z':    
		    data.f=d;
		    us fw=data.Z[0];
		    data.Z[0]=data.Z[1];
		    data.Z[1]=fw;
		    formatLen=4; ss=ss1+"D"+ss; break;    
	    }
	    setSource(p->sheet,cell,ss);
	    if (doWrite) {
		if (formatLen) {
		    LOG2("calling protocolWrite for cell:%s\n",cellN(cell));
		    LOG5("data:%02X %02X %02X %02X\n",data.b[0],data.b[1], data.b[2],data.b[3]);  
		    int res=protocolWrite(p, PLCadr, &data.ucv, formatLen);
		    if (0==res) {	
			clrPushButton(p->sheet, cell);
		    } else {
			LOG2("got error code %d from protocolWrite\n",res);
			FLUSH;
//			setCell(p->sheet, p->errorCell, res, soIoState);
			return plcExecWriteError;
		    }
		}  else {
		    LOG2("fornatLen=0, specifier:%c\n",specifier);
		}
	    }
	    PLCadr+=8*formatLen/bitsPerItem;
	    cell.row++;	// next row in spreadsheet
	    fpos++;	// next letter in format
	    i+=formatLen;
	}
//	setCell(p->sheet, p->errorCell, 0, soIoState);
    } else {
/*
	Read Instructions:
*/    
#if debug>0    
	LOG1("execTI read.\n");
	FLUSH;
#endif	
	int res=protocolRead(p);
	if (-1111==res) {
/*
	    Just a kludge to allow very special drivers to do special handling of data storage
	    without overwriting this long method...
*/	    
	    return plcExecOk;
	}
	if (0==res) {
#if debug>0	
	    LOG1("after protocolRead().\n");
	    FLUSH;
#endif	    
	    uc * b= getProtocolData();
	    if (b!=NULL) {
		storeByFormat(b, p, 1, 0);
#if debug>0			
		LOG1("after storeByFormat().\n");
		FLUSH;
#endif		
	    } else {
		LOG1("got NULL pointer to data.\n");
		FLUSH;
		goto readError;
	    }		
//	    setCell(p->sheet,p->errCell,0,soExternal); // can do this in a central point
///	    setCell(p->sheet, p->errorCell, 0, soIoState);
	    return plcExecOk;	
	} else {
readError:	
	    LOG3("got error code %d from protocolRead %s\n",res, getProtocolErrorMessage(res));
	    FLUSH;
	    markCellErrorStates(p);
///	    setCell(p->sheet, p->errorCell, res, soIoState);
	    return plcExecReadError;
	} 
    }
    return plcExecOk;
}

/*
    store the result of a read into spreadshead according to format specifiers:
*/
void PLCdriver::storeByFormat(uc * byteArray, TFI * ti, int rowIncrement, int colIncrement) {
//#if debug >1
    if(driverDebugLevel & debugControlFlow) {
	hexDump("enter storeByFormat() raw Bytes:", byteArray,ti->byteCount);
	FLUSH;
    }
//#endif    
    int i=0;
    int fp=0;
    int bits=0;
    int bitNumber=0;
    double d;
    coord cell=ti->startCell; 
    unsigned char bitContainer;
    union {
	float a;
	int b;
	unsigned int c;
    } f;
    string ss1=getPLCAreaName(ti);
    int bitsPerItem = getPLCBitsPerItem(ti);
    while (i<ti->byteCount) {
	char fs=ti->format[fp]; 	//char fs=pa->format.at(fp);
#if debug >6
	LOG3("using format: %c for address %d\n",fs,i);
#endif		
	char ic[10];
	sprintf(ic,"%d", (8*i/bitsPerItem)+ti->startAddr);
	string ss=ic;
	switch(fs) {
	    case 'x':
    	    case 'X':
		if (bits==0) {
		    bitContainer=*byteArray; byteArray++;
		}	
		if((bitContainer & (1 << bits))!=0) {
		    d=1.0;
		} else {
		    d=0.0;
		}
		ss=ss1+"X"+ss+".";
		sprintf(ic,"%d", bitNumber+firstBitNumber);
		//ss=ss+(char)('0'+bits); 
		ss=ss+=ic;
	    bits++;
	    bitNumber++;
	    if (bitNumber==bitsPerItem) {
		bitNumber=0;
	    }
	    if (bits==8) {
	        i+=1;
		bits=0;
	    }	
	break;
	    case 'b': 
	    case 'B': d=*byteArray; byteArray++; i+=1; ss=ss1+"B"+ss;  break;
	    case 'c': 
	    case 'C': d=*(char*)byteArray; byteArray++; i+=1; ss=ss1+"B"+ss;  break;

	    case 'i': d=*(short*)byteArray; byteArray+=2; i+=2; ss=ss1+"W"+ss;break;
	    case 'I': d=bswap_16(*((short*)byteArray)); byteArray+=2; i+=2; ss=ss1+"W"+ss;break;
	    case 'w': d=*(unsigned short*)byteArray; byteArray+=2; i+=2; ss=ss1+"W"+ss;break;
	    case 'W': d=bswap_16(*((unsigned short*)byteArray)); byteArray+=2; i+=2; ss=ss1+"W"+ss;break;
	    case 'F':
	    case 'R': f.b=bswap_32(* (int*)( (int*)byteArray));
		      d=f.a;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    case 'f':
	    case 'r': f.b=(* (int*)( (int*)byteArray));
		      d=f.a;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    break;	    
	    case 'L': f.b=bswap_32(* (int*)( (int*)byteArray));
		      d=f.b;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    break;
	    case 'l': f.b=(* (int*)( (int*)byteArray));
		      d=f.b;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    break;
	    case 'U': f.b=bswap_32(* (int*)( (int*)byteArray));
		      d=f.c;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    break;
	    case 'u': f.b=(* (int*)( (int*)byteArray));
		      d=f.c;
		      byteArray+=4; i+=4; ss=ss1+"D"+ss;break;
	    break;
	    default: LOG2("unimplemented format specifier: %c\n",fs);
		      i++;	
	}
	setCell(ti->sheet,cell,d,soExternal);
	setSource(ti->sheet,cell,ss);
#if debug >5				
	LOG4("I set cell %s = %g from %s.\n",cellN(cell),d,ss.c_str());
#endif		
	cell.row+=rowIncrement;
	cell.col+=colIncrement;
	fp++;
    }
} 

int PLCdriver::execList()
{
//#if debug >0
    if(driverDebugLevel & debugControlFlow) {
	LOG2("enter exec list @%p\n",il);
	if(il)LOG2("%d items: \n",il->count);
	FLUSH;
    }
//#endif
//    il->forEach(&execTI1, (void *) this);
    for (int i=0;i<il->count;i++) {
	TFI * p=(TFI*)il->items[i];
	if(p) {
	    int res=execTI(p);
	    setCell(p->sheet, p->errorCell, res, soIoState);
//	    if (res!=0) return res;
	}    
    }	
    
    if(DI){
	LOG2("execute: looking for commands %d.\n",PLCdriver::DI->commands);
	if(DI->commands & 1) {
	    LOG2("Ooops, commands %d, beginning to reload.\n",DI->commands);
	    unRegisterDriver(bitMask);
	    if (rfd>0){	
	        if ((wfd!=rfd) && (wfd>0)) close(wfd);
	        close(rfd);
	    }	
	    DI->commands=0;
	    intSetup(DI);
	    LOG2("Reload done. Commands: %d\n",DI->commands=0);
	} else if(DI->commands & 2) {
	    LOG2("Ooops, commands %d, beginning to reconnect.\n",DI->commands);
	    DI->commands=0;
	    invalidateConnection();
	    LOG1("Connection closed.\n");
	}
    } else {
        LOG1("Ooops, thought I kept my DI!\n");
    }
    return 0;
}

/**
    After a successful write to PLC, reset the "changed" flag of the source 
    cell. If source cell is coupled to a pushbutton in an hmi page, we set
    the cell to zero, simulating the "springing back" of the button. 
    In setting it with setCell, it is marked as "changedAll" again, so we 
    will handle it in the next cycle again and send the 0. We mark the data
    source type as "soPushBAck" (ackknowledge for pushbutton, so next time we 
    only reset the flag.
*/
void PLCdriver::clrPushButton(int sheet, coord & acell)
{
//#if debug >10
    if(driverDebugLevel & debugControlFlow) {
	LOG2("clrPushButton sheet %d\n",sheet);
	FLUSH;
    }
//#endif    
    if (sheets[sheet]!=NULL) {
#if debug >0
	LOG3("clrPushButton cell %d:%s\n",sheet,cellN(acell));
	FLUSH;
#endif    
	cell * c=sheets[sheet]->find(acell);
	if (c!=NULL) {
	    c->setFlags(c->getFlags() & ~bitMask);
	    if (c->getSourceType()==soPushButton) {
		setCell(sheet, acell, 0, soPushBAck);
	    }  
	}
    }
}

/*
    Internal setup. Does all preparations necessary to run the driver.
*/
void PLCdriver::intSetup(DriverInfo * i) {
    DI=i;
    string params=DI->s,name,value,s1;
    LOG2("intSetup: got setup information:%s\n",params.c_str());
    string driverParameters=getToken(params);
    LOG2("driver parameters: %s\n", driverParameters.c_str());
    while (s1!=driverParameters) {    
	s1=getToken2(driverParameters);
	LOG2("driver parameter: %s\n", s1.c_str());
	int p = s1.Find("=");
        if(p != 0) {
	    name = s1.before(p);
	    value = s1.after(p);
	}    
	LOG3("name: %s value: %s\n", name.c_str(),value.c_str());
	parseNameValuePair(name, value);
    }
    portParams=params;
}

/*
    Write a block of data to the exteranal device.
*/
int PLCdriver::protocolWrite(TFI * p, int PLCadr, uc * data, int len) {
    LOG1("Hello from protocolWrite(). Overwrite me!\n");
    return -1;
}

/*
    Read a block of data from the exteranal device.
*/
int PLCdriver::protocolRead(TFI * p) {
    LOG1("Hello from protocolRead(). Overwrite me!\n");
    return -1;
}

void PLCdriver::markCellErrorStates(TFI * p) {
    LOG1("enter markCellErrorStates().\n");
    FLUSH;
    coord cell=p->startCell; 
    for (ui i=0;i<p->format.length(); i++) {
	setState(p->sheet, cell, bitMask, true);
#if debug >5				
	LOG2("I mark cell %s\n",cellN(cell));
#endif		
	cell.row++;
    }
}

uc * PLCdriver::getProtocolData() {
    LOG1("Hello from getProtocolData(). Overwrite me!\n");
    return NULL;
}

void PLCdriver::hexDump(char * text,uc * buffer, int len) {
	int i;
	LOG2("%s",text);
	for (i = 0; i < len; i++) {
	    LOG_2("%x,", buffer[i]);
	    if ((i % 16) == 15)
	    LOG_1("\n");
	}
    }					

void PLCdriver::shutdown() {
    LOG1("Hello from shutdown!\n");
}

void PLCdriver::invalidateConnection() {
    LOG1("Hello from invalidateConnection!\n");
}


IMPLEMENT_PERSISTENCE(DLY, "Delay")

void DLY::Read(Engine & e)
{
    string ctc;		// temporary strings for coordinates
    LOG1("Hello from DLY::Read\n");
    FLUSH;
    e >> comment >>sheet >> ctc >> msecs;
    ctrlCell = getPos(ctc);
    isWrite=false;
};

bool DLY::isDelay(){
    return true;
}


string PLCdriver::getToken2(string &params)
{
    int p = 0;
    do {
	p = params.Find(',');
	if (p == 0)
	    params.erase(0, 1);
    } while (p == 0);
    string s = params.before(p);
    params.erase(0, p + 1);
    LOG3("%s rest %s\n", s.c_str(), params.c_str());
    return s;
}

void PLCdriver::parseNameValuePair(string name,string value) {
    if (name=="debug") driverDebugLevel = atol(value.c_str()); else
    if (name=="libDebug") libraryDebugLevel = atol(value.c_str()); else
    LOG1("Hello from parseNameValuePair(). Overwrite me!\n");
}
/*
    Changes:
    09/26/2003 	implemented new basic PLC driver.
    09/29/2003 	fixed bug in expandFormat().
    09/29/2003 	Mark corresponding cells also for write instructions. 
		fixme: The 'DBx' notation is only appropriate for Siemens. 
		fixme: Source name is updated every cycle. Too much work?
    09/29/2003 	fixed bug in bit writing.
    09/29/2003 	fixed bug in bit writing.
    01/23/2004 	clear DI, show DI-> commands if nonzero
    03/18/2004 	fixed bug in expandFormat
    06/23/2004 	added hexDump()
    08/17/2004 	fixed numbering of source addresses from startAddr
    08/17/2004 	changed definition of storeByFormat to take row and column increment steps.
    03/11/2005	added more comments.
*/
