/*
 Definitions of a skeleton PLC driver and the instructions it can process.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001..2003

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#ifndef __PLCdriver
#define __PLCdriver

#include <sys/time.h>
//#include "base.h"
#include "persist.h"
#include "mytypes.h"
#include "list.h"
#include "persist.h"
#include "vi_string.h"

#include "setport.h"

#define plcExecFileClosed 1
#define plcExecOk 0
#define plcExecWriteError 2
#define plcExecReadError 4

struct driverError {
    time_t time;
    int level1;
    int level2;
};

/*
    Basic operation of a PLC driver:
    Get loaded by loader.
    Get configuration information from loader. (Coming from visual.cfg)
    Load the tfi-list.
    Prepare tfis.
    Open a pair of file descrptors for I/O 1)
init:    
    Initialize logical connection.
    Loop:
	Execute first tfi.
	    Execute transfer
	    Store result in <b>error cell</b>
	    Ok?
	    Yes: Write instructions reset "changed"	flags on spreadsheet data.
	    No:	1. Set error state on target data
		2. Now, if we have a logical connection, does the error mean
		   it's disrupted?
		Yes: 
		    cleanupConnection; 
		    goto init:
		No:    
	Execute next tfi;
	    ....
	Execute last tfi;
	check for commands, e.g. reload
	goto loop;		    
    
*/

/*
    Definition of a transport instruction for all PLC drivers.
    One instruction describes a block of data in PLC which is copied
    to or from a column of the spreadsheet beginning at coordinate
    startCell. 
    Transport instructions are held in a list, which is executed round 
    and round by the driver. Conditional execution can be achieved using 
    <condition>.
    
    Instruction syntax:
    <comment>		Some comment to remember what an instruction does.
    <condition>		A cell coordinate that must exist, be numerical and 
			and nonzero in order to execute this instruction.
    <error cell>	A cell that will take error codes.
    
    The PLC-side of an instruction defines the data source as:
    <device address>	Address of the physical device,
			IP address when TCP/IP used or MAC address 
			for "RAW" ethernet.
    <subdevice>		reserved for future use, in the hope I shall not
			have to redo this.
    <area>		A code for the memory area in the PLC
			e.g. Siemens: Data block, flags, inputs, outputs, etc.
    			e.g. GE Fanuc: Markers, Registers, inputs, outputs, etc.
			e.g. Allen Bradley: File type code
    <area number>	Numeric id of multiple similar areas
			e.g. Siemens: Data block number
    			e.g. GE Fanuc: not applicable. (at least in 9030)
			e.g. Allen Bradley: File number
    <start>		Start address (offset) of PLC data in mermory area.
    
    <method>		'R' for read or 'W' for write.
			For some PLC types, this will take up special codes 
			for the transport. 
			e.g. Allen Bradley: The command and function code to
			use. A very "userfriendly" driver could determine this 
			from a model code and a database, but who wants to
			keep this up to date? Some VERY old DOS programs had
			a list of printer control codes to adapt them to your 
			printer. I should very much like this approach instead
			of googling for the most recent drivers and repeating 
			it until I can live with the most recent bugs...
			
    VISUALS side of an instruction defines the corresponding spreadsheet area:
    <sheet>		The number of the sheet to use.
    <startCell>		The cell where the first data item of this Block goes 
			to. Next item goes to next row.
    <format>		A sequence of letters that let the driver extract a
			sequence of bytes as a corresponding data type.
*/

/* 
    The format specifier string allows to read mixed type data from a byte 
    array. It consists of the following characters that extract one Variable 
    of <length> byte and put it to a cell. Cell row is incremented by one. For
    multi-byte types, a lower case letter means little endian (Intel) byte 
    order, while upper case letters mean big endian (Motorola).
   
    character	mnemonic	gcc C-type			length (byte)
    X,x		bit		bit				1/8	
    c,C		char		unsigned char			1
    b,B		byte		signed char			1
    i		integer		signed short, little endian	2
    I		integer		signed short, big endian	2
    w		word		unsigned short, little endian	2
    W		word		unsigned short, big endian	2
    l		longint		signed int, little endian	4
    L		longint		signed int, big endian		4
    u		unsigned long	unsigned int, little endian	4
    U		unsigned long	unsigned int, big endian	4
    w		word		unsigned short, little endian	2
    W		word		unsigned short, big endian	2
    f		float		float, little endian		4
    F		float		float, big endian		4
    r		real		float, little endian		4	same as f
    R		real		float, big endian		4	same as F
    d		double		double, little endian		8
    D		double		double, big endian		8
    
    The syntax <number><specifier> replaces a sequence of equal specifiers.
    '6F5I4b' is equivalent to 'FFFFFFIIIIIbbbb'
    The former syntax that defined a single type for all data elements of a block
    can be replaced by a single combination of number and specifier.
*/

/*	
    Example: 'BWwCLFf', target cell B2 will extract the first 18 bytes from a byte array ba[]
    and store it to cells:
    
    ba[0]->B2			signed
    256*ba[1]+ba[2]->B3		unsigned
    ba[3]+256*ba[4]->B4		unsigned
    ba[5]->B5			unsigned
    16777216*ba[6]+65536*ba[7]+256*ba[8]+ba[9]->B6	signed
    cast (ba[13],ba[12],ba[11]+ba[10]) to float->B7	signed    
    cast (ba[14],ba[15],ba[16]+ba[17]) to float->B8	signed    
*/

/*
    The transfer instruction. This is the binary representation of a single row
    in the tfi-list.
*/

class TFI: public base {
public:
    string comment;
    coord ctrlCell;	// <condition>
    coord errorCell;	// <error cell>	
    string deviceAddress;
    string subAddress;
    string area;
    int areaNumber;
    ui startAddr;
    string method;
    ui sheet;
    coord startCell;
    string format;
    bool isWrite;	// true if this is a write instruction
    int byteCount;	// the number of bytes we need when we concatenate
			// the binary representation of the data specified 
			// by format. (Hope this is a valid definition even when
			// we deal with ASCII data.
    TFI();
    virtual void Read(Engine & e);
    virtual void show();
    virtual bool isDelay();
    int formatBytes();
    void expandFormat();	// translates multiplier notation ('6F') to explicit notation ('FFFFFF').
};
/*
    A special transfer instruction. This will delay execution for msecs millisecods. If you have fast drivers
    (normally on ethernet connections), insert a delay instruction into the list.
*/
class DLY: public TFI {
public:
    int msecs;			// delay in mikroseconds
    virtual bool isDelay();
    virtual void Read(Engine & e);
    DECLARE_PERSISTENCE(DLY)
};

#include "driverinfo.h"
/*
//
// what the loader passes to a PLC driver:
//
struct DriverInfo {
    pthread_t th;
    string s1;
    string s;
};
*/

//
// PLC driver prototype:
//
class PLCdriver:public base {
  public:
    DriverInfo * DI;
    int rfd;			// a file Descriptor to read from
    int wfd;			// a file Descriptor to write to
    int bitMask;		// bit mask to mark errors as originating from this driver
				// bit mask to extract drivers "new value flag"
				// the same mask can be used for both
    int firstBitNumber;		// the number of the lowest bit, normally 0, but e.g. GE starts with 1				
    pthread_t handle;		// the handle from libpthread
    int	driverDebugLevel;	// the debug level for driver code
    int	libraryDebugLevel;	// the debug level for an external communication library
public:    
    int pid;			// save, we need it to recognize a particular instance of the driver
    int execCount;		// counts executions of the tfi list
    int consecutiveErrors;
    bool doRun;			// used to end driver execution from another thread
    bool isRunning;		// used to see whether driver is still executing
    string portParams;		// the part of parameter string from .cfg used to set up a connection
    PLCdriver();		
    list *il;			// list of transfer instructions
    
    virtual void loadList(DriverInfo *);	// Get loaded by loader.
    
    virtual int execList();
    virtual int execTI(TFI * p);
    void clrPushButton(int sheet, coord & cell);
    virtual void intSetup(DriverInfo * i);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data, int len);
    virtual int protocolRead(TFI * p);
    virtual void markCellErrorStates(TFI * p);
    virtual void storeByFormat(uc * byteArray, TFI * ti, int rowIncrement, int colIncrement);
    virtual uc * getProtocolData();	// pointer to an array of bytes. Either the protocol 
					// library or the driver must provide it.
    virtual void hexDump(char * text,uc * buffer, int len);
    virtual const char * getPLCAreaName(TFI * p);
    virtual int getPLCBitsPerItem(TFI * p);
    virtual char * getProtocolErrorMessage(int code);
    virtual void shutdown();
    virtual void invalidateConnection();
    virtual string getToken2(string &params);
    virtual void parseNameValuePair(string name,string value);
/*    
    Basic operation of a PLC driver:
    Get loaded by loader.
    Get configuration information from loader. (Coming from visual.cfg)
    Load the tfi-list.
    Prepare tfis.
    Open a pair of file descrptors for I/O 1)
init:    
    Initialize logical connection.
    Loop:
	Execute first tfi.
	    Execute transfer
	    Store result in <b>error cell</b>
	    Ok?
	    Yes: Write instructions reset "changed"	flags on spreadsheet data.
	    No:	1. Set error state on target data
		2. Now, if we have a logical connection, does the error mean
		   it's disrupted?
		Yes: 
		    cleanupConnection; 
		    goto init:
		No:    
	Execute next tfi;
	    ....
	Execute last tfi;
	check for commands, e.g. reload
	goto loop;
*/	
    
};

#endif
/*
    Changes:
    
    05/13/2002  Object Read & write functions return void now. 
    07/25/2002  Mechanism to reload on command (normally given by loader on SIGHUP).
    06/23/2004 	added hexDump()
    08/17/2004 	changed definition of storeByFormat to take row and column increment steps.
*/
