/*
 Implementation of a serial line driver for Allen Bradley PLCs

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include "plclog.h"

#include "vi_string.h"
#define debug 10

#include <stdio.h>
#include "AB_lib.h"
#define ThisModule "Allen_B  : "

#define tmo_normal 331000
#include "allen_b.h"

#define want_write    
/*
    I do not yet know, how Allen Bradley's addressing types and transfer modes
    fit into a closed concept.
    So, to minimize the chaos, we define directly the CMD and FNC and
    other codes here, leaving blank many "tradional" fields of the transfer
    instruction.
*/

class AllenBradleyTI:public TFI {
  public:
    int CMD;			//      Allen B's CMD code
    int FNC;			//      Allen B's FNC code
    int AB_type;		//      Allen B's data type code
    DECLARE_PERSISTENCE(AllenBradleyTI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(AllenBradleyTI, "ALLENB")

void AllenBradleyTI::Read(Engine & e)
{
/*
    We define directly the CMD and FNC codes in method field.
*/
    TFI::Read(e);
    FNC=0;
    char *sep;
    CMD=strtol(method.c_str(), &sep ,0);
    if (*sep!=0) {
	sep++;
	FNC=strtol(sep, &sep ,0);
    }	
    
    AB_type=0x89;  // integer;
    if ((CMD == 0xF) && (FNC == 0xA2)) isWrite=false; else // protected typed read with 3 byte address
    if ((CMD == 0x6) && (FNC == 0x03)) isWrite=false; else // diagnostic status
    if ((CMD == 0x1) && (FNC == 0x00)) isWrite=false; else // unprotected read from CIF file, command has no FNC
    if ((CMD == 0xF) && (FNC == 0xAA)) isWrite=true;  else // protected typed write with 3 byte address
    if ((CMD == 0x8) && (FNC == 0x00)) isWrite=true;  else // unprotected write to CIF file
    LOG1("** Warning: unsupprted method. Supported methods are (15,162),(1,0),(6,3) for read and (15,170) or (8,0) for write.\n");
#if debug >3
    show();
    LOG2("CMD  	      	 :%0X\n", CMD);
    LOG2("FNC  	 	 :%0X\n", FNC);
    LOG2("AB's type      :%0X\n", AB_type);
    LOG2("AB's file no.  :%d\n", areaNumber);
    LOG2("length in bytes:%d\n", byteCount);
#endif
}

class AllenBradleyDriver:public AllenBradley {
  public:
    AllenBradleyTI *pa;
    AllenBradleyDriver() {};
    virtual int execTI44(TFI * p);
    unsigned short master_send(AB_block *, int size);
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    execList();
	    execCount++;
	}
    };
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    virtual uc * getProtocolData(){
	return result;
    }
    uc * result;
};

void AllenBradleyDriver::intSetup(DriverInfo * i)
{
    rfd = setPort(i->s, &wfd);
    LOG2("intSetup fd: %d\n", wfd);
    TypeManager::Add("ALLENB", CreateNewAllenBradleyTI);
    loadList(i);
    bitMask = registerDriver("AllenBradley", this);
    TNS = 97 * 256;
};

unsigned short AllenBradleyDriver::master_send(AB_block * ab, int size)
{
    ab->loTNS = TNS % 256;
//    ab->hiTNS=TNS/256;
    TNS++;
    return ab_send(ab, size);    
}


int AllenBradleyDriver::protocolRead(TFI * p) {
    us rcv_tns, sentTNS;
    unsigned char b1[1024];
    unsigned char *bs;
    int errorState = 0, res;
    AB_block *ab;
    pa = (AllenBradleyTI *) p;
    
    if ((pa->CMD == 0xF) && (pa->FNC == 0xA2)) {	// it's that sort...
#if debug >2
	LOG1("execTI protected typed read with 3 byte address\n");
#endif
	AB_block abb;
	memset(&abb, 0, sizeof(abb));
	abb.CMD = pa->CMD;
	abb.FNC = pa->FNC;
	abb.byteSize = p->byteCount;
	abb.fileNo = p->areaNumber;
	abb.fileType = pa->AB_type;
	abb.eleNo = 0xFF;
	abb.XeleNo = p->startAddr;
	sentTNS = master_send(&abb, 14);
    }	
    
    if ((pa->CMD == 0x6) && (pa->FNC == 0x03)) {	// it's that sort...
#if debug >2
	LOG1("diagnostic status\n");
#endif
	AB_block abb;
	memset(&abb, 0, sizeof(abb));
	abb.CMD = pa->CMD;
	abb.FNC = pa->FNC;
	sentTNS = master_send(&abb, 7);
    }
    if ((pa->CMD == 0x1) /*&& (pa->FNC == 0xA2)*/) {	// command has no FNC
#if debug >2
	LOG1("execTI unprotected read from CIF file\n");
#endif
	AB_block2 abb2;
	memset(&abb2, 0, sizeof(abb2));
	abb2.CMD = pa->CMD;
	abb2.byteSize = p->byteCount;
	abb2.loADR = p->startAddr % 256;
	abb2.hiADR = p->startAddr / 256;
	sentTNS = master_send( (AB_block*) &abb2, 9/*14*/); // is this ok??
    }    
    
    res=readAnswer3(b1,&bs,&errorState);
    if (errorState)	{
        errorState=0;
        LOG1("sending NAK and retry read.\n");
//	sendNAK();
	res=readAnswer3(b1,&bs,&errorState);
    }
    if (errorState)	{
        errorState=0;
        LOG1("sending NAK and retry read.\n");
//	    sendNAK();
        res=readAnswer3(b1,&bs,&errorState);
    }    
    if (errorState)	goto error;
    ab = (AB_block *) & bs[2];
    if ((pa->CMD | 0x40) != ab->CMD) {
        LOG3("*** error Command:%x reply:%x.\n", pa->CMD, ab->CMD);
        errorState |= 2;
        goto error;
    }
    rcv_tns = ab->loTNS + 256 * ab->hiTNS;
#if debug>4
    LOG2("protocolRead() TNS of answer %d.\n", rcv_tns);
#endif
    if (sentTNS != rcv_tns) {
        LOG3("*** error wrong TNS: mine %d reply:%d.\n", sentTNS, rcv_tns);
        errorState |= 4;
        goto error;
    }
//	sendACK();
#if debug>4
    LOG3("execTI() bytes required: %d got:%d.\n", p->byteCount, res - 12);
#endif
    result= (uc *) &(ab->FNC);
    return errorState;
    
error:    
    res = readChars(b1, tmo_normal);
//	if (res) sendACK();
    usleep(500000);
    res = readChars(b1, tmo_normal);
//	if (res) sendACK();
    usleep(500000);
    res = readChars(b1, tmo_normal);
//	if (res) sendACK();
    usleep(500000);
    return errorState;
}

int AllenBradleyDriver::protocolWrite(TFI * p, int PLCadr, uc * data,int length) {
    pa = (AllenBradleyTI *) p;
    uc b1[1024], *bs;
    int res, errorState = 0;
    us sentTNS;
    
    if ((pa->CMD == 0xF) && (pa->FNC == 0xAA)) {	// it's a write command
#if debug >1
	LOG1("execTI protected typed write with 3 byte address\n");
	FLUSH;
#endif
/*
    Here we fill the structure AB_block	with the values that do not depend on the value to
    send:
*/	    
	AB_block abb;
	memset(&abb, 0, sizeof(abb));	
	abb.CMD = pa->CMD;			
	abb.FNC = pa->FNC;
	abb.byteSize = length;
	abb.fileNo = p->areaNumber;
	abb.fileType = pa->AB_type;
	abb.eleNo = 0xFF;
	abb.XeleNo = PLCadr;
//	short *w = (short *) &(abb.data);
//	float *wf = (float *) &(abb.data);
	memcpy(&abb.data, data, length);
#if debug >3		    	    
	LOG3("bytes %d varSize %d \n", p->byteCount, length);
#endif	    
        sentTNS = master_send(&abb, 14+length);
	res = readAnswer3(b1,&bs,&errorState);
//	sendACK();
    }
    if (pa->CMD == 0x8) {	// it's an unprotected write command
#if debug >1
	LOG1("execTI unprotected write to CIF file\n");
	FLUSH;
#endif
/*
    Here we fill the structure AB_block	with the values that do not depend on the value to
    send:
*/	    
	AB_block2 abb2;
	memset(&abb2, 0, sizeof(abb2));	
	abb2.CMD = pa->CMD;		
	abb2.byteSize = length;
	int PLCadr = p->startAddr;
	memcpy(&abb2.byteSize, data, length);	// There is no size field here
#if debug >3		    	    
	LOG2("bytes %d\n",length);
#endif
        abb2.loADR = PLCadr % 256;
        abb2.hiADR = PLCadr / 256;
        sentTNS = master_send( (AB_block*) &abb2, 8+length); // is this ok??
	res = readAnswer3(b1,&bs,&errorState);
    }

}

int AllenBradleyDriver::execTI44(TFI * p)
{
    uc b1[1024];
    int errorState = 0, res;
    pa = (AllenBradleyTI *) p;
#if debug >2
    LOG3("execTI CMD %x FNC %x\n", pa->CMD, pa->FNC);
    FLUSH;
#endif
 error:
    LOG4("*** error %d CMD %d FNC %d !\n", errorState, pa->CMD, pa->FNC);
    res = readChars(b1, tmo_normal);
    usleep(500000);
    res = readChars(b1, tmo_normal);
    usleep(500000);
    res = readChars(b1, tmo_normal);
    usleep(500000);
}

#include "thread_makro.cc"
#if debug>0
DTDebug(AllenBradleyDriver)
#else
DT(AllenBradleyDriver)
#endif
/*
    Changes:
    
    10/05/2003  ported to the new driver model.
*/
