/*
 Implementation of a PLC driver for GE Fanuc SNP-X protocol.
 Was originally written for and tested with ALSTOM/Cegelec ALSPA C80-35 PLC. 
 Then I found out that ALSTOM just sells GE Fanuc 90-30 components under
 this name. Everything is identical up to the part numbers.
 
 The SNP-X Protocol is documented by GE in this document:
 
 http://www.geindustrial.com/support/documents/gfk0528d.pdf

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include <stdio.h>
#include "plclog.h"

#include "vi_string.h"
#include "debuglevels.h"
#define debug 17

#define ThisModule "GEFanuc : "


#define read_from_PLC 0x01
#define write_to_PLC 0x02
#define typ_registers 0x08
#define typ_flags 0x16

class GEFanucTI:public TFI {
  public:
/*  
    ui lastSheet;		// last of multiple sheets when we are 
				// connected to multiple PLCs on RS485.
*/				
    int dataArea;		// 0=Data, 1=Flags, 2=Outputs, 3=inputs				
    DECLARE_PERSISTENCE(GEFanucTI)
    void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(GEFanucTI, "GEFanuc")

void GEFanucTI::Read(Engine & e)
{
    TFI::Read(e);
#if debug >3
    show();
#endif
    if (area=="R") dataArea=typ_registers;
    else if (area=="M") dataArea=typ_flags;
/*    
    else if (area=="AI") dataArea=typ_analogIn
    else if (area=="AO") dataArea=typ_analogOut
    TO DO
*/
}

#pragma pack (1)

struct GEFanucMessage {
    unsigned char esc;		// start of message, allways ESCape 
    unsigned char code;		// allways 0x58 = 'X'
    unsigned char SNP_id[8];	// identify PLC on a RS485 bus
    unsigned char read_write;	// direction
    unsigned char typ;
    word adr;
    word size;
    word w1;
    unsigned char b17;
    unsigned char ddd[8];
};

#pragma pack (1)

struct GEFanucAnswer {
    char esc;
    char code;
    char q;
    char a[6];
    char by[1024];
};

class GEFanucDriver:public PLCdriver {
  public:
    int firstSheet, lastSheet;
//    char aMachine[9];
    string snp;
    GEFanucAnswer fa;
    GEFanucMessage msg;
    GEFanucDriver() {};
    int readChars(int size);
    void sendGEFanuc(int len);
    bool GEFanucOpen();
    virtual void intSetup(DriverInfo * i);
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    void execute() {
//	ui i;
	while (1) {
	    execList();
	    execCount++;
	}
    };
//    void coreOfsetLoop(TFI*p, int &i, coord & cell);
    void writeByte(word a, char &b, char t);
    void readByte(word a, char *b, char t);
    void parseNameValuePair(string name,string value);
    int execList();	// we override this to handle multiple sheets
};

int GEFanucDriver::execList()
{
    for (int is=firstSheet; is<lastSheet; is++) {
//	if(driverDebugLevel & debugControlFlow) {
	    LOG3("enter exec list @%p with sheet:%d\n",il,is);
	    if(il) LOG2("%d items in list\n",il->count);
	    FLUSH;
//	}
	for (int i=0;i<il->count;i++) {
	    TFI * p=(TFI*)il->items[i];
	    if(p) {
		p->sheet=is;	
		int res=execTI(p);
		setCell(p->sheet, p->errorCell, res, soIoState);
//	    	if (res!=0) return res;
	    }    
	}
	if(DI){
    	    LOG2("execute: looking for commands %d.\n",PLCdriver::DI->commands);
    	    if(DI->commands & 1) {
	    LOG2("Ooops, commands %d, beginning to reload.\n",DI->commands);
	    unRegisterDriver(bitMask);
	    if (rfd>0){	
	        if ((wfd!=rfd) && (wfd>0)) close(wfd);
	        close(rfd);
	    }	
		DI->commands=0;
		intSetup(DI);
		LOG2("Reload done. Commands: %d\n",DI->commands=0);
	    } else if(DI->commands & 2) {
		LOG2("Ooops, commands %d, beginning to reconnect.\n",DI->commands);
		DI->commands=0;
		invalidateConnection();
	        LOG1("Connection closed.\n");
	    }
	} else {
    	    LOG1("Ooops, thought I kept my DI!\n");
	}
    }	
    return 0;
}


int GEFanucDriver::protocolRead(TFI * p) {
    LOG1("Hello from protocolRead()\n");
    GEFanucTI *pa=(GEFanucTI *)p;
    int wordCount;
    
/*    
    if (snp!=p->deviceAddress) {
	if (!GEFanucOpen()) {
	    return -2;	//failed
	}
    }
*/    
    int bytesRead;
    memset(&msg, 0, sizeof(GEFanucMessage));
    msg.read_write = read_from_PLC;
    msg.typ = pa->dataArea;
    wordCount = p->byteCount;
    switch (pa->dataArea) {
	case typ_flags:
	msg.adr = ((p->startAddr - 1) / 8);
	wordCount = p->byteCount;
    break;	
    case typ_registers:
	msg.adr = 2*(p->startAddr-1);
	wordCount = p->byteCount;
    break;	
    default:    
	LOG2("invalid Data Type: %d\n", p->areaNumber);
    }
    msg.size = wordCount;
    LOG2("wordCount: %d\n", wordCount);
    sendGEFanuc(23);
    
    bytesRead = readChars(p->byteCount + 15);
	    
    LOG2("bytes read from PLC: %d\n", bytesRead);    
    if(bytesRead==p->byteCount + 15) return 0; else  return -1;
}

int GEFanucDriver::protocolWrite(TFI * p, int PLCadr, uc*data, int length) {
    LOG3("protocolWrite(dat:%p, len:%d)\n",data,length);
    GEFanucTI *pa=(GEFanucTI *)p;
    memset(&msg, 0, sizeof(GEFanucMessage));
    msg.read_write = write_to_PLC;
    msg.typ = pa->dataArea;
    int byteStart,byteOffset;
    switch (pa->dataArea) {
	case typ_flags:
	    byteStart = ((p->startAddr - 1) / 8);
	    byteOffset=PLCadr-p->startAddr;
	break;	
	case typ_registers:
	    byteStart = 2*(p->startAddr - 1);
	    byteOffset=PLCadr-p->startAddr;
	break;	
    default:    
	LOG2("invalid Data Type: %d\n", p->areaNumber);
	return -1;
    }
    
    msg.adr = byteStart+byteOffset;
    msg.size = length;
    sendGEFanuc(23);
    int res=readChars(15);
    if (res==15) return 0; else return -2;
}

void GEFanucDriver::parseNameValuePair(string name,string value) {
    if (name=="first") firstSheet = atol(value.c_str()); else
    if (name=="last") lastSheet = atol(value.c_str()); else
    if (name=="debug") driverDebugLevel = atol(value.c_str()); else
    if (name=="libraryDebug") libraryDebugLevel = atol(value.c_str()); else
    LOG2("unsupported parameter: %s\n", name.c_str());
}


void GEFanucDriver::intSetup(DriverInfo * ddi)
{
//    aMachine[0] = 0; !!!
    PLCdriver::intSetup(ddi);
//    srtpSetDebug(libraryDebugLevel);
    rfd = setPort(portParams, &wfd);
    LOG2("intSetup write fd: %d\n", wfd);
//    retryCount=0;
//    openCount=0;
    TypeManager::Add("GEFANUC", CreateNewGEFanucTI);
    loadList(DI);
    bitMask = registerDriver("GEFanuc SNPX", this);
    LOG3("intSetup done fd: %d errorMask %x\n", wfd, bitMask);
    FLUSH;
};

#define tmo_answ 291
#define tmo_read 291
#define tmo_open 291

int GEFanucDriver::readChars(int size)
{
    fd_set FDS;
    bool charsAvail;
    unsigned char msb, sum, *b;
    int i, res;
//    *real_size = 0;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
    usleep(1000 * tmo_read);
    FD_ZERO(&FDS);
    FD_SET(rfd, &FDS);
    charsAvail = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
    t.tv_sec = 0;
    t.tv_usec = 1000 * tmo_read;
#if debug>4
    LOG2("readChars %d\n", charsAvail);
#endif     
    if (!charsAvail) {
	LOG3("%s timeout %d\n", snp.c_str(), charsAvail);
    }
    if (charsAvail) {
#if debug>4    
	LOG2("readChars reading up to %d chars.\n", size);
#endif	
	res = read(rfd, &fa, size);
/*	*real_size = res; */
	b = (unsigned char *) &fa;
#if debug >3    	
#include "dumpchars.cc"
#endif
	LOG2("readChars %d chars read.\n", res);
	sum = 0;
	for (i = 0; i < res - 1; i++) {
//          LOG2(ThisModule"readChars %d \n",i);
	    sum = sum ^ b[i];
	    msb = sum & 0x80;
	    sum = sum << 1;
	    if (msb != 0)
		sum |= 1;
	}
#if debug > 0
	int chs = b[res - 1];
	LOG3("Checksums %d , %d\n", chs, int (sum));
#endif
	LOG3("%s got %d chars\n", snp.c_str(), res);
	FLUSH;
	return (res);
    } else
	return (0);
}

void GEFanucDriver::sendGEFanuc(int len)
{
    int i;
    unsigned char msb, sum;
#if debug > 0
    LOG1("sendGEFanuc enter.\n");
    FLUSH;
#endif				/*  */

    msg.esc = 0x1b;
    msg.code = 0x58;
    memset (msg.SNP_id,0,9);
    strncpy((char*)msg.SNP_id, snp.c_str(), 8);
    msg.b17 = 0x17;
    sum = 0;
    char *f = (char *) &msg;
    for (i = 0; i < len; i++) {
	sum = sum ^ ((unsigned char *) f)[i];
	msb = sum & 0x80;
	sum = sum << 1;
	if (msb != 0)
	    sum = sum | 1;
    }
    ((unsigned char *) f)[len] = sum;
    int res = write(wfd, f, len + 1);
#if debug > 0
    LOG2("sendGEFanuc: %d chars send.\n", res);
#endif				/*  */
}

bool GEFanucDriver::GEFanucOpen()
{
    bool res;
    int res1;
#if debug > 0
    LOG1("GEFanucOpen enter.\n");
    FLUSH;
#endif				/*  */
    while (readChars(255) > 0) {
	LOG1("buffer not empty\n");
    }
    memset(&msg, 0, sizeof(msg));
#if debug > 0
    LOG1("GEFanucOpen fillChar done.\n");
    FLUSH;
#endif				/*  */
    sendGEFanuc(23);
#if debug > 0
    LOG1("GEFanucOpen sendGEFanuc done.\n");
    FLUSH;
#endif
    usleep(1000 * tmo_open);
#if debug > 0
    LOG1("GEFanucOpen usleep(1000*tmo_open) done.\n");
    FLUSH;
#endif
    res1 = readChars(64);
#if debug > 0
    LOG2("GEFanucOpen read done. %d\n", res1);
    FLUSH;
#endif	
    res=(res1 == 24);			/*  */
    return (res);
}

#define read_from_PLC 0x01
#define write_to_PLC 0x02
#define typ_registers 0x08
#define typ_flags 0x16

void GEFanucDriver::writeByte(word a, char &b, char t)
{
    memset(&msg, 0, sizeof(GEFanucMessage));
    msg.read_write = write_to_PLC;
    msg.typ = t;
    msg.adr = a;
    msg.size = 1;
    msg.w1 = b;
    sendGEFanuc(23);
    readChars(15);
}

void GEFanucDriver::readByte(word a, char *b, char t)
{
    int res;
    memset(&msg, 0, sizeof(GEFanucMessage));
    msg.read_write = read_from_PLC;
    msg.typ = t;
    msg.adr = a;
    msg.size = 1;
    sendGEFanuc(23);
    res = readChars(1 + 15);
    *b = (char) fa.by[0];
}



#include "thread_makro.cc"
#if debug>0
DTDebug(GEFanucDriver)
#else
DT(GEFanucDriver)
#endif
/*
    Changes:
    
    02/12/2002 	change output statements to printf, this gives 10% less code size.
    02/28/2002 	some changes and corrections to debug output
    05/13/2002 	Object Read & write functions return void now. cout<<... replaced
		by LOG2(..
    06/20/2002 	* got rid of some warnings
		* uses cellN makro in printf
    07/02/2002 	* thread_macro		
    02/23/2003  got rid of unused vars
*/
