/*
 Implementation of a driver for BECK / FESTO PLCs via Ethernet and IBH-NetLink.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2004.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
//#include "setport.h"
#include "lib0.h"

#include "vi_string.h"

#define debug 6

#include <stdio.h>

extern "C" {
#include "easyIP.h"
}

#include "plclog.h"

//#define sim_broken_transport 

class BECKTI:public TFI {
  public:
    int dataArea;	// 0=Data, 1=Flags, 2=Outputs, 3=inputs
    DECLARE_PERSISTENCE(BECKTI)
    void BECKTI::Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(BECKTI, "BECK")

char * areaName(uc n) {
    switch (n) {
	case easyTypeREGISTER:	return "Register";
	case easyTypeINPUT:	return "Inputs";
	case easyTypeOUTPUT:	return "Outputs";
	case easyTypeMEM:	return "Memory";
	case easyTypeTIMER:	return "Timers";
	default:return "unknown area!";
    }	
}

siEntry areat[] = {
    "R",easyTypeREGISTER,
    "M",easyTypeMEM,
    "A",easyTypeOUTPUT,
    "O",easyTypeOUTPUT,
    "Q",easyTypeOUTPUT,
    "E",easyTypeINPUT,
    "I",easyTypeINPUT,
    "T",easyTypeTIMER,
     };    

    
void BECKTI::Read(Engine & e)
{
    TFI::Read(e);
    show();
    dataArea=translateByTable(area.c_str(), areat, sizeof(areat)/sizeof(siEntry));
    if (dataArea==-1)
	LOG1("Warning: data area code must be in M,R,T,O or Q,I (english) or M,R,T,A,E (german)");
#if debug >3
    LOG3("Data area      :%d (%s)\n", dataArea, areaName(dataArea));
    LOG2("first word     :%d\n", startAddr);
    LOG2("Sheet          :%d\n", sheet);
    LOG2("Bytes          :%d\n", byteCount);
#endif
}

class BECKDriver:public PLCdriver {
  public:
    int res;
    word DriverNr;
    BECKTI *pa;
    word DriverState;
    int openCount;
    int connectionError;
    uc * dataPointer;
    uc buffer[1524];	// max Len of ethernet packet
    BECKDriver() {};
    virtual void intSetup(DriverInfo * i);
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    virtual uc * getProtocolData(){
	return (uc*) dataPointer;
    }
    bool protOK;
    int retryCount;
    void execute() {
	while(1){
	    if(connectionError) {
	        LOG1("execute: di error\n");
	        if (rfd>0){	
		    if ((wfd!=rfd) && (wfd>0)) close(wfd);
		    close(rfd);
		}	
		LOG2("execute: write fd: %d trying to connect\n", wfd);
		res=beckConnect(DI->s.c_str(), &rfd, &wfd);
		if(wfd>0) {
		    connectionError=0;
		    openCount=0;
		}	
		LOG2("execute: write fd: %d after connect\n", wfd);
	    }    
	    if (connectionError==0) {
	        protOK=true;
	        DI->commands=0;
	        while (protOK) {
		    LOG1("execute: before execList()\n");
		    res=execList();
		    if (res!=0) protOK=false;
		    LOG1("execute: after execList()\n");
		    execCount++;
		    if(!protOK) {
			LOG1("execute: error detected.\n");
		        DI->commands|=1;
		        close (wfd);
		        connectionError=1;
		        wfd=-1;
		        rfd=-1;
		        usleep(999999);
		    }    
		}
		LOG1("execute: protOK is false.\n");
		FLUSH;
	    } else {
	        usleep(200000);
	    }            
	}    
    };
};

void BECKDriver::intSetup(DriverInfo * ddi)
{
    beckEasyDebug=100;
    DI=ddi;
    int res=beckConnect(DI->s.c_str(), &rfd, &wfd);
    LOG2("intSetup write fd: %d\n", wfd);
    retryCount=0;
    openCount=0;
    DI=ddi;
    TypeManager::Add("BECK2", CreateNewBECKTI);
    loadList(DI);
    bitMask = registerDriver("BECK/FESTO", this);
    LOG3("intSetup done fd: %d errorMask %x\n", wfd, bitMask);
    FLUSH;
};

int BECKDriver::protocolRead(TFI * p) {
    LOG1("Hello from protocolRead()\n");
    
    BECKTI *pa=(BECKTI *)p;
    int res=beckReadBytes(rfd,wfd, 
	pa->dataArea, /*p->areaNumber,*/ p->startAddr, (p->byteCount+1)/2, buffer);
    dataPointer=(uc*)buffer;	
#ifdef sim_broken_transport
    
    double pr=random();
    pr/=RAND_MAX;
	
    if (pr>=0.95) {
        LOG2("faking broken transport: %0.2f >= 0.95 \n",pr);
	res=-1;
    }    
#endif	

#if debug >5
    LOG2("protocolRead: beckReadBytes returned:%d\n", res); 
#endif	    	
    if(res==0) return 0; 
    else{
	protOK=false;
      return -1;
    }  
//    return res;
}

int BECKDriver::protocolWrite(TFI * p, int PLCadr, uc*data, int length) {
/*
    We can read single bytes, reading a word and return only it's LSB or MSB,
    but we cannot do this with write. What would we write to the the other half?
*/
    if ((length % 2)!=0	) return -2;
    
    LOG3("protocolWrite(dat:%p, len:%d)\n",data,length);
    BECKTI *pa=(BECKTI *)p;
/*
    I think wee need to correct the PLC address here:
    If we specify format 'w','i', the offset is incremented by 2 (bytes), if we specify 
    format 'l' 'u' or 'f', the offset is incremented by 4 (bytes), 
    It is the start address + number of bytes, but we need start + number of words.
    We recalculate the word offset from byte offset and startAddr.
    This may lead to "fractional" word addresses if byte offset is odd.
    Report an error then. This makes format specifiers 'b','c' and x unusable for write.
    TO DO: Allow use of single bits by using a word sized "bitContainer". See PLCdriver.
*/    
    int offset=PLCadr-p->startAddr;
    if ((offset & 1)!=0) {
	LOG2("Cannot write single bytes or to half word Adresses! Check the format filed in %s !\n",p->comment.c_str());
	return -10;
    }
    offset/=2;
    int res=beckWriteBytes(rfd, wfd, pa->dataArea, /*p->areaNumber,*/ 
	p->startAddr+offset/*PLCadr*/, length/2, data);
    if (res>0) res=0;	// this version of nodave returns <0 for errors;
    else {
//	protOK=false;
    }
    return res;
}

#include "thread_makro.cc"
#if debug>0
DTDebug(BECKDriver)
#else
DT(BECKDriver)
#endif
