/*
 Implementation of a PLC driver for the BECKHOFF BC9000/BK9000 controller. 

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include "lib0.h"
#include <stdio.h>
#include "plclog.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <errno.h>
#include "nethelp.h"

#include "byteswap.h"

#include "vi_string.h"

#include "modbus.h"

#define uc unsigned char

class modbusTI:public TFI {
  public:
    int wordCount;
    DECLARE_PERSISTENCE(modbusTI)
    void Read(Engine & e);
    int function;
};

IMPLEMENT_PERSISTENCE(modbusTI, "BECKHOFF")

void modbusTI::Read(Engine & e)
{
    TFI::Read(e);
    function=atol(method.c_str()); 
    isWrite=((function==5)||(function==6)||(function==15)||(function==16));
    wordCount = (byteCount+1)/2;
    show();
#if debug >3
    LOG3("Modbus function :%d = %s\n", function, modbusFunctionName(function));
    LOG2("Bytes          :%d\n", byteCount);
#endif
}

class modbusDriver:public PLCdriver {
  public:
    word DriverNr;
    modbusTI *pa;
    word DriverState;
    modbusMessage msg;
    int transaction_id;
    uc ibuf[1500];
    int tmo_read;
    int errorCode;
    string peer;
    modbusDriver() {};
    int readChars(int size);
    void sendmodbus(int len);
    void watchdog_reset();
    int checkAnswer(int i, int count, int func);
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (doRun) {
	    if (rfd>0) {	// will never be 0
		execList();
		execCount++;
	    } else {
		usleep(500000);	// delay between connect trials
		openSock();	// try to connect
	    }
	}
    };
    void openSock();
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    virtual uc * getProtocolData(){
	return ibuf+9;
    }
};

int modbusDriver::protocolRead(TFI * p) {
    modbusTI * pa = (modbusTI *) p;
    LOG3("function %d = %s\n",pa->function, modbusFunctionName(pa->function));
    memset(&msg, 0, sizeof(modbusMessage));
    msg.function = pa->function;
    msg.ddd[0] = p->startAddr / 256;
    msg.ddd[1] = p->startAddr % 256;
    msg.ddd[2] = pa->wordCount / 256;
    msg.ddd[3] = pa->wordCount % 256;
    sendmodbus(12);
    int i=readChars(1024);
    while (i==1024) i=readChars(1024);
    errorCode=checkAnswer(i, 2*pa->wordCount, pa->function);
    LOG3("function %d = %s\n",pa->function, modbusFunctionName(pa->function));
    LOG2("error: %d\n",errorCode);
    return errorCode;
}

int modbusDriver::protocolWrite(TFI * p, int PLCadr, uc * data,int length) {
    int offset=PLCadr-p->startAddr;
    offset=(offset+1)/2;
    int useAdr=p->startAddr+offset;
    modbusTI * pa = (modbusTI *) p;
    int wordCount = (length+1)/2;
    memset(&msg, 0, sizeof(modbusMessage));
    msg.function = pa->function;
    if (driverDebugLevel>1)
	LOG4("register address %d length:%d words:%d\n",PLCadr,length,wordCount);
    msg.ddd[0] = (useAdr) / 256;
    msg.ddd[1] = (useAdr) % 256;
    msg.ddd[2] = wordCount/256;
    msg.ddd[3] = wordCount%256;
    msg.ddd[4] = length;
    memcpy(&msg.ddd[5],data,length);
//    sendmodbus(12);
    sendmodbus(13+length);
    int i=readChars(1024);
    errorCode=checkAnswer(i, 3, pa->function);
    return errorCode;
}

void modbusDriver::openSock()
{
    tmo_read = 16291;
    rfd = wfd= openSocket(502, peer.c_str());
    if (driverDebugLevel>1) {
	LOG2("OpenSocket: socket is %d\n", wfd);
        FLUSH;
    }	
}

void modbusDriver::intSetup(DriverInfo * di)
{
    DI=di;
    TypeManager::Add("BECKHOFF", CreateNewmodbusTI);
    peer=noSpace(DI->s);
    openSock();
    loadList(DI);
    bitMask = registerDriver("BECKHOFF", this);
};

int modbusDriver::readChars(int size)
{
    fd_set FDS;
    bool charsAvail;
    unsigned char *b;
    int res;
    timeval t;
    t.tv_sec = 0;
    t.tv_usec = tmo_read;
    usleep(tmo_read);
    FD_ZERO(&FDS);
    FD_SET(rfd, &FDS);
    charsAvail = (select(rfd + 1, &FDS, NULL, NULL, &t) > 0);
    if (driverDebugLevel>1)    
        LOG2("readChars %d\n", charsAvail);
    if (!charsAvail) {
	LOG2(" timeout %d \n", charsAvail);
    }
    if (charsAvail) {
	if (driverDebugLevel>1)
	    LOG2("readChars reading up to %d chars.\n", size);
	res = read(rfd, &ibuf, size);
	b = (unsigned char *) &ibuf;
	if (driverDebugLevel>1) {
#include "dumpchars.cc"
	    LOG2("readChars %d chars read.\n", res);
	    FLUSH;
	}
	return (res);
    } else
	return (0);
}

void modbusDriver::sendmodbus(int len) {
    int res;    
    res=modbusSend(len, &transaction_id, tmo_read, &msg, wfd);    
    if (res<0) {
	LOG1("sendmodbus: cannot write. Must reconnect\n");
	openSock();
    }	
}

/*
    This is very special to BECKHOFF couplers:
*/

void modbusDriver::watchdog_reset()
{
    int i;
    LOG1("watchdogReset\n");
    msg.unit_id = 1;
    msg.function = 6;
    msg.ddd[0] = 0x11;
    msg.ddd[1] = 0x21;
    msg.ddd[2] = 0xbe;
    msg.ddd[3] = 0xcf;
    sendmodbus(12);
    i=readChars(261);
    checkAnswer(i, 0, 6);

    msg.unit_id = 1;
    msg.function = 6;
    msg.ddd[0] = 0x11;
    msg.ddd[1] = 0x21;
    msg.ddd[2] = 0xaf;
    msg.ddd[3] = 0xfe;
    sendmodbus(12);
    i=readChars(261);
    checkAnswer(i, 0, 6);
}

int modbusDriver::checkAnswer(int i, int num_regs, int func) {
    int errC=0;
    if (i < 9) {
	if (i == 0) {
	    LOG1("unexpected close of connection at remote end\n");
	    errC|=512;
	} else {
	    LOG2("response was too short - %d chars\n", i);
	    errC|=256;
	}
    } else if (0 != (ibuf[7] & 0x80)) {
	LOG2("MODBUS exception response - type %d\n", ibuf[8]);
	errC|=ibuf[8];
	if (ibuf[8] == 4) {
	    if (num_regs) watchdog_reset();
	}
    } else if (ibuf[7] != func) {
	LOG3("wrong response %d for function %d\n", func, ibuf[7]);
	tmo_read = 10 + (3 * tmo_read) / 2;
	errC|=1024;
	return errC;
    } else if (i != (9 + num_regs)) {
	LOG3("incorrect response size is %d expected%d\n",i, (9 + num_regs));
	errC|=2048;
    } else {
    if (driverDebugLevel>11) {
	for (i = 0; i < num_regs; i++) {
	    int w = (ibuf[9 + i]);	
	    // + (ibuf[10+i+i]& 0x00FF);
	    LOG3("byte %d = %x\n", i, w);
	} 
    }
	if (!tmo_read) tmo_read = 1;
//	return errC;
    }
    if (driverDebugLevel>2)
	LOG2("about to return error code %d \n", errC);
    return errC;
}

#include "thread_makro.cc"
#if debug>0
DTDebug(modbusDriver)
#else
DT(modbusDriver)
#endif
/*
    Changes:
    
    04/30/2002	written
    06/29/2002	got rid of some warnings
    01/17/2003 	implemented function 6, preset single register
    01/28/2003	reconnect automatically
		implemented function 3, read holding registers
		implemented function 4, read input registers
    01/29/2003	connect later if device wasn't available at startup
    09/26/2003	adaptions to new basic PLCdriver
    04/01/2004	inserted some delays. With serial drivers, they should
		try to get maximum communication performance, but ethernet drivers may
		repeat at unreasonable rates. 
		TODO: find a solution to configure delays.
    04/01/2004	mark errors in output with *** 
*/
