/*
 Implementation of a serial line slave driver for OMRON PLCs using
 hostlink protocol.
 This should make visual behave more or less like an OMRON PLC. The first
 gola of this is to communicate with SOMETHING because i've no OMRON
 PLC to test. As with the Allen Bradley slave, one can also use it
 as a kind of bridge:
 Disconnect an existing SCAD/HMI or other communication device from
 the PLC, connect VISUAL to the PLC, let VISUAL get all the data from 
 the PLC, including what the other device needs and connect the other 
 device to a serial interface handled by this slave driver.


 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2002.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "plcdrv.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"
#include "plclog.h"

#include "vi_string.h"

#define debug 4

#include <stdio.h>

#define tmo_normal 999999
#define ThisModule "OMRON_SLV: "

#include "omron_lib.h"

//#define sim_broken_transport
//#include "omron.h"


/*
    The Slave maps a combination of Unit ID and a header Code to a combination
    of a sheet and a column, beginning at startCell. 
    So lets assume we get a "read 20 bytes as integers from 
    N7:50" command. 
    Then we look for a TI that maps RR command somewhere. 
    Lets assume we find one that maps RR to sheet 1, startCell Q17.
    Now the first value is in Q17 corresponding to word 0.
    We start reading data at Q67 corresponding to 50 th word of RR.
    As 20 bytes are 10 Integers, we take them from Q67 to Q76.
    
*/

class Omron_Slave_TI:public transferInstruction {
  public:
    int unitID;		//      Omron's numerical unit ID
    string headerCode;
    DECLARE_PERSISTENCE(Omron_Slave_TI)
    void Omron_Slave_TI::Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(Omron_Slave_TI, "OMRON_SL")

void Omron_Slave_TI::Read(Engine & e)
{
    string ctc, stc, erc;
    e >> comment >> unitID >> headerCode >> sheet >> stc;
    startCell = getPos(stc);
#if debug >3
    LOG2("comment        :%s\n", comment.c_str());
    LOG2("Unit ID        :%d\n", unitID);
    LOG2("Header         :%s\n", headerCode.c_str());
    LOG2("sheet          :%d\n", sheet);
    LOG2("Map to Column  :%s\n", cellN(startCell));
#endif
}

class OmronSlave:public OmronDriver {
  public:
    OmronSlave() { };
    void execSlave();
    virtual void intSetup(DriverInfo * i);
    void execute() {
	while (1) {
	    execSlave();
	    execCount++;
	}
    }
};

void OmronSlave::intSetup(DriverInfo * i)
{
    rfd= setPort(i->s, &wfd);
    LOG2("intSetup write fd: %d\n",wfd);
    TypeManager::Add("OMRON_SL", CreateNewOmron_Slave_TI);
    PLCdrv::load(i);
    changedMask = registerDriver("OmronSlave", this);
    errorMask = changedMask;
};

void OmronSlave::execSlave()
{
    LOG1("execSlave() starts.\n");
    FLUSH;
    double rvalue;
    coord cell;
    uc b1[1024];
    int i,errorState=0;
    bool ok;
    errorState=false;
    int res=readAnswer(b1);
    if (!errorState){
	int unit=ascii_decode(b1+1,2);
	LOG3("got command: %c%c \n",b1[3],b1[4]);	
	int start=ascii_decode(b1+5,4);
	int byteCount=ascii_decode(b1+5+4,4);
	LOG4("unit: %d 1st word: %d count: %d\n",unit,start,byteCount);	
//	string headerCode=string((char*) *(b1+3))+string((char*) *(b1+4));
	b1[5]=0;
	Omron_Slave_TI *ent;
	LOG1("checking TFIs\n");	
        for (i = 0; i < il->count; i++) {
	    ent = (Omron_Slave_TI *) (il->items[i]);
#if debug >3
	    LOG5("entry: %d %s<->%d:%s\n",
		ent->unitID, ent->headerCode.c_str(),ent->sheet, cellN(ent->startCell)
	    );
#endif
	    if ((ent->unitID == unit)&&(strcmp(ent->headerCode.c_str(),(char*)(b1+3))==0)) {
#if debug >0	    
	        LOG2("entry no. %d is the right statement\n", i);
	        FLUSH;
#endif		
	        break;
	    }
	    ent = NULL;
	}
	if (ent) {
	    int start=ascii_decode( b1+5,4);
	    int count=ascii_decode( b1+9,4);
#if debug >3	    
	    LOG3("start %d, count %d\n", start, count);
#endif			    
	    cell=ent->startCell;
	    cell.row+=start;
#if debug >3	    
	    LOG2("start with cell %s\n", cellN(cell));
#endif			   
	    ascii_encode(b1+5,0,2);
	    int omlen=7;
	    for (i=0;i<count;i++) {
		rvalue=numContents(ent->sheet,cell,ok);
		cell.row++;
		ascii_encode(b1+omlen,(int)rvalue,4);
		omlen+=4;
	    }
	    omron_send(b1,omlen);	    
	}
    }
}

#include "thread_makro.cc"
#if debug>0
DTDebug(OmronSlave)
#else
DT(OmronSlave)
#endif

