/*
 Implementation of a PLC driver for the SIEMENS S5 series via its programmer
 port. See www.runmode.com for information about the underlying AS511
 protocol and a proposal of circuit to connect the current loop programmer
 port to a common V24.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001..2005.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "PLCdriver.h"
#include "s5.h"
#include "s5l.h"
#include "lib0.h"
#include "setport.h"

#include "vi_string.h"

#include "plclog.h"
#define debug 5

class S5TFI:public TFI {
public:
    int dataArea;
    blockInfo BI;
    bool blockinfoValid;
    DECLARE_PERSISTENCE(S5TFI)
    virtual void Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(S5TFI, "S5")


#define daveDB 1
#define daveFlags 2
#define daveOutputs 3
#define daveInputs 4

char * daveAreaName(uc n) {
    switch (n) {
	case daveDB:		return "DB";
	case daveInputs:	return "Inputs";
	case daveOutputs:	return "Outputs";
	case daveFlags:		return "Flags";
	default:return "unknown area!";
    }	
}

siEntry areat[] = {
     {"DB",daveDB},
     {"F",daveFlags},
     {"M",daveFlags},
     {"A",daveOutputs},
     {"O",daveOutputs},
     {"Q",daveOutputs},
     {"E",daveInputs},
     {"I",daveInputs},
     };    

void S5TFI::Read(Engine & e)
{
    TFI::Read(e);    
    blockinfoValid=false;
    dataArea=translateByTable(area.c_str(), areat, sizeof(areat)/sizeof(siEntry));
    if (dataArea==-1)
	 LOG1("Warning: data area code must be in DB,F,O or Q,I (english) or DB,M,A,E (german)");
#if debug >3
    LOG3("Data area      :%d (%s)\n", dataArea, daveAreaName(dataArea));
    LOG2("Data Block     :%d\n", areaNumber);
    LOG2("first word     :%d\n", startAddr);
    LOG2("Sheet          :%d\n", sheet);
    LOG2("Bytes          :%d\n", byteCount);
#endif
#if debug >3
    show();
    LOG2("direction:     :%s\n", method.c_str());
#endif
}

class S5drv:public PLCdriver{
  public:
    bool doVerifySyspars;	// Read system parameters twice and verify. Slower.
    bool doVerifyDBaddress;	// Allways read DB address before accessing a DB. Slower.
    bool sysParsOk;			
    char flagsOk;
    char inputsOk;
    char outputsOk;
    int firstFlag;		//the first byte of flags we copy
    int zFlags;			//number of flag bytes we copy
    int firstOutput;		//the first byte of outputs we copy
    int zOutputs;		//number of outputs bytes we copy
    int firstInput;		//the first byte of inputs we copy
    int zInputs;		//number of inputs bytes we copy
    sysPars sp;
    bool execError;
    S5drv();
    unsigned char DB_buffer[1024];
    virtual void execute();
    virtual void intSetup(DriverInfo * i);
    
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    virtual uc * getProtocolData() 	// pointer to an array of bytes. Either the protocol 
    					// library or the driver must provide it.
    
    {
	return DB_buffer;
    }
    virtual bool getSysParams();
    void parseNameValuePair(string name,string value);
};


int S5drv::protocolRead(TFI * p) {
//    LOG1("Hello from protocolRead()\n");
    S5TFI *pa=(S5TFI *)p;
    switch(pa->dataArea) {
	case daveFlags:
	    if (readData (rfd, wfd, sp.Merker + p->startAddr, p->byteCount, DB_buffer)) {
		return 0;
	    }
	break;
	case daveInputs:
	    if (readData (rfd, wfd, sp.Inputs + p->startAddr, p->byteCount, DB_buffer)) {
		return 0;
	    }
	break;
	case daveOutputs:
	    if (readData (rfd, wfd, sp.Outputs + p->startAddr, p->byteCount, DB_buffer)) {
		return 0;
	    }
	break;
	case daveDB:
	    if (doVerifyDBaddress) pa->blockinfoValid=false;
	    if(!pa->blockinfoValid) {
		pa->blockinfoValid=(getBlockInfo(rfd,wfd, bt_DB, p->areaNumber, &(pa->BI)));
	    }
	    if(pa->blockinfoValid) {
		if (readData(rfd,wfd, pa->BI.address + 2 * p->startAddr, p->byteCount,  DB_buffer)) {
		    return 0;
		}
		else {
		    execError = true;
		    sBreak(rfd,wfd, DBReadFault);
		}
	    } else {
		execError = true;
		sBreak(rfd,wfd, DBNotPresent);
	    };
	break;
    }
    return -1;
}


int writePLCdata(word memAddress, uc* value,/* sysPars * sp,*/ int rfd, int wfd, int length) {
	uc x6[] = {0x10, 0x06, 0x02}; 
	uc x7[] = {0x12, 0x10, 0x03}; 
	if (!prolog(rfd,wfd, 0x03)) {
	    sBreak(rfd,wfd, SVAR_SF + 0x200);
	    return -1; 
	}
	sendWordData(wfd, memAddress);
	for (int i=0; i<length; i++)
	    sendData(wfd, value[i]);
	s1004(wfd);
	if (!testChars(rfd, x6, sizeof(x6), true)) {
	    sBreak(rfd,wfd, SVAR_T100602 + 0x200);
	    return -2; 
	}
	s1006(wfd);
	if (!testChars(rfd, x7, sizeof(x7), true)) {
	    sBreak(rfd,wfd, SVAR_T121003 + 0x100);
	    return -3; 
	}
	s1006(wfd);
	return 0; 
    }


int S5drv::protocolWrite(TFI * p, int PLCadr, uc*data, int length) {
    LOG3("protocolWrite(dat:%p, len:%d)\n",data,length);
    S5TFI *pa=(S5TFI *)p;
    blockInfo BI;
    switch(pa->dataArea) {
	case daveFlags:
	    return writePLCdata(PLCadr+sp.Merker, data, rfd, wfd, length);
	case daveInputs:
	    return writePLCdata(PLCadr+sp.Inputs, data, rfd, wfd, length);
	case daveOutputs:
	    return writePLCdata(PLCadr+sp.Outputs, data, rfd, wfd, length);
	case daveDB:
	    if (getBlockInfo(rfd,wfd, bt_DB, p->areaNumber, &BI)) {
		return writePLCdata(BI.address + 2 *PLCadr, data, rfd, wfd, length);
	    } else {
	        sBreak(rfd,wfd, DBReadFault);
		return -2;
	    }
    }	
    return -1;
}

bool S5drv::getSysParams() {
    bool spOk;
    spOk=getSysPars(rfd, wfd, &sp, false);
    if (spOk) spOk=getSysPars(rfd, wfd, &sp, true);
    sp.Inputs = _swap(sp.Inputs);
    sp.Outputs = _swap(sp.Outputs);
    sp.Merker = _swap(sp.Merker);
    sp.Zeiten = _swap(sp.Zeiten);
    sp.Zaehler = _swap(sp.Zaehler);	    
    return spOk;	    
}    

void S5drv::execute()
{
    int errorCode;
    do {
	if (!sysParsOk || doVerifySyspars) sysParsOk=getSysParams();
	if (!sysParsOk) {
#if debug > 0
	    LOG1("sysParsError!\n");
	    FLUSH;
#endif				/*  */
	    sBreak(rfd, wfd, errorCode + 0x100);
#if debug > 0
	    LOG1("SysParsError , sBreak done.\n");
	    FLUSH;
#endif				/*  */
//	    il->forEach(globalErrorSet, this);
	    tcsendbreak(wfd, 410);
#if debug > 0
	    LOG1("TCSendBreak after global error done.\n");
	    FLUSH;
#endif
	    usleep(500 * 1000);
	    execError = true;
	} else {		// System parameters read ok
#if debug > 0
	    LOG2("SysPars ok. before execList() %d\n", firstFlag + zFlags);
	    FLUSH;
#endif
	    execList();
	}
	execCount++;
    }
    while (doRun);
}

S5drv::S5drv()
{
}

void S5drv::parseNameValuePair(string name,string value) {
    if (name=="debug") driverDebugLevel = atol(value.c_str()); else
    if (name=="daveDebug") libraryDebugLevel = atol(value.c_str()); else
    if (name=="verifySys") doVerifySyspars = (atol(value.c_str())>0); else
    if (name=="verifyDBaddress") doVerifyDBaddress = (atol(value.c_str())>0); else
    LOG2("unsupported parameter: %s\n", name.c_str());
}

void S5drv::intSetup(DriverInfo * i)
{
    doVerifySyspars = false;
    doVerifyDBaddress = false;
    sysParsOk=false;
//    PLCdriver::intSetup(i);
    rfd = setPort(i->s, &wfd);
    LOG2("intSetup wfd: %d\n", wfd);
    TypeManager::Add("S5", CreateNewS5TFI);
    loadList(i);
    bitMask = registerDriver("Simatic S5", this);
}

#include "thread_makro.cc"
#if debug>0
DTDebug(S5drv)
#else
DT(S5drv)
#endif
/*
    Changes:
    31/12/2003 new style driver
*/
