/*
 Implementation of a driver for Siemens S7-300/400 PLCs via Ethernet,MPI,PPI and IBH-NetLink.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2003.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "PLCdriver.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include "setport.h"
#include "lib0.h"

#include "vi_string.h"

#define debug 6

#include <stdio.h>

extern "C" {
#define LINUX
#include <nodave.h>
}

#include "plclog.h"

class S7_TI:public TFI {
  public:
    int MPI_Adr;	// MPI address of PLC
    int dataArea;	// 0=Data, 1=Flags, 2=Outputs, 3=inputs
    DECLARE_PERSISTENCE(S7_TI)
    void S7_TI::Read(Engine & e);
};

IMPLEMENT_PERSISTENCE(S7_TI, "S7")

siEntry areat[] = {
     {"DB",daveDB},
     {"F",daveFlags},
     {"M",daveFlags},
     {"A",daveOutputs},
     {"O",daveOutputs},
     {"Q",daveOutputs},
     {"E",daveInputs},
     {"I",daveInputs},
     };    

void S7_TI::Read(Engine & e)
{
    TFI::Read(e);
    show();
    MPI_Adr=atol(deviceAddress.c_str());
    dataArea=translateByTable(area.c_str(), areat, sizeof(areat)/sizeof(siEntry));
    if (dataArea==-1)
	 LOG1("Warning: data area code must be in DB,F,O or Q,I (english) or DB,M,A,E (german)");
#if debug >3
    LOG2("MPI Address    :%d (not implemented)\n", MPI_Adr);
    LOG3("Data area      :%d (%s)\n", dataArea, daveAreaName(dataArea));
    LOG2("Data Block     :%d\n", areaNumber);
    LOG2("first word     :%d\n", startAddr);
    LOG2("Sheet          :%d\n", sheet);
    LOG2("Bytes          :%d\n", byteCount);
#endif
}

class S7MPIDriver:public PLCdriver {
  public:
    int protocol,rack,slot,speed,plcMPI;
    bool di_error;
    _daveOSserialType fd;
    daveInterface * di;
    daveConnection * dc;
    
    int openCount;
    S7MPIDriver() {};
    bool openMPI();
    virtual void intSetup(DriverInfo * i);
    virtual void shutdown();
    virtual void invalidateConnection();
    virtual int protocolRead(TFI * p);
    virtual int protocolWrite(TFI * p, int PLCadr, uc * data,int length);
    virtual uc * getProtocolData(){
	return (uc*) dc->resultPointer;
    }
    bool protOK;
    int retryCount;
    void execute() {
	while(doRun){
	    if(di_error) {
		LOG1("execute: di error\n");
	        if (rfd>0){	
		    if ((wfd!=rfd) && (wfd>0)) close(wfd);
		    close(rfd);
		}	
		LOG2("execute: write fd: %d trying to connect\n", di->fd.wfd);
		    
		di->fd.rfd=setPort(portParams, &di->fd.wfd);
		if(di->fd.wfd>0) {
		    di_error=false;
		    openCount=0;
		}	
		    LOG2("execute: write fd: %d after connect\n", di->fd.wfd);
		}    
		if (openMPI()) {
		    protOK=true;
		    DI->commands=0;
		    while (protOK && doRun) {
			execList();
			execCount++;
			if(!protOK) {
			    DI->commands|=1;
			    close (di->fd.wfd);
			    close (di->fd.rfd);
			    di_error=true;
//			    di->fd.wfd=-1;
//			    di->fd.rfd=-1;
			    usleep(999999);
			}    
		    }
		    LOG1("execute: protOK is false.\n");
		    FLUSH;
		} else {
		    usleep(200000);
		}            
	}    
	shutdown();
    };

    virtual const char * getPLCAreaName(TFI * p) {
	return p->area.c_str();
    }    
    void parseNameValuePair(string name,string value);
    virtual char * getProtocolErrorMessage(int code) {
	return daveStrerror(code);
    }
    
};

siEntry protot[] = {
    {"PPI",    daveProtoPPI},
    {"MPI",    daveProtoMPI},
    {"MPI2",   daveProtoMPI2},
    {"ISO243", daveProtoISOTCP243},
    {"ISOTCP243", daveProtoISOTCP243},
    {"ISOTCP", daveProtoISOTCP},
    {"IBH",    daveProtoMPI_IBH},
    {"IBHMPI", daveProtoMPI_IBH},
    {"MPIIBH", daveProtoMPI_IBH},
    {"PPIIBH", daveProtoPPI_IBH},
    {"IBHPPI", daveProtoPPI_IBH},
};    

void S7MPIDriver::parseNameValuePair(string name,string value) {
    if (name=="protocol") protocol= translateByTable(value.c_str(), protot, sizeof(protot)/sizeof(siEntry));else
    if (name=="slot") slot = atol(value.c_str()); else
    if (name=="rack") rack = atol(value.c_str()); else
    if (name=="debug") driverDebugLevel = atol(value.c_str()); else
    if (name=="daveDebug") libraryDebugLevel = atol(value.c_str()); else
    if (name=="MPI") plcMPI = atol(value.c_str()); else
    LOG2("unsupported parameter: %s\n", name.c_str());
}
    
void S7MPIDriver::intSetup(DriverInfo * ddi) {
    protocol=daveProtoISOTCP;
    slot=2; 
    rack=0;
    plcMPI=2;
    PLCdriver::intSetup(ddi);
    daveSetDebug(libraryDebugLevel);
    LOG2("protocol: %d\n", protocol);
    rfd = setPort(portParams, &fd.wfd);
    fd.rfd=rfd;
    wfd=fd.wfd;
    LOG2("intSetup write fd: %d\n", fd.wfd);
    retryCount=0;
    openCount=0;
    di=daveNewInterface(fd,"IF1",0, protocol, daveSpeed187k);
    LOG3("protocol: %d di:%d\n", protocol,di->protocol);
    dc=daveNewConnection(di, plcMPI, rack, slot);
    TypeManager::Add("S7", CreateNewS7_TI);
    loadList(DI);
    bitMask = registerDriver("S7", this);
    LOG3("intSetup done fd: %d errorMask %x\n", wfd, bitMask);
    FLUSH;
};

bool S7MPIDriver::openMPI() {
    int res;
    LOG2("openMPI trial %d\n",openCount);
    if (openCount>2) { 
	di_error=true;
	return false;
    }
    LOG2("openMPI protocol %d\n",di->protocol);
    if (daveInitAdapter(di)) return false;
//    dc=daveNewConnection(di, MPIADR, 0,2);
    res=daveConnectPLC(dc);
    if (res==0) {
	openCount=0;
        return true;
    } else {
	openCount++;
	return false;
    }
}    

int S7MPIDriver::protocolRead(TFI * p) {
    LOG1("Hello from protocolRead()\n");
    S7_TI *pa=(S7_TI *)p;
    int res=daveReadBytes(dc, 
	pa->dataArea, p->areaNumber, p->startAddr, p->byteCount, NULL);
//    res=dc->AnswLen;
    if (res<0) protOK=false;
    if(p->byteCount!=dc->AnswLen) protOK=false;
#if debug >5
    LOG4("protocolRead: need:%d got:%d error:%d\n",p->byteCount, dc->AnswLen, res); 
#endif	    	
    return res;
}

int S7MPIDriver::protocolWrite(TFI * p, int PLCadr, uc*data, int length) {
    LOG3("protocolWrite(dat:%p, len:%d)\n",data,length);
    S7_TI *pa=(S7_TI *)p;
    int res=daveWriteBytes(dc, pa->dataArea, p->areaNumber, PLCadr, length, data);
    if (res>0) res=0;	// this version of nodave returns <0 for errors;
    else {
//	protOK=false;
    }
    return res;
}

void S7MPIDriver::shutdown() {
    LOG1("shutting down driver\n");
    daveDisconnectPLC(dc);
    daveDisconnectAdapter(di);
}

void S7MPIDriver::invalidateConnection() {
    protOK=false;
    LOG1("Hello from invalidateConnection!\n");
}

#include "thread_makro.cc"
#if debug>0
DTDebug(S7MPIDriver)
#else
DT(S7MPIDriver)
#endif
