/*
 Definitions for BECKHOFF's ADS protocol. 
 ADS = Automation Device Specification
 Implemented according to specifications given in TwinCAT Information System Nov 2002.
 TwinCAT, ADS and maybe other terms used herein are registered trademarks of BECKHOFF 
 Company. www.beckhoff.de

 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2003.

 This is free software; you can redistribute it and/or modify
 it under the terms of the GNU Library General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This Software is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU Library General Public License
 along with this; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#ifndef ADS__
#define ADS__

/* 
    Some simple types:
*/
#define uc unsigned char //characters
#define us unsigned short 
#define u32 unsigned int

#define USHORT unsigned short 
#define DWORD unsigned int

#pragma pack (1)

typedef struct _AMS_TCPheader {
    us	reserved;	// must be zero for now;
    u32 length;		//length in bytes
} AMS_TCPheader;

/*
    According to BECKHOFF, AMS Id is a six byte field. In case of IP transport, the first
    4 bytes are identical to the IP Address.
*/
typedef struct _AMSNetID {
    uc b1,b2,b3,b4;	// normally equal to IP address
    uc  b5;		// extra bytes
    uc  b6;
} AMSNetID, AmsNetId;


typedef struct _AmsAddr {
    AmsNetId netId;
    USHORT port;
} AmsAddr, * PAmsAddr;

/*
    The AMS header.
*/

typedef struct _AMSheader {
    AMSNetID	targetId;
    us		targetPort;
    AMSNetID	sourceId;
    us		sourcePort;
    us		commandId;
    us		stateFlags;
    u32 	dataLength;
    u32 	errorCode;
    u32 	invokeId;	//user-defined 32-bit field. Usually it is used to identify
				// a response as belonging to a certain request sent with the 
				// same value
} AMSheader;

/*
    Definitions of commands:
*/
#define cmdADSinvalid		0x0000
#define cmdADSreadDevInfo	0x0001
#define cmdADSread		0x0002
#define cmdADSwrite		0x0003
#define cmdADSreadState		0x0004
#define cmdADSwriteControl	0x0005
#define cmdADSaddDeviceNotification	0x0006
#define cmdADSdeleteDeviceNotification	0x0007
#define cmdADSdevNotify		0x0008
#define cmdADSreadWrite		0x0009

/*
    Definitions of bits in stateFlags:
*/
#define sfAMSresponse	0x001		// zero means request
#define sfAMScommand	0x004		// indicates data exchange via ADS command
#define sfAMSudp	0x400		// indicates transport over UDP (0 mens TCP)

#define maxDataLen 1234

typedef struct _ADSpacket {
    AMS_TCPheader	adsHeader;
    AMSheader		amsHeader;
    char data[maxDataLen];
} ADSpacket;


typedef struct _ADSdeviceInfo {
    u32 ADSerror;
    uc majorVersion;	// major version number
    uc minorVersion;	// minor version number
    us build;		// build number
    char name[16];	// rely on 16 byte, not terminating #0
} ADSdeviceInfo;

typedef struct _ADSreadRequest {
    u32 indexGroup;	// index Group the data belongs to
    u32 indexOffset;	// offset within index Group 
    u32 length;		// length in bytes to be read
} ADSreadRequest;

/*
    Definitions for index Group:
*/
#define igADSfirstReserved	0x0000
#define igADSlastReserved	0x0FFF
#define igADSParameter		0x1000
#define igADSState		0x2000
#define igADSdevFunction	0x3000
#define igADScommonFunction	0x4000	//contains functions to access PLC process data image
#define igADSfirstReservedExt	0x6000	// first iG reserved for extensions
#define igADSlastReservedExt	0xEFFF	// last iG reserved for extensions
#define igADSfirstTwinCATsysService	0xF000	// general TwinCAT ADS system services...
#define igADSlastTwinCATsysService	0xFFFF	// ...containing functions to access PLC I/O image


#define errADSok	0x0000
#define errADSInternal	0x0001
#define errADSNoRuntime	0x0002
#define errADSAllocLockedMemory	0x0003


#define ADS_TRANS_NOTRANS	0x0000
#define ADS_TRANS_CLIENTCYCLE	0x0001
#define ADS_TRANS_CLIENT1REQ	0x0002
#define ADS_TRANS_SERVERCYCLE	0x0003
#define ADS_TRANS_SERVERONCHA	0x0004

typedef enum nAdsTransmode {
    ADSTRANS_NOTRANS=0,
    ADSTRANS_CLIENTCYCLE=1,
    ADSTRANS_CLIENT1REQ=2,
    ADSTRANS_SERVERCYCLE=3,
    ADSTRANS_SERVERONCHA=4
} ADSTRANSMODE;    

typedef struct _ADSreadResponse {
    u32 result;		// error code
    u32 length;		// length in bytes of response
    char data[maxDataLen];
} ADSreadResponse;

typedef struct _ADSwriteRequest {
    u32 indexGroup;
    u32 indexOffset;
    u32 length;		// length in bytes of response
    char data[maxDataLen];
} ADSwriteRequest;

typedef struct _ADSwriteResponse {
    u32 result;		// error code
} ADSwriteResponse;


typedef struct _ADSstateResponse {
    u32 result;
    us	ADSstate;
    us devState;
} ADSstateResponse;

typedef struct _ADSwriteControlRequest {
    us	ADSstate;
    us devState;
    u32 length;
    uc data[];
} ADSwriteControlRequest;


typedef struct _ADSaddNotificationRequest {
    u32 indexGroup;	// index Group the data belongs to
    u32 indexOffset;	// offset within index Group 
    u32 length;		// length in bytes to be read
    u32 transmissionMode; //
    u32 maxDelay;	// after this time the notification will be called up.
			// unit is 100ns
    u32 cycleTime;	// after this time the server teests whether values did change
			// unit is 100ns			
    uc reserved[16];
} ADSaddDeviceNotificationRequest;


typedef struct _ADSaddNotificationResponse {
    u32 result;
    u32 notificationHandle;	// offset within index Group 
} ADSaddDeviceNotificationResponse;


typedef struct _ADSreadWriteRequest {
    u32 indexGroup;
    u32 indexOffset;
    u32 readLength;		// length in bytes of response
    u32 writeLength;		// length in bytes of response
    char data[maxDataLen];
} ADSreadWriteRequest;

typedef struct _ADSreadWriteResponse {
    u32 result;		// length in bytes of response
    u32 length;		// length in bytes of response
    char data[maxDataLen];
} ADSreadWriteResponse;

/*
    Library specific stuff:
*/

#define LINUX

#ifdef LINUX
typedef struct _osst{
    int rfd;
    int wfd;
} osst;
#define _ADSOSserialType osst

#else    
#error Fill in what you need for your OS or API.
#endif

extern int ADSDebug;

#define ADSDebugOpen 0x10
#define ADSDebugPacket 0x20
#define ADSDebugByte 0x40
#define ADSDebugAnalyze 0x80
#define ADSDebugPrintErrors 0x8
#define ADSDebugAll ADSDebugOpen|ADSDebugPacket|ADSDebugByte|ADSDebugAnalyze|ADSDebugPrintErrors



#define ADSPortLogger 100
#define ADSPortRTCore 200
#define ADSPortIO 300
#define ADSPortReserved 400
#define ADSPortNC 500

#define ADSPortBC 800

#define AMSPORT_RO_PLC_RTS0 801
#define AMSPORT_RO_PLC_RTS1 811
#define AMSPORT_RO_PLC_RTS2 821
#define AMSPORT_RO_PLC_RTS3 831

#define ADSPortPCRuntime1 801
#define ADSPortPCRuntime2 811
#define ADSPortPCRuntime3 821
#define ADSPortPCRuntime4 831

#define ADSPortCamshaftController 900

/* 
    This is a wrapper for the serial or network interface. 
*/
typedef struct {
    _ADSOSserialType fd; // some handle for the serial/network interface
    int error;	// Set when read/write errors occur. You will have to do something
		// specific to your OS to make transort work again.
    int timeout;// Timeout in microseconds used in transort.
    char * name;// this	name is used in error outut, so you can identify the interface	
    AMSNetID me;
    int AMSport;
}
ADSInterface;

/* 
    This holds data for a connection;
*/
typedef struct {
    ADSInterface * iface; // pointer to used interface
    int AnswLen;	// length of last message
    int invokeID;	// packetNumber in transport layer
    uc * dataPointer;	// pointer to result data, if present
    uc msgIn[maxDataLen];
    uc msgOut[maxDataLen];
    AMSNetID partner;
    int AMSport;
    int invokeId;
}
ADSConnection;
/* 
    This will setup a new interface structure from an initialized
    serial interface's handle and a name.
*/
ADSInterface * ADSNewInterface(_ADSOSserialType nfd, AMSNetID me, int port, char * nname);

/* 
    This will setup a new connection structure using an initialized
    ADSInterface and PLC's MPI address.
*/
ADSConnection * ADSNewConnection(ADSInterface * di,AMSNetID partner, int port);

/*
    Hex dump:
*/
void _ADSDump(char * name, void * v,int len);


void _ADSDumpAMSNetId(AMSNetID * id);

void _ADSDumpAMSheader(AMSheader * h);

/*
    Naming
*/
char * ADSCommandName(int c);

char * ADSerrorText(int err);

int _ADSReadOne(ADSInterface * di, uc *b);

int _ADSReadPacket(ADSInterface * di,uc *b);

void analyze(uc * p1);

int ADSreadBytes(ADSConnection *dc, int indexGroup, int offset, int length, void * buffer);
int ADSreadDeviceInfo(ADSConnection *dc);
int ADSwriteBytes(ADSConnection *dc, int indexGroup, int offset, int length, void * data);
int ADSreadState(ADSConnection *dc, int * ADSstate, int * devState);
int ADSwriteControl(ADSConnection *dc, int ADSstate, int devState, void * data, int length);
int ADSaddDeviceNotification(ADSConnection *dc, 
    int indexGroup, int offset, int length,
    int transmissionMode, int maxDelay, int cycleTime);

int AdsPortOpen();
int AdsGetLocalAddress(PAmsAddr p);
int _ADSwrite(ADSConnection *dc);

int ADSparseNetID(const char * NetIDstring, AMSNetID * id);
#endif
/*
    Changes:
    03/12/2005 	added ADSparseNetID
*/
