/*
 Definitions of basic functions for a PLC driver for the SIEMENS S5 
 series via its programmer port.
 See www.runmode.com for information about the underlying AS511
 protocol and a proposal of circuit to connect the current loop programmer
 port to a common V24.

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#ifndef __s5l
#define __s5l

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "mytypes.h"
#include "cellList.h"
#include "cell.h"
#include <termios.h>
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include "s5.h"

#define uc unsigned char

inline uc hi(word w)
{
    return (w / 256);
}
inline uc lo(word w)
{
    return (w % 256);
}

word _swap(word w);

void sendChar(int fd, uc c);	// sends a single character
void sendData(int fd, uc n);	// sends a single character, treats 0x10
void s1006(int fd);		// sends the frequently used sequence 0x10 0x06
void s1004(int fd);		// sends the frequently used sequence 0x10 0x04
/*
   Send a 16bit word. SIEMENS S5 uses big endian.
*/ 
void sendWordData(int fd, word w);

bool readChars(int fd, uc *target, size_t size);	// read s block of chars
/* 
 * verifys a block of chars, expects repitition of 0x10 as in data,
 * if cntrl is not set
*/
bool testChars(int fd, uc *target, int len, bool cntrl);
/*
  sBreak inteerupts the cummunication in progress, waits 500 ms
  if the PLC might stil send something and empties the receive
  buffer then.
*/
void sBreak(int rfd, int wfd, word lev);

/*
  Prolog to start function number n on a S5 PLC:
  
  PC to PLC: 0x02                                                      
  PLC to PC: 0x10 0x06                                                   
  PC to PLC: function number
  PLC to PC: 0x02           
  PC to PLC: 0x10 0x06
  PLC to PC: 0x16 0x10 0x03
  PC to PLC: 0x10 0x06
  
  ...execution of function n...
  
  Epilog after execution:
    
  PLC to PC: 0x10 0x03
  PC to PLC: 0x10 0x06
  PLC to PC: 0x02
  PC to PLC: 0x10 0x06
  PLC to PC: 0x12 0x10 0x03
  PC to PLC: 0x10 0x06

*/ 
bool prolog(int rfd, int wfd, unsigned char n);
bool epilog(int rfd, int wfd);
//
// Get or verify the systems parameter block. Verify is done to be sure that
// the connection to PLC is working. The addresses of internal data are used
// to access inputs, outputs and flags
//
bool getSysPars(int rfd, int wfd, sysPars * sp, bool comp);

bool getBlockInfo(int rfd, int wfd, uc type, uc n, blockInfo * BI);
bool setPLCvar(char ot, word address, word value, sysPars * sp, int rfd, int wfd);
bool readData(int rfd, int wfd, word sa, word count, uc *target);

#endif				/* _s5l_ */
