/*
 Implementation of a recursive parser which shifts cell coordinates
 when copying formulas. 

 Part of VISUAL, a human machine interface and data acquisition program
 
 (C) Thomas Hergenhahn (thomas.hergenhahn@web.de) 2001.

 VISUAL is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 VISUAL is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Visual; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
*/
#include <stdlib.h>
#include "pString.h"
#include "lib0.h"
#include "ptokens.h"

#include "vi_string.h"

#include <stdio.h>
#include "plclog.h"
#define ThisModule "shifter : "
#define debug 0

extern void rparse(string & s, int colShift, int rowShift);

int sop(string & s, string op, int colShift, int rowShift)
{
    int pos = s.Find(op);
    if (pos >= 0) {
#if debug >5
	LOG2(ThisModule"pos: %d\n",pos);
#endif
	string sr = s.from( (int) (pos + op.length()));
	rparse(sr, colShift, rowShift);
	string sl = s.before(pos);
	rparse(sl, colShift, rowShift);
	s = sl + op + sr;
	return (1);
    }
    return (0);
}

int sof(string & s, string op, int colShift, int rowShift)
{
    int pos = s.Find(op);
    if (pos >= 0) {
#if debug >5
	LOG2(ThisModule"pos: %d\n",pos);
#endif
	string sr = s.from( (int) (pos + op.length()));
	rparse(sr, colShift, rowShift);
	s = op + sr;
	return (1);
    }

    return (0);
}


void rparse(string & s, int colShift, int rowShift)
{
#if debug >0
    LOG2(ThisModule"rparse line: %s\n",s.c_ctr());
    FLUSH;
#endif
    if (s.length() == 0)
	return;
    int pos;
    int brLev = 0;
    int i;
    char c;
    coord co;
    string sl, sr, si;

    pos = s.Find('(');
#if debug >0
    LOG1(ThisModule"position of first '(': %d\n",pos);
    FLUSH;
#endif
    if (pos >= 0) {
	brLev = 1;		// brLev is 1, 1st open '('
	i = pos;		// search for ')' from here
	do {
	    i++;		// nchstes Zeichen 
	    c = s[i];
	    switch (c) {
	    case ')':
		brLev--;
		break;
	    case '(':
		brLev++;
		break;
	    }
	}
	while ((brLev != 0) && ((unsigned int) i < s.length()));	// closing parenthesis corresponds to 1st open one
	
// brake into substrins and evaluate separately, beginnig with the inner
// expression:
#if debug >0
        LOG1(ThisModule"calling rparse inner\n");
	FLUSH;
#endif
	si = s.substr(pos + 1, i - pos - 1);
	rparse(si, colShift, rowShift);
	
// evaluate left subexpression:
	    if (pos > 0) {
#if debug >0
	    LOG2(ThisModule "calling rparse left %d\n",pos);
	    FLUSH;
#endif
	    sl = s.before(pos);
	    rparse(sl, colShift, rowShift);
	}
#if debug >0
	else {
	    LOG2(ThisModule"nothing left of %d\n",pos);
	    FLUSH;
	}
#endif
#if debug >0
// evaluate right subexpression:

	LOG1(ThisModule "calling rparse right\n");
	FLUSH;
#endif
	sr = s.after(i);	// 02/07/02
	rparse(sr, colShift, rowShift);

	s = sl + '(' + si + ')' + sr;
	return;		// all substrings done, leave rparse
    }
    
#if debug >0
    LOG1(ThisModule "rparse after processing brackets: %s\n" s.c_str());
    FLUSH;
#endif

    if (sop(s, "<>", colShift, rowShift))
	return;
    if (sop(s, "=", colShift, rowShift))
	return;
    if (sop(s, "<", colShift, rowShift))
	return;
    if (sop(s, ">", colShift, rowShift))
	return;

    // colon separated arguments:
    pos = s.Find(',');
    if (pos >= 0) {
	sl = s.before(pos);
	rparse(sl, colShift, rowShift);
	sr = s.after(pos);	// 02/07/02
	rparse(sr, colShift, rowShift);
	s = sl + ',' + sr;	// 02/07/02
	return;
    }
    if (sop(s, "+", colShift, rowShift))
	return;
    if (sop(s, "-", colShift, rowShift))
	return;
    if (sop(s, "*", colShift, rowShift))
	return;
    if (sop(s, "/", colShift, rowShift))
	return;

    if (sop(s, "MOD", colShift, rowShift))
	return;
    if (sop(s, "AND", colShift, rowShift))
	return;
    if (sop(s, "XOR", colShift, rowShift))
	return;
    if (sof(s, "ERROR", colShift, rowShift))
	return;
    if (sop(s, "OR", colShift, rowShift))
	return;

    if (sof(s, "SIN", colShift, rowShift))
	return;
    if (sof(s, "COS", colShift, rowShift))
	return;
    if (sof(s, "EXP", colShift, rowShift))
	return;	
    if (sof(s, "TAN", colShift, rowShift))
	return;
    if (sof(s, "ASIN", colShift, rowShift))
	return;
    if (sof(s, "ACOS", colShift, rowShift))
	return;
    if (sof(s, "ATAN", colShift, rowShift))
	return;
    if (sof(s, "INT", colShift, rowShift))
	return;
    if (sof(s, "SECONDS", colShift, rowShift))
	return;
    if (sof(s, "SEKUNDEN", colShift, rowShift))
	return;
    if (sof(s, "MINUTES", colShift, rowShift))
	return;
    if (sof(s, "MINUTEN", colShift, rowShift))
	return;
    if (sof(s, "MAX", colShift, rowShift))
	return;
    if (sof(s, "MIN", colShift, rowShift))
	return;
    if (sof(s, "SQR", colShift, rowShift))
	return;
    if (sof(s, "SQRT", colShift, rowShift))
	return;
    if (sof(s, "NOISE", colShift, rowShift))
	return;
    if (sof(s, "TIME", colShift, rowShift))
	return;
    if (sof(s, "STATE", colShift, rowShift))
	return;
    if (sof(s, "PREV", colShift, rowShift))
	return;

#if debug >0
    LOG2(ThisModule "rparse after processing operator and functions: %s\n",s.c_str());
    FLUSH;	
#endif
//  
// here we should have left cell coordinates and numbers only:
//
    pos = s.Find(':');		// has it at sheet prefix ?
    if (pos >= 0) {
	s = s.after(pos);
    }
    if ((s[0] > '9') || (s[0] == '$')) {
	string tl = s;
	bool doColShift = (s[0] != '$');
	
//
// when I didn't cast the result from find to int, it gave me an unsigned!
//    
	    bool doRowShift = ((int) (string(s.from(1)).Find('$')) < 0);
//    	LOG4(ThisModule" %s %d %d\n" ,s.substr(1),doRowShift,(int)(s.substr(1).Find('$')));
	    co = getPos(s);
	if (doColShift)
	    co.col += colShift;
	if (doRowShift)
	    co.row += rowShift;
	s = cellName(co);
	if (!doRowShift) {
	    LOG3(ThisModule"%s  %d\n",s.from(1).c_str(),doRowShift);
	    i = 0;
	    while ((s[i] >= 'A') && (s[i] <= 'Z')) {
		i++;
	    }
/*	    
	    s.insert(i, "$");
#else
*/
	    s = s.before(i) + "$" + s.after(i);
	}
	if (!doColShift)
#ifndef use_string_2
	    s = '$' + s;
#else
	    s.prepend('$');
#endif
#if debug >0
	LOG3(ThisModule"shifted: %s --> %s\n",tl.c_str(),s.c_str());
#endif
#if debug >4
	LOG1(ThisModule"cell\n");
#endif
/*      all = all + "PUSHC " + s + " ";
 *      p += ptCELL;
 *      p += getPos (s);
*/
	return;
    } else {
#if debug >4
	LOG1(ThisModule"number\n");	
#endif
//      all = all + "PUSHN " + s + " ";
//      p += ptDOUBLE;
//      p += atof (s.c_str ());
    }
}


void shift(string & s, int c, int r)
{
#if debug >0
    LOG4("shift %s %d %d\n");
#endif
    s = upperString(s);
    rparse(s, c, r);
}

/*
    Changes:
    05/06/2002  STATE() and TIME() functions added.
    06/20/2002 	* got rid of warning about signed/unsigned
		  if string index is SIGNED, why the hack is length() UNSIGNED?
    12/14/2002 	* replaced cout<< by LOGx(
		* adaptions for gcc 3.0.4		  
    07/22/2003 	* added missing function EXP		
    
*/
